-----------------------------------------------------------------------------
-- | 
-- Module      :  LRSX.Language.Alpha 
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--
--  Alpha-Renaming  and Alpha-Equivalence
--
-----------------------------------------------------------------------------
module LRSX.Language.Alpha where
import Data.List
import Data.Function
import Data.Maybe
import LRSX.Util.Util
import LRSX.Language.Syntax
import qualified Data.Map.Strict as Map
import qualified Data.Set as Set
import Debug.Trace
import LRSX.Unification.UnificationProblem
import LRSX.Matching.ConstraintInclusion
import LRSX.Interface.Parser
import LRSX.Interface.Lexer
mytrace a b = b



-- ============================================================================= 
-- Simplification of alpha-renamings (after they have been applied)
-- using the constraint base to remove unneccary meta alpha renamings
-- ============================================================================= 

-- Representing Domains:
-- data Dom = Vars SyntaxComponent         -- All Variables of a Component
--          | FreeVars SyntaxComponent     -- Free Variables of a Component
--          | Cod SubstItem                -- Codomain of a fresh renaming 
--          | SingleVar VarLift            -- A single variable
--          | SingleVarCod VarLift         -- A single variable which is fresh
--  deriving(Eq,Show)

 
 
 
-- Checking the influence of substituions to other syntax parts
-- canRename takes 
--  * a SubstItem  e (a part of a substitution)
--  * a domain of type Dom
--  * the current alpha renaming, to 'know' which names are fresh
--  * and a knowledge base (Splitted Non-Capture Constraints)
-- and checks whether the SubstItem may rename the Domain
-- the inference is based on the one hand on the Domain and on the other hand on the knowledge base
--  e.g. for the domain:
--     * sub can influence domain if domain is the codomain of a fresh renaming, i.e. a Cod ... or a SingleVarCod ...
--       and for the knowledge base, it can be inferred that if sub renames only bound variables (e.g. Let, or CapVars),
--       and there is an NCC ensuring that Dom is not captured by the corresponding renamed construct, then the 
--       substitution sub cannot rename dom.


-- canRename :: SubstItem -> Dom -> SplittedNCCs -> Bool   

canRename e (Vars com) abase  kn        = canRenameExt e com abase kn
-- canRename e (FreeVars com) abase kn     = canRenameExt e com abase kn
canRename e (SingleVar v) abase kn      = canRenameExt e (SCVar v) abase kn 
canRename e (Cod com) abase kn          = False 
canRename e (SingleVarCod com) abase kn = False 

canRenameExt (SubstName (LetVars e) _) comp abase splittedNccs =
 case Map.lookup comp splittedNccs of 
  Nothing -> True 
  Just set -> if e `elem` set then False else True

--  
-- A-Contexts do not contain binders which capture the whole  (note that T-context can capture the hole, since letrec is included)
canRenameExt (SubstName (CapVars (SCCtxt (MetaCtxt sub srt name))) _) com abase splittedNccs 
    | srt `elem` [ACtxt] = False
canRenameExt (SubstName (CapVars (SCCtxt (ConstantCtxt sub srt name))) _) com abase splittedNccs 
    | srt `elem` [ACtxt] = False
--
--    
    
canRenameExt (SubstName (CapVars e) _) comp abase splittedNccs =
 case Map.lookup comp splittedNccs of 
  Nothing -> True 
  Just set -> if e `elem` set then False else True

canRenameExt (SubstName (None e) _) comp abase splittedNccs = True
    
canRenameExt u@(SubstVar v w) comp abase splittedNccs =
        case Map.lookup comp splittedNccs of 
        Nothing -> checkAlphaBase  
        Just set -> if (SCVar v) `elem` set then False else checkAlphaBase
   where
     checkAlphaBase = 
       case comp of 
         (SCVar u) -> 
           null    -- if the set is empty, then True is delivered, otherwise no renaming is possible 
            [() |
              (SCVar x1,SubstVar old new) <- abase
            , new == u -- new is u and is fresh 
            , null [() | (_,SubstVar old1 new1) <- abase, new == old1 ] -- no further renaming of new
            ] 
         _ -> True -- nothing inferred            
 


simplifyEnv  abase knbase env =
 let
   mc' = [simplifyMetaChain abase knbase m | m <- metaChains env]
   cc' = [simplifyConstantChain abase knbase m | m <- constantChains env]
   me' = [simplifyMetaEnv abase knbase m | m <- metaEnvVars env]
   ce' = [simplifyConstantEnv abase knbase m | m <- constantEnvVars env]
   b'  = [simplifyBinding abase knbase b | b <- bindings env]
 in env{bindings=b', constantEnvVars = ce', metaEnvVars = me', constantChains = cc', metaChains = mc'}
 
simplifyConstantEnv abase knbase (e@(Env (Substitution sub) name))  =
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCConstantEnv (Env emptySubstitution name))]
                                 ,[Vars (SCConstantEnv (Env emptySubstitution name))]) sub
 in (Env (Substitution sub') name)
 
simplifyMetaEnv abase knbase (e@(Env (Substitution sub) name))  =
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCMetaEnv (Env emptySubstitution name))]
                                 ,[Vars (SCMetaEnv (Env emptySubstitution name))]) sub
 in (Env (Substitution sub') name)
 
simplifyMetaChain abase knbase (e@(MetaChain (Substitution sub) typ name z s)) = 
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCMetaChain (emptySubstitution) typ name)]
                                 ,[Vars (SCMetaChain (emptySubstitution) typ name)]) sub
                                        
     s' = simplifyExpr abase knbase s
     z' = simplifyVar abase knbase z
 in (MetaChain (Substitution sub')   typ name z' s')
 
simplifyConstantChain abase  knbase (e@(ConstantChain (Substitution sub) typ name z s)) = 
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCConstantChain (emptySubstitution) typ name)]
                                 ,[Vars (SCConstantChain (emptySubstitution) typ name)]) sub
                                        
     s' = simplifyExpr abase knbase s
     z' = simplifyVar abase knbase z
 in (ConstantChain (Substitution sub')   typ name z' s')
 
simplifyBinding abase knbase (e@(v := r))  =
 let r' = simplifyExpr abase knbase r 
     v' = simplifyVar abase knbase v
 in v' := r'
 
simplifyVar abase knbase (MetaVarLift (Substitution sub) name) = 
    let (v,f,sub') = 
         simplifySub abase knbase ([SingleVar (MetaVarLift emptySubstitution name)],[SingleVar (MetaVarLift emptySubstitution name)]) sub
    in  tryShift abase knbase (MetaVarLift (Substitution sub') name)
    
simplifyVar abase knbase (ConstantVarLift (Substitution sub) name) = 
    let (v,f,sub') = 
         simplifySub abase knbase ([SingleVar (ConstantVarLift emptySubstitution name)],[SingleVar (ConstantVarLift emptySubstitution name)]) sub
    in  tryShift abase knbase (ConstantVarLift (Substitution sub') name)

simplifyVar abase knbase (VarLift (Substitution sub) name) = 
    let (v,f,sub') = 
         simplifySub abase knbase ([SingleVar (VarLift emptySubstitution name)],[SingleVar (VarLift emptySubstitution name)]) sub
    in  tryShift abase knbase (VarLift (Substitution sub') name)
    
-- try shift is used to apply variables renamings after simplyVar was applied,
-- if successful it again calls simplifyVar, since more simplifications may be possible    

tryShift abase knbase (MetaVarLift (Substitution ((SubstVar (MetaVarLift (Substitution []) v1) (MetaVarLift (Substitution []) v2)):sub)) v3)
  | v1 == v3 =  simplifyVar abase knbase (MetaVarLift (Substitution sub) v2)

  
tryShift abase knbase (ConstantVarLift (Substitution ((SubstVar (ConstantVarLift (Substitution []) v1) (ConstantVarLift (Substitution []) v2)):sub)) v3)
  | v1 == v3 =  simplifyVar abase knbase (ConstantVarLift (Substitution sub) v2)
tryShift abase knbase (VarLift (Substitution ((SubstVar (VarLift (Substitution []) v1) (VarLift (Substitution []) v2)):sub)) v3)
  | v1 == v3 =  simplifyVar abase knbase (VarLift (Substitution sub) v2)

-- cases if the substitution is a set  
  
tryShift abase knbase (MetaVarLift (Substitution ((SubstSet set):sub)) v3) 
   | ((SubstVar (MetaVarLift (Substitution []) vv1) (MetaVarLift (Substitution []) vv2),rrest):_) <-
        [el | el@(SubstVar (MetaVarLift (Substitution []) v1) (MetaVarLift (Substitution []) v2),rest) <- splits set, v1 == v3]
   = case rrest of 
             [] ->  simplifyVar abase knbase (MetaVarLift (Substitution sub) vv2)
             _  -> simplifyVar abase knbase (MetaVarLift (Substitution ((SubstSet rrest):sub)) vv2)
  
tryShift abase knbase (ConstantVarLift (Substitution ((SubstSet set):sub)) v3) 
   | ((SubstVar (ConstantVarLift (Substitution []) vv1) (ConstantVarLift (Substitution []) vv2),rrest):_) <-
        [el | el@(SubstVar (ConstantVarLift (Substitution []) v1) (ConstantVarLift (Substitution []) v2),rest) <- splits set, v1 == v3]
   = case rrest of 
             [] ->  simplifyVar abase knbase (ConstantVarLift (Substitution sub) vv2)
             _  -> simplifyVar abase knbase (ConstantVarLift (Substitution ((SubstSet rrest):sub)) vv2)
             
tryShift abase knbase (VarLift (Substitution ((SubstSet set):sub)) v3) 
   | ((SubstVar (VarLift (Substitution []) vv1) (VarLift (Substitution []) vv2),rrest):_) <-
        [el | el@(SubstVar (VarLift (Substitution []) v1) (VarLift (Substitution []) v2),rest) <- splits set, v1 == v3]
   = case rrest of 
             [] ->  simplifyVar abase knbase (VarLift (Substitution sub) vv2)
             _  -> simplifyVar abase knbase (VarLift (Substitution ((SubstSet rrest):sub)) vv2)
  
tryShift abase knbase v = v 



simplifyExpr abase knbase (VarL v) = VarL (simplifyVar abase knbase v)
 
simplifyExpr  abase  knbase (e@(MetaExpression (Substitution sub) name))  =
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCExpr (MetaExpression emptySubstitution name))]
                                 ,[Vars (SCExpr (MetaExpression emptySubstitution name))]) sub
 in (MetaExpression (Substitution sub') name)
 
simplifyExpr abase knbase (e@(Constant (Substitution sub) name))  =
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCExpr (Constant emptySubstitution name))]
                                 ,[Vars (SCExpr (Constant emptySubstitution name))]) sub
 in (Constant (Substitution sub') name)

simplifyExpr abase knbase (e@(Fn f args))
 = Fn f $ map (simplifyArgPos abase knbase) args
 
simplifyExpr abase knbase (e@(Letrec env expr))
 = Letrec (simplifyEnv abase knbase env) (simplifyExpr abase knbase expr)

simplifyExpr abase knbase (e@(MetaCtxtSubst (Substitution sub) typ name s)) =
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCCtxt (MetaCtxt emptySubstitution typ name))]
                                 ,[Vars (SCCtxt (MetaCtxt emptySubstitution typ name))]) sub
 in  (MetaCtxtSubst (Substitution sub') typ name (simplifyExpr abase knbase s))
 
simplifyExpr abase knbase (e@(ConstantCtxtSubst (Substitution sub) typ name s)) =
 let (v,f,sub') =
         simplifySub abase knbase ([Vars (SCCtxt (ConstantCtxt emptySubstitution typ name))]
                                 ,[Vars (SCCtxt (ConstantCtxt emptySubstitution typ name))]) sub
 in  (ConstantCtxtSubst (Substitution sub') typ name (simplifyExpr abase knbase s))
     
simplifyExpr abase knbase (Hole (Substitution sub)) =
 let (v,f,sub') =
         simplifySub abase knbase ([]
                                 ,[]) sub
 in  Hole (Substitution sub')
     
simplifyExpr abase knbase other = error $ "in l.182 Alpha.hs:" ++ show other
 
 
simplifyArgPos abase knbase (e@(Binder v r))  =
 let r' = simplifyExpr abase knbase r 
     v' = map (simplifyVar abase knbase) v
 in Binder v' r'

simplifyArgPos abase knbase (e@(VarPos v)) = (VarPos (simplifyVar abase knbase v))

simplifySub abase knbase (vars,freevars) sub =
   let (a,b,c) = simplifySubOld abase knbase (vars,freevars) sub
       (a',c') = simplifySub' abase knbase vars sub 
   in if a == a' && c == c' then (a,b,c) else error (show (a,b,c,a',c'))

-- simplifySubOld :: Map.Map SyntaxComponent (Set.Set SyntaxComponent) -> ([Dom],[Dom]) -> [SubstItem] -> ([Dom],[Dom],[SubstItem]) 
simplifySubOld abase knbase (vars,freevars) [] = (vars,freevars,[]) 
simplifySubOld abase knbase (vars,freevars) (s@(SubstVar x y):subs) 
 | all (== True) $ map (\a -> not $ canRename (s) a abase knbase) vars 
 = simplifySubOld abase knbase (vars,freevars) subs
 | all (== True) $ map (\a -> not $ canRename (s) a abase knbase) freevars
 =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[SingleVarCod y],freevars) subs
 in  (a,b,s:c)
 | otherwise =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[SingleVarCod y],freevars++[SingleVarCod y]) subs
 in  (a,b,s:c)
 
  
simplifySubOld abase knbase (vars,freevars) ((SubstSet s):subs) =
 let (v',f',set') = simplifySubOld abase knbase (vars,freevars) s
     (v'',f'',subs') = simplifySubOld abase knbase (v',f') subs
 in if null set' then 
      (v'',f'',subs')
      else (v'',f'',(SubstSet set'):subs')
 
simplifySubOld abase knbase (vars,freevars) (s@(SubstName (None e') new):subs) 
--  | (e:_) <- [e | Vars e <- vars, e == e'] =
--   simplifySubOld abase knbase ((delete (Vars e) vars)++[Vars e,Cod s],freevars) (s:subs)
 | otherwise =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[Cod s],freevars++[Cod s]) subs
 in  (a,b,s:c)
 
simplifySubOld abase knbase (vars,freevars) ss@(s@(SubstName (LetVars e') new):subs)
 | all (== True) $ map (\x -> not $ canRename (s) x abase knbase) vars
 = simplifySubOld abase knbase (vars,freevars) subs
 | all (== True) $ map (\x -> not $ canRename (s) x abase knbase) freevars
 =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[Cod s],freevars) subs
 in  (a,b,s:c)
 | otherwise =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[Cod s],freevars++[Cod s]) subs
 in  (a,b,s:c)
 
simplifySubOld abase knbase (vars,freevars) (s@(SubstName (CapVars e') new):subs)
 | all (== True) $ map (\x -> not $ canRename (s) x abase knbase) vars
 = simplifySubOld abase knbase (vars,freevars) subs
 | all (== True) $ map (\x -> not $ canRename (s) x abase knbase) freevars
 =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[Cod s],freevars) subs
 in  (a,b,s:c)
 | otherwise =
 let (a,b,c) =  simplifySubOld abase knbase (vars++[Cod s],freevars++[Cod s]) subs
 in  (a,b,s:c)
 

 
 
 
 
 
 
 
simplifySub' abase knbase vars [] = (vars,[])
simplifySub' abase knbase vars (s@(SubstVar x y):subs) 
 | all (== True) $ map (\a -> not $ canRename (s) a abase knbase) vars 
 = simplifySub' abase knbase vars subs
 | otherwise =
 let (a,c) =  simplifySub' abase knbase (vars++[SingleVarCod y]) subs
 in  (a,s:c)
 
  
simplifySub' abase knbase vars ((SubstSet s):subs) =
 let (v',set') = simplifySub' abase knbase vars s
     (v'',subs') = simplifySub' abase knbase v' subs
 in if null set' then 
      (v'',subs')
      else (v'',(SubstSet set'):subs')
 
simplifySub' abase knbase vars (s@(SubstName (None e') new):subs) 
--  | (e:_) <- [e | Vars e <- vars, e == e'] =
--   simplifySub' abase knbase ((delete (Vars e) vars)++[Vars e,Cod s]) (s:subs)
 | otherwise =
 let (a,c) =  simplifySub' abase knbase (vars++[Cod s]) subs
 in  (a,s:c)
 
simplifySub' abase knbase vars ss@(s@(SubstName (LetVars e') new):subs)
 | all (== True) $ map (\x -> not $ canRename (s) x abase knbase) vars
 = simplifySub' abase knbase vars subs
 | otherwise =
 let (a,c) =  simplifySub' abase knbase (vars++[Cod s]) subs
 in  (a,s:c)
 
simplifySub' abase knbase vars (s@(SubstName (CapVars e') new):subs)
 | all (== True) $ map (\x -> not $ canRename (s) x abase knbase) vars
 = simplifySub' abase knbase (vars) subs
 | otherwise =
 let (a,c) =  simplifySub' abase knbase (vars++[Cod s]) subs
 in  (a,s:c)
 
 
-- for debugging:    
-- simplifySub abase knbase (vars,freevars) sub = (vars,freevars,sub) 
  
-- ====================================================================       
-- Meta-Alpha-Renaming
-- ====================================================================       
-- renameExpr takes a list of fresh names and an expression,
-- it performs meta alpha-renaming of the expression, by
-- introducing new meta renamings
-- the output is a triple (fresh,expr',subst)
-- where 
--  * fresh are the unused fresh names
--  * expr' is the renamed expr
--  * subst is the substitution of variables by fresh variables and 
--    meta variables by meta variables and a alpha renaming
--    which can be seen as a 'making fresh constraint', 
--    e.g. S |-> S_{alpha} means, that alpha makes the bound variables
--    of S fresh.
--    Note that in the expression S may be replaced something 
--    like S_{alpha,furtherrenamings} if S is in scope of other renamings
--    introduced by making fresh constraints    
  
renameExpr :: NonCaptureConstraints -> [String] -> Expression -> ([String],Expression,[(SyntaxComponent,SubstItem)])        
renameExpr nccs fresh (VarL var) = (fresh,(VarL var),[])
        
renameExpr nccs (s':fresh) e@(MetaExpression substold s) =
  let
    newExpr  = (MetaExpression subst s) 
    subst    = (substold{substList=(substList substold) ++ [SubstName (None $ SCExpr e)  ("ALPHA"++s')]})
    renaming = [(SCExpr (MetaExpression substold s),SubstName (None $ SCExpr e) ("ALPHA"++s'))]
  in (fresh,newExpr,renaming)

renameExpr nccs (s':fresh) e@(Constant substold s) =
  let
    newExpr = (Constant subst s) 
    subst   = (substold{substList=(substList substold) ++ [SubstName (None $ SCExpr e) ("ALPHA"++s')]})
    renaming =  [(SCExpr  $ Constant substold s, SubstName (None $ SCExpr e) ("ALPHA"++s'))]
  in (fresh,newExpr,renaming)
  
renameExpr nccs fresh (MetaCtxtSubst substold typ name s) =
 let
  (f:fresh',s',renaming) = renameExpr nccs  fresh s
  subst= (substold{substList= (substList substold) ++ [SubstName (None $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)]})
  substNew = Substitution{substList = [SubstName (None $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)]}
  substNewCV = Substitution{substList = [SubstName (CapVars $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)]}
  renaming' = (SCCtxt (MetaCtxt substold typ name), SubstName (None $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)):renaming
  s'' = siftDownWithScope nccs []  substNewCV s'
 in (fresh', (MetaCtxtSubst subst typ name s''), renaming')                

renameExpr nccs fresh (ConstantCtxtSubst substold typ name s) =
 let
  (f:fresh',s',renaming) = renameExpr nccs  fresh s
  subst= (substold{substList= (substList substold) ++ [SubstName (None $ SCCtxt (ConstantCtxt substold typ name)) ("ALPHA"++f)]})
  substNew = Substitution{substList = [SubstName (None $ SCCtxt (ConstantCtxt substold typ name)) ("ALPHA"++f)]}
  substNewCV = Substitution{substList = [SubstName (CapVars $ SCCtxt (ConstantCtxt substold typ name)) ("ALPHA"++f)]}
  renaming' = (SCCtxt (ConstantCtxt substold typ name),SubstName (None $ SCCtxt (ConstantCtxt substold typ name)) ("ALPHA"++f)):renaming
  s'' = siftDownWithScope nccs []  substNewCV s'
 in (fresh', (ConstantCtxtSubst subst typ name s''), renaming')                

renameExpr nccs  fresh (MetaCtxt substold typ name) =
 let
  (f:fresh',s',renaming) = renameExpr nccs  fresh (Hole substold)
  subst= (substold{substList= (substList substold) ++ [SubstName (None $ (SCCtxt (MetaCtxt substold typ name))) ("ALPHA"++f)]})
  substNew = Substitution{substList = [SubstName (None $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)]}
  substNewCV = Substitution{substList = [SubstName (CapVars $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)]}
  renaming' = (SCCtxt (MetaCtxt substold typ name), SubstName (None $ SCCtxt (MetaCtxt substold typ name)) ("ALPHA"++f)):renaming
  s'' = siftDownWithScope nccs []  substNewCV s'
 in (fresh', (MetaCtxtSubst subst typ name s''), renaming')                

renameExpr nccs  fresh (ConstantCtxt substold typ name) =
 let
  (f:fresh',s',renaming) = renameExpr nccs  fresh (Hole substold)
  subst= (substold{substList= (substList substold) ++ [SubstName (None $ (SCCtxt (ConstantCtxt substold typ name))) ("ALPHA"++f)]})
  substNew = Substitution{substList = [SubstName (None $ (SCCtxt (ConstantCtxt substold typ name))) ("ALPHA"++f)]}
  substNewCV = Substitution{substList = [SubstName (CapVars $ (SCCtxt (ConstantCtxt substold typ name))) ("ALPHA"++f)]}
  renaming' = ((SCCtxt (ConstantCtxt substold typ name)),SubstName (None (SCCtxt (ConstantCtxt substold typ name))) ("ALPHA"++f) ):renaming
  s'' = siftDownWithScope nccs []  substNewCV s'
 in (fresh', (ConstantCtxtSubst subst typ name s''), renaming')                
 
renameExpr nccs  fresh (Letrec env s) =
 let 
  mc = metaChains env
  cc = constantChains env
  me = metaEnvVars env
  ce = constantEnvVars env
  -- Bindings
  (vars1,expr1) = unzip [(x,e) | x:=e <- bindings env]
  (f1,expr1',ren1) = renameList nccs fresh expr1
  vars1' = mkVar vars1 (map ("X" ++) $ take (length vars1) f1)
  subst1 = zipWith SubstVar vars1 vars1'
  ren1' = [(SCVar vold, SubstVar vold vnew) | (vold,vnew) <- zip vars1 vars1']
  f2     = drop (length vars1) f1
  binds' =  zipWith 
                (\((oldv,oldvs),e) (v,vs) ->
                 v := (siftDownWithScope nccs [] 
                            (Substitution [SubstSet 
                                            (   
                                                [SubstVar oldv v]
                                                ++(zipWith SubstVar oldvs vs)
                                                ++ subst2 ++ subst3 ++ subst4 ++ subst5
                                                 
                                            )]) e))
                (zip (splits vars1) expr1')
                (splits vars1')
  -- MetaChains
  (vars2,expr2) = unzip [(z,e) | MetaChain substold typ name z e <- metaChains env]
  (f3,expr2',ren2) = renameList nccs f2 expr2
  metaAlpha1 = map ("ALPHA" ++) $ take (length vars2) f3
  f4 = drop (length vars2) f3
  vars2' = mkVar vars2 $ map ("X" ++) $ take (length vars2) f4
  f5 = drop (length vars2) f4
  substChains2 = [SubstName (None $ SCMetaChain subold typ name) ma 
                  | (ma,MetaChain subold typ name z e) <- zip metaAlpha1 (metaChains env)
                 ]
  substVars2   = zipWith SubstVar vars2 vars2'
  allSubsts2   = [x: (toSubstSet ((map noneToLetVars xs) ++ substVars2 ++ otherSubst2)) | (x,xs) <- splits substChains2]
  
  subst2 = (map noneToLetVars substChains2)++substVars2
  ren2' = [ (SCMetaChain old typ name, SubstName (None $ SCMetaChain old typ name) alp)
           | (MetaChain old typ name _ _, alp) <- zip (metaChains env) metaAlpha1
          ]
          ++ [(SCVar vold, SubstVar vold vnew) | (vold,vnew) <- zip vars2 vars2']
  otherSubst2 =  concat [subst1,subst3,subst4,subst5]          
  mc' = [(MetaChain  Substitution{substList=old ++ new} typ name z' (siftDownWithScope nccs [] subst e'))
           | (MetaChain (Substitution {substList=old}) typ name z e,(new,(z',e'))) <- zip (metaChains env) (zip allSubsts2 (zip vars2' expr2'))]          
          
  -- ConstantChains
  (vars3,expr3) = unzip [(z,e) | ConstantChain substold typ name z e <- constantChains env]
  (f6,expr3',ren3) = renameList nccs f5 expr3
  metaAlpha2 = map ("ALPHA" ++) $ take (length vars3) f6
  f7 = drop (length vars3) f6
  vars3' = mkVar vars3 $ map ("X" ++) $ take (length vars3) f7
  f8 = drop (length vars3) f7
  substChains3 = [SubstName (None $ SCConstantChain subold typ name) ma 
                   | (ma,ConstantChain subold typ name z e) <- zip metaAlpha2 (constantChains env)
                 ]
  substVars3   = zipWith SubstVar vars3 vars3'
  allSubsts3   =  [x:(toSubstSet $ ((map noneToLetVars xs) ++ substVars3) ++ otherSubst3)| (x,xs) <- splits substChains3]
  subst3 = (map noneToLetVars substChains3)++substVars3
  ren3' = [ (SCConstantChain old typ name, SubstName (None $ SCConstantChain old typ name) alp)
           | (ConstantChain old typ name _ _, alp) <- zip (constantChains env) metaAlpha2
          ]
          ++ [(SCVar vold, SubstVar vold vnew) | (vold,vnew) <- zip vars3 vars3']
  otherSubst3 = concat [subst1,subst2,subst4,subst5]          
  cc' = [(ConstantChain  Substitution{substList=old ++ new} typ name z' (siftDownWithScope nccs [] subst e'))
           | (ConstantChain (Substitution {substList=old}) typ name z e,(new,(z',e'))) <- zip (constantChains env) (zip allSubsts3 (zip vars3' expr3'))]          
  -- MetaEnvs
  envVars = metaEnvVars env
  metaAlpha3 = map ("ALPHA" ++) $ take (length envVars) f8
  f9 = drop (length envVars) f8
  substEnvs4 = [ SubstName (None $ SCMetaEnv (Env subold name)) ma 
                  | (ma,Env subold name) <- zip metaAlpha3 (metaEnvVars env)
               ]
  subst4 = (map noneToLetVars substEnvs4)
  ren4' = [(SCMetaEnv (Env sub name), SubstName (None $ SCMetaEnv (Env sub name)) alp) | ((Env sub name),alp) <- zip envVars metaAlpha3]
  allSubsts4 = [x:(toSubstSet $ ((map noneToLetVars xs) ++ otherSubst4) ) | (x,xs) <- splits substEnvs4]
  otherSubst4 = concat [subst1,subst2,subst3,subst5]          
  envVars' = 
     [ Env (Substitution (old++xs) ) name
     | xs@((SubstName (None (SCMetaEnv (Env (Substitution old) name))) ma):rest) <- allSubsts4
     ]
  -- ConstantEnvs
  cenvVars = constantEnvVars env
  metaAlpha4 = map ("ALPHA" ++) $ take (length cenvVars) f9
  f10 = drop (length cenvVars) f9
  substEnvs5 = [SubstName (None $ SCConstantEnv (Env subold name)) ma | (ma,Env subold name) <- zip metaAlpha4 (constantEnvVars env)]
  subst5 = map noneToLetVars substEnvs5
  ren5' = [ (SCConstantEnv (Env sub name), SubstName (None $ SCConstantEnv (Env sub name)) alp) | ((Env sub name),alp) <- zip cenvVars metaAlpha4]
  allSubsts5 = [x:(toSubstSet $ (map noneToLetVars xs) ++ otherSubst5) | (x,xs) <- splits substEnvs5]
  otherSubst5 = concat [subst1,subst2,subst3,subst4]          
  cenvVars' = 
     [ Env (Substitution (old++xs) ) name
     | xs@((SubstName (None (SCConstantEnv (Env (Substitution old) name))) ma):rest) <- allSubsts5
     ]
  -- InExpr  
  (f11,s',ren4)    = renameExpr nccs  f10 s 
  s'' = siftDownWithScope nccs [] subst s'
  subst = Substitution{substList = (toSubstSet $ concat [subst1,subst2,subst3,subst4,subst5])}
  env' = env{bindings=binds', metaChains=mc', constantChains=cc',metaEnvVars=envVars',constantEnvVars=cenvVars'}
  renaming' = ren1 ++ ren2 ++ ren3 ++ ren4  ++  ren1' ++ ren2' ++ ren3' ++ ren4' ++ ren5'
 in  
     (f11,Letrec env' s'',renaming')                       

renameExpr nccs  fresh (Fn f args) =
    let (fresh',args',ren) = renameArgPosList nccs fresh args 
    in (fresh',Fn f args',ren)
 
renameExpr nccs  fresh (Hole subs) = 
  (fresh,Hole subs,[]) 

  
renameArgPos nccs fresh (Binder ys s) =
 let
  (fresh',s',renaming) = renameExpr nccs  fresh s
  newnames = [(varPrefix y)  ++ name| (name,y) <-  (zip (take (length ys) fresh') ys)]
  fresh'' = drop (length ys) fresh'
  varSubst = zipWith (\old new -> case old of 
                                    VarLift substold a -> (old,VarLift emptySubstitution new)
                                    MetaVarLift substold a -> (old,MetaVarLift emptySubstitution new)
                                    ConstantVarLift substold a -> (old,ConstantVarLift emptySubstitution new)) ys newnames
  newVars = map snd varSubst                                   
  s'' = siftDownWithScope nccs [] (Substitution{substList = map (\(x,y) -> SubstVar x y) (reverse $ varSubst)}) s'
  newexpr = Binder newVars s''  
  renaming' = (reverse $ zipWith (\x y ->  (SCVar x, SubstVar x y)) ys newVars) ++ renaming
 in (fresh'',newexpr,renaming')           

  
renameList nccs fresh [] =
 (fresh,[],[])                  
renameList nccs fresh (e:es) =
 let (f',e',r) = renameExpr nccs  fresh e 
     (f'',es',r') = renameList nccs f' es
 in (f'',e':es', r ++ r')

renameArgPosList nccs fresh [] =
 (fresh,[],[])                  
renameArgPosList nccs fresh (e:es) =
 let (f',e',r) = renameArgPos nccs  fresh e 
     (f'',es',r') = renameArgPosList nccs f' es
 in (f'',e':es', r ++ r')


siftDownWithScopeArgPos :: NonCaptureConstraints -> [String] -> Substitution -> ArgPos -> ArgPos        
siftDownWithScopeArgPos nccs scope sub (Binder v e) = Binder v (siftDownWithScope nccs scope' sub e)
  where scope' = scope ++ (map varName v)
        
siftDownWithScope nccs scope sub (Hole subold) = Hole (append subold sub)        
siftDownWithScope nccs scope sub (Fn f args) = 
 Fn f (map (siftDownWithScopeArgPos nccs scope sub) args)
siftDownWithScope nccs scope sub (MetaExpression subold s) = MetaExpression (append subold sub) s
siftDownWithScope nccs scope sub (Constant subold s) = Constant (append subold sub) s

siftDownWithScope nccs scope sub (MetaCtxtSubst subold typ name s) = 
    MetaCtxtSubst (append subold sub) typ name (siftDownWithScope nccs scope sub s)
    
siftDownWithScope nccs scope sub (ConstantCtxtSubst subold typ name s) =
    ConstantCtxtSubst (append subold sub) typ name (siftDownWithScope nccs scope sub s)
    
siftDownWithScope nccs scope sub (Letrec env s) =
  Letrec
    ( env {
         metaChains = [MetaChain (append subold sub) t ee z (siftDownWithScope nccs scope' sub s) | MetaChain subold t ee z s <- metaChains env]
        ,constantChains = [ConstantChain (append subold sub) t ee z (siftDownWithScope nccs scope' sub s) | ConstantChain subold t ee z s <- constantChains env]
        ,metaEnvVars = [Env (append oldsub sub) name| (Env oldsub name) <- metaEnvVars env]
        ,constantEnvVars = [Env (append oldsub sub) name| (Env oldsub name) <- constantEnvVars env]      
        ,bindings = [x := (siftDownWithScope nccs scope' sub e) | x := e <- bindings env]
    })
    (siftDownWithScope nccs scope' sub s)
 where
   scope' = scope ++ map varName ([z | MetaChain sub t ee z s <- metaChains env]
                                  ++
                                  [z | ConstantChain sub t ee z s <- constantChains env]
                                  ++
                                  [z | z:=_ <- bindings env])

   
siftDownWithScope nccs scope sub (VarL  var)  = VarL (siftDownWithScopeVar nccs scope sub  var)

-- ========================================================================
--
-- this seems to be quite unsafe and wrong:
-- example:
--  \X.C[X]
-- gets first alpha renamed into
-- \X.C_{alphaC}[X_{CV(alphaC)}]
-- and then (after renaming X) into
-- \Y.C_{alphaC,X|->Y}[Y]  which is wrong, since C may capture X,
-- assume that C captures X,
-- then the right renaming would be
-- \X.C[X] -> \Y.C_alphaC[X_CV(alphaC)] 
-- or also
-- \X.C[X] -> \Y.C_alphaC[X_{[CV(alphaC), X |-> Y]}] 
-- thus the second expression is the right one for renaming independently
-- whether C captures or not captures X, perhaps the NCCs then can be used to 
-- simplify the renaming
--
-- A further problem here is how the matching and reductions works, if e.g.
-- letrec X=v in C[X] is copied after alpha-renaming
-- Assume that the NCC (X,C) holds 
-- then alpha-renaming must lead to
--  letrec Y=v in C_{alphaC,X|->Y}[Y]
-- to enable the copy operation, since  for
-- letrec Y=v in C_{alphaC,X|->Y}[CV(alphaC), X |-> Y]}


siftDownWithScopeVar nccs scope sub v
 = let res = siftDownWithScopeVarR nccs scope sub v
   in 
#ifdef DEBUG       
                
                mytrace 
            (("\nSIFTDOWN AT VAR:" ++ show (sub,v))
            ++
            ("\nwith nccs:" ++ show (splitCapCon nccs))
            ++ 
            "\n resulting in " ++ show res) 
#endif            
            res
            
-- -- new: Sep16:            
siftDownWithScopeVarR nccs scope sub (VarLift oldsub v)
   |  v `elem` scope = simplifyVar [] (splitCapCon nccs) $ VarLift oldsub v
siftDownWithScopeVarR nccs scope sub (MetaVarLift oldsub v)
   | v `elem` scope = simplifyVar [] (splitCapCon nccs) $ MetaVarLift oldsub v
siftDownWithScopeVarR nccs scope sub (ConstantVarLift oldsub v)
   | v `elem` scope = simplifyVar [] (splitCapCon nccs) $ ConstantVarLift oldsub v
  
  
siftDownWithScopeVarR nccs scope (Substitution ((SubstVar v1 v2@(VarLift (Substitution sub2) v2')):sub)) v@(VarLift (Substitution []) _)
    | v1 == v = simplifyVar [] (splitCapCon nccs) $ VarLift (Substitution ((sub2++sub))) v2'
siftDownWithScopeVarR nccs scope (Substitution ((SubstVar v1 v2@(MetaVarLift (Substitution sub2) v2')):sub)) v@(MetaVarLift (Substitution []) _)
    | v1 == v = simplifyVar [] (splitCapCon nccs)  $MetaVarLift (Substitution (sub2++sub)) v2'
siftDownWithScopeVarR nccs scope (Substitution ((SubstVar v1 v2@(ConstantVarLift (Substitution sub2) v2')):sub)) v@(ConstantVarLift (Substitution []) _)
    | v1 == v = simplifyVar [] (splitCapCon nccs) $ ConstantVarLift (Substitution (sub2++sub)) v2'
    
siftDownWithScopeVarR nccs scope (Substitution ((SubstSet set):sub)) v@(VarLift (Substitution []) _)
    | (SubstVar v1 v2@(VarLift (Substitution sub2) v2')):set' <- [(SubstVar v1 v2)| (SubstVar v1 v2@(VarLift (Substitution sub2) v2')) <- set, v1 == v]
    =  simplifyVar [] (splitCapCon nccs) $ VarLift (Substitution (sub2++[SubstSet set'] ++ sub)) v2'

siftDownWithScopeVarR nccs scope (Substitution ((SubstSet set):sub)) v@(MetaVarLift (Substitution []) _)
    | (SubstVar v1 v2@(MetaVarLift (Substitution sub2) v2')):set' <- [(SubstVar v1 v2)| (SubstVar v1 v2@(MetaVarLift (Substitution sub2) v2')) <- set, v1 == v]
    =  simplifyVar [] (splitCapCon nccs) $ MetaVarLift (Substitution (sub2++[SubstSet set'] ++ sub)) v2'
    
siftDownWithScopeVarR nccs scope (Substitution ((SubstSet set):sub)) v@(ConstantVarLift (Substitution []) _)
    | (SubstVar v1 v2@(ConstantVarLift (Substitution sub2) v2')):set' <- [(SubstVar v1 v2)| (SubstVar v1 v2@(ConstantVarLift (Substitution sub2) v2')) <- set, v1 == v]
    =  simplifyVar [] (splitCapCon nccs) $ ConstantVarLift (Substitution (sub2++[SubstSet set'] ++ sub)) v2'    

siftDownWithScopeVarR nccs scope sub (VarLift oldsub v)
--   |  Just x <- lookupS v sub = VarLift emptySubstitution x
   
--  |  v `elem` scope = VarLift oldsub v
  | otherwise = (VarLift (append oldsub sub) v)
siftDownWithScopeVarR nccs scope sub (MetaVarLift oldsub v)
--  | Just x <- lookupS v sub = MetaVarLift emptySubstitution x
--  | v `elem` scope = simplifyVar [] (splitCapCon nccs) $ MetaVarLift oldsub v
  | otherwise = simplifyVar [] (splitCapCon nccs) $ (MetaVarLift (append oldsub sub) v)
siftDownWithScopeVarR nccs scope sub (ConstantVarLift oldsub v)
--  | Just x <- lookupS v sub = ConstantVarLift emptySubstitution x
--  | v `elem` scope = simplifyVar [] (splitCapCon nccs) $ ConstantVarLift oldsub v
  | otherwise = simplifyVar [] (splitCapCon nccs) $ (ConstantVarLift (append oldsub sub) v)
--
-- =============================================================================
 
toSubstSet [] = []
toSubstSet [x] = [x]
toSubstSet xs = [SubstSet xs] 
  
deleteRek = deleteRekBy (==   )  
deleteRekBy p a [] = []  
deleteRekBy p a (b@(SubstSet set):rest)
 = case (deleteRekBy p a set) of
    [] -> (deleteRekBy p a rest) 
    xs -> (SubstSet xs):(deleteRekBy p a rest)
deleteRekBy p a (b:rest)
 | p a b = deleteRekBy p a rest
 | otherwise = b:(deleteRekBy p a rest)
  
append subst1 subst2 = subst1{substList = clean $ substList subst1 ++ substList subst2}
  where
   clean [] = []
   clean ((SubstSet s):xs) =  (SubstSet (clean s)):(clean (deleteFn (SubstSet s) xs))
   clean ((SubstName s a):xs) = (SubstName s a):(clean (deleteFn (SubstName s a) xs))
   clean ((SubstVar x y):xs) = (SubstVar x y):(clean (deleteFn (SubstVar x y) xs))
   deleteFn (SubstVar x y) xs = deleteRekBy (\x y ->
       case (x,y) of 
        (SubstVar x1 x2,SubstVar y1 y2) -> x1 == y1
        _ -> False) (SubstVar x y) xs
   deleteFn (SubstName s a) xs = deleteRek (SubstName s a) xs 
   deleteFn (SubstSet s) xs = foldr (\a b -> deleteFn a b) xs s
  
   
mkVar [] []= []
mkVar ((VarLift subst x):xs)  (y:ys) = (VarLift emptySubstitution y):(mkVar xs ys)
mkVar ((MetaVarLift subst x):xs)  (y:ys) = (MetaVarLift emptySubstitution y):(mkVar xs ys)
mkVar ((ConstantVarLift subst x):xs) (y:ys) = (ConstantVarLift emptySubstitution y):(mkVar xs ys)
 
freshtest =[ "##" ++ show i | i <- [1..10]]           

varName (MetaVarLift _ n) = n
varName (VarLift _ n) = n
varName (ConstantVarLift _ n) = n

 
lookupS v sub' = 
 let sub = flatten $ substList sub'
     flatten ((SubstSet set):rest) = flatten (set ++ rest)
     flatten (x:xs) = x:(flatten xs)
     flatten [] = []
 in
  case [v2 | SubstVar v1 v2 <-  sub, varName v1 == v] of 
    [] -> Nothing
    (x:xs) -> Just (varName x)

    

varPrefix (VarLift _ _) = "x"
varPrefix (MetaVarLift _ _) = "X"
varPrefix (ConstantVarLift _ _) = "X"




-- =============================================================================
-- Applying a name-to-name substitution (used for checking alpha-equivalence
-- =============================================================================
type NameMap = Map.Map String String         

changeNamesSyntaxComponent :: NameMap -> SyntaxComponent -> SyntaxComponent
changeNamesSyntaxComponent mp (SCVar v)                       = SCVar (changeNamesVars mp v)
changeNamesSyntaxComponent mp (SCExpr e)                      = SCExpr (changeNamesExpr mp e)
changeNamesSyntaxComponent mp (SCCtxt c)                      = SCCtxt (changeNamesExpr mp c)
changeNamesSyntaxComponent mp (SCMetaEnv (Env sub name))      = SCMetaEnv (Env (changeNamesSubst mp sub) name)
changeNamesSyntaxComponent mp (SCMetaChain sub typ name)      = SCMetaChain (changeNamesSubst mp sub) typ name
changeNamesSyntaxComponent mp (SCConstantEnv (Env sub name))  = SCConstantEnv (Env (changeNamesSubst mp sub) name)
changeNamesSyntaxComponent mp (SCConstantChain sub typ name ) = SCConstantChain (changeNamesSubst mp sub) typ name

changeNamesSubst :: NameMap -> Substitution -> Substitution
changeNamesSubst mp (Substitution sub) = Substitution $ map (changeNamesSubstItem mp) sub

changeNamesSubstItem :: NameMap -> SubstItem -> SubstItem
changeNamesSubstItem mp (SubstName (None expr) name) =
 let name' = lookupOrId name mp
     expr' = changeNamesSyntaxComponent mp expr
 in SubstName (None expr') name'

changeNamesSubstItem mp (SubstName (LetVars expr) name) =
 let name' = lookupOrId name mp
     expr' = changeNamesSyntaxComponent mp expr
 in SubstName (LetVars expr') name'
 
changeNamesSubstItem mp (SubstName (CapVars expr) name) =
 let name' = lookupOrId name mp
     expr' = changeNamesSyntaxComponent mp expr
 in SubstName (CapVars expr') name'
 
changeNamesSubstItem mp (SubstVar (VarLift sub1 name1) (VarLift sub2 name2)) =
 let name2' = lookupOrId name2 mp
     sub1' = changeNamesSubst mp sub1
     sub2' = changeNamesSubst mp sub2  
 in SubstVar (VarLift sub1' name1) (VarLift sub2' name2')

changeNamesSubstItem mp (SubstVar (MetaVarLift sub1 name1) (MetaVarLift sub2 name2)) =
 let name2' = lookupOrId name2 mp
     sub1' = changeNamesSubst mp sub1
     sub2' = changeNamesSubst mp sub2  
 in SubstVar (MetaVarLift sub1' name1) (MetaVarLift sub2' name2')
 
changeNamesSubstItem mp bef@(SubstVar (ConstantVarLift sub1 name1) (ConstantVarLift sub2 name2)) =
 let name2' = lookupOrId name2 mp
     sub1' = changeNamesSubst mp sub1
     sub2' = changeNamesSubst mp sub2  
     after = SubstVar (ConstantVarLift sub1' name1) (ConstantVarLift sub2' name2')
 in  after
 
 
changeNamesSubstItem mp (SubstSet set) = SubstSet $ map (changeNamesSubstItem mp) set

changeNamesVars :: NameMap -> VarLift -> VarLift
changeNamesVars mp (VarLift sub name) =
    let sub' = changeNamesSubst mp sub 
        name' = lookupOrId name mp
    in VarLift sub' name'

changeNamesVars mp (MetaVarLift sub name) =
    let sub' = changeNamesSubst mp sub 
        name' = lookupOrId name mp
    in MetaVarLift sub' name'

changeNamesVars mp (ConstantVarLift sub name) =
    let sub' = changeNamesSubst mp sub 
        name' = lookupOrId name mp
    in ConstantVarLift sub' name'
    
changeNamesExpr mp (VarL var) = VarL (changeNamesVars mp var)

changeNamesExpr mp e@(MetaExpression sub s) =
  MetaExpression (changeNamesSubst mp sub) s
    
changeNamesExpr mp e@(Constant sub s) =
  Constant (changeNamesSubst mp sub) s
 
changeNamesExpr mp e@(MetaCtxtSubst sub typ name s) =
  MetaCtxtSubst (changeNamesSubst mp sub) typ name (changeNamesExpr mp s)
  
changeNamesExpr mp e@(ConstantCtxtSubst sub typ name s) =
  ConstantCtxtSubst (changeNamesSubst mp sub) typ name (changeNamesExpr mp s)

changeNamesExpr mp e@(MetaCtxt sub typ name) =
  MetaCtxt (changeNamesSubst mp sub) typ name
  
changeNamesExpr mp e@(ConstantCtxt sub typ name ) =
  ConstantCtxt (changeNamesSubst mp sub) typ name
   
changeNamesExpr mp (Letrec env s) =
 let 
  s' = changeNamesExpr mp s
  env' = changeNamesEnv mp env
 in Letrec env' s'

  
changeNamesExpr mp (Fn f args) = Fn f (map (changeNamesArgPos mp) args)
changeNamesExpr mp (Hole subs) = Hole (changeNamesSubst mp subs)
changeNamesEnv mp env =
  let
    mc = [MetaChain (changeNamesSubst mp sub) typ name (changeNamesVars mp z) (changeNamesExpr mp s) | (MetaChain sub typ name z s) <- metaChains env]
    cc = [ConstantChain (changeNamesSubst mp sub) typ name (changeNamesVars mp z) (changeNamesExpr mp s) | (ConstantChain sub typ name z s) <- constantChains env]
    me = [Env (changeNamesSubst mp sub) name | (Env sub name) <- metaEnvVars env]
    ce = [Env (changeNamesSubst mp sub) name | (Env sub name) <- constantEnvVars env]    
    b  = [(changeNamesVars mp x) := (changeNamesExpr mp e) | x:=e <- bindings env]
  in env{metaChains = mc, constantChains = cc, metaEnvVars = me, constantEnvVars = ce, bindings = b}

changeNamesArgPos mp (Binder ys s) =
 let
     ys' = map (changeNamesVars mp) ys
     s' = changeNamesExpr mp s
 in Binder ys' s'

 
-- ============================================================================= 
-- Checking simlet and simsubst
-- ============================================================================= 
 
simlet (MetaExpression sub1 n1) (MetaExpression sub2 n2)
    | n1 == n2  =  sub1 `simsubst` sub2
    | otherwise = False
    
simlet (Constant sub1 n1) (Constant sub2 n2)
    | n1 == n2  =  sub1 `simsubst` sub2 
    | otherwise = False
    
simlet (Fn f1 args1) (Fn f2 args2)    
    | f1 == f2  = and (map (\(x,y) -> x `simletArgpos` y) (zip args1 args2))
    | otherwise = False

simlet (MetaCtxtSubst sub1 typ1 name1 s1)  (MetaCtxtSubst sub2 typ2 name2 s2)
    | name1 == name2
    , s1 `simlet` s2
    , typ1 == typ2   = sub1 `simsubst` sub2
    | otherwise = False
    
simlet (ConstantCtxtSubst sub1 typ1 name1 s1)  (ConstantCtxtSubst sub2 typ2 name2 s2)
    | name1 == name2
    , s1 `simlet` s2
    , typ1 == typ2   = sub1 `simsubst` sub2
    | otherwise = False
       
simlet (MetaCtxt sub1 typ1 name1)  (MetaCtxt sub2 typ2 name2)
    | name1 == name2
    , typ1 == typ2   = sub1 `simsubst` sub2
    | otherwise = False
       
simlet (ConstantCtxt sub1 typ1 name1)  (ConstantCtxt sub2 typ2 name2)
    | name1 == name2
    , typ1 == typ2   = sub1 `simsubst` sub2
    | otherwise = False
       
simlet (VarL v1) (VarL v2) = simletVar v1 v2
simlet (Hole sub1) (Hole sub2) = simsubst sub1 sub2

simlet (Letrec env1 s1) (Letrec env2 s2) 
    | s1 `simlet` s2 =
          let me1 = metaEnvVars env1
              me2 = metaEnvVars env2
              meCheck = (length me1 == length me2) &&  
                          
              
                        (let 
                            isSimLetEnvVar xs = and (map (\(x,y) -> x `simLetEnvVar` y) xs)
                            simLetEnvVar (Env sub1 e1) (Env sub2 e2) = e1 == e2 && sub1 `simsubst` sub2
                         in
                          isSimLetEnvVar (zip (sort me1) (sort me2))
                         )
                 
                             
--                              any (isSimLetEnvVar) [zip me1 p | p <- (permutations me2)])
              ce1 = constantEnvVars env1
              ce2 = constantEnvVars env2
              ceCheck = (length ce1 == length ce2) &&  
                        (let 
                            isSimLetEnvVar xs = and (map (\(x,y) -> x `simLetEnvVar` y) xs)
                            simLetEnvVar (Env sub1 e1) (Env sub2 e2) = e1 == e2 && sub1 `simsubst` sub2
                         in isSimLetEnvVar (zip (sort ce1) (sort ce2)))
                             
--                              any (isSimLetEnvVar) [zip ce1 p | p <- (permutations ce2)])                         
              mc1 = metaChains env1
              mc2 = metaChains env2
              mcCheck = (length mc1 == length mc2) &&  
                        (let 
                            isSimChain xs = and (map (\(x,y) -> x `simChain` y) xs)
                            simChain (MetaChain sub1 typ1 name1 z1 s1) (MetaChain sub2 typ2 name2 z2 s2) = 
                                        sub1 `simsubst` sub2 && typ1 == typ2 && name1 == name2 && z1 `simletVar` z2 && s1 `simlet` s2
                         in isSimChain (zip (sort mc1) (sort mc2)))
                         -- any (isSimChain) [zip mc1 p | p <- (permutations mc2)])                         
              cc1 = constantChains env1
              cc2 = constantChains env2
              ccCheck = (length cc1 == length cc2) &&  
                        (let 
                            isSimChain xs = and (map (\(x,y) -> x `simChain` y) xs)
                            simChain (ConstantChain sub1 typ1 name1 z1 s1) (ConstantChain sub2 typ2 name2 z2 s2) = 
                                        sub1 `simsubst` sub2 && typ1 == typ2 && name1 == name2 && z1 `simletVar` z2 && s1 `simlet` s2
                         in isSimChain (zip (sort cc1) (sort cc2))) 
                         -- any (isSimChain) [zip cc1 p | p <- (permutations cc2)])                         
              b1 = bindings env1
              b2 = bindings env2
              bCheck = (length b1 == length b2) &&  
                        (let 
                            isSimBinds xs = and (map (\(x:=s,y:=t) -> (x `simletVar` y && s `simlet` t)) xs)
                         in isSimBinds (zip (sort b1) (sort b2)))
                             -- any (isSimBinds) [zip b1 p | p <- (permutations b2)])                         
                         
                         
          in meCheck && ceCheck && mcCheck && ccCheck && bCheck
simlet (Letrec env1 s1) (Letrec env2 s2) = False
simlet _ _ = False          
               
               
simletArgpos (VarPos v1) (VarPos v2) = simletVar v1 v2
simletArgpos (Binder xs1 s1) (Binder xs2 s2) 
 = 
  (length xs1 == length xs2) 
  &&
  all  (== True) [simletVar x1 x2 | (x1,x2) <- zip xs1 xs2]
  &&
  simlet s1 s2

simletArgpos _ _ = False  


simletVar (VarLift sub1 v1) (VarLift sub2 v2)
    | v1 == v2 = sub1 `simsubst` sub2
simletVar (MetaVarLift sub1 v1) (MetaVarLift sub2 v2)
    | v1 == v2 = sub1 `simsubst` sub2
simletVar (ConstantVarLift sub1 v1) (ConstantVarLift sub2 v2)
    | v1 == v2 = sub1 `simsubst` sub2
simletVar _ _ = False


simsubst (Substitution sub1) (Substitution sub2) =
   let sub1' = toItemList sub1
       sub2' = toItemList sub2
   in similarListsOfListItems sub1' sub2' 
      
    
    

dropAllSubsVars  (VarLift sub name) =
    VarLift emptySubstitution name

dropAllSubsVars  (MetaVarLift sub name) =
    MetaVarLift emptySubstitution name
    
dropAllSubsVars  (ConstantVarLift sub name) =
    ConstantVarLift emptySubstitution name
    
dropAllSubsExpr  (VarL var) = VarL (dropAllSubsVars  var)

dropAllSubsExpr  e@(MetaExpression sub s) =
  MetaExpression emptySubstitution s
    
dropAllSubsExpr  e@(Constant sub s) =
  Constant emptySubstitution s
 
dropAllSubsExpr  e@(MetaCtxtSubst sub typ name s) =
  MetaCtxtSubst emptySubstitution typ name (dropAllSubsExpr  s)
  
dropAllSubsExpr  e@(ConstantCtxtSubst sub typ name s) =
  ConstantCtxtSubst emptySubstitution typ name (dropAllSubsExpr  s)

dropAllSubsExpr  e@(MetaCtxt sub typ name) =
  MetaCtxt emptySubstitution typ name
  
dropAllSubsExpr  e@(ConstantCtxt sub typ name ) =
  ConstantCtxt emptySubstitution typ name
   
dropAllSubsExpr  (Letrec env s) =
 let 
  s' = dropAllSubsExpr  s
  env' = dropAllSubsEnv  env
 in Letrec env' s'

  
dropAllSubsExpr  (Fn f args) = Fn f (map (dropAllSubsArgPos ) args)
dropAllSubsExpr  (Hole subs) = Hole emptySubstitution
dropAllSubsEnv  env =
  let
    mc = [MetaChain emptySubstitution typ name (dropAllSubsVars  z) (dropAllSubsExpr  s) | (MetaChain sub typ name z s) <- metaChains env]
    cc = [ConstantChain emptySubstitution typ name (dropAllSubsVars  z) (dropAllSubsExpr  s) | (ConstantChain sub typ name z s) <- constantChains env]
    me = [Env emptySubstitution name | (Env sub name) <- metaEnvVars env]
    ce = [Env emptySubstitution name | (Env sub name) <- constantEnvVars env]    
    b  = [(dropAllSubsVars  x) := (dropAllSubsExpr  e) | x:=e <- bindings env]
  in env{metaChains = mc, constantChains = cc, metaEnvVars = me, constantEnvVars = ce, bindings = b}

dropAllSubsArgPos  (Binder ys s) =
 let
     ys' = map (dropAllSubsVars ) ys
     s' = dropAllSubsExpr  s
 in Binder ys' s'    
 
 
 
 

similar (MetaExpression sub1 n1) (MetaExpression sub2 n2) = n1 == n2
similar (Constant sub1 n1) (Constant sub2 n2) = n1 == n2
similar (Fn f1 args1) (Fn f2 args2)    
    | f1 == f2  = and (map (\(x,y) -> x `similarArgpos` y) (zip args1 args2))
    | otherwise = False

similar (MetaCtxtSubst sub1 typ1 name1 s1)  (MetaCtxtSubst sub2 typ2 name2 s2)
    | name1 == name2
    , s1 `similar` s2
    , typ1 == typ2   = True
    | otherwise = False
    
similar (ConstantCtxtSubst sub1 typ1 name1 s1)  (ConstantCtxtSubst sub2 typ2 name2 s2)
    | name1 == name2
    , s1 `similar` s2
    , typ1 == typ2   = True 
    | otherwise = False
       
similar (MetaCtxt sub1 typ1 name1)  (MetaCtxt sub2 typ2 name2)
    | name1 == name2
    , typ1 == typ2   = True
    | otherwise = False
       
similar (ConstantCtxt sub1 typ1 name1)  (ConstantCtxt sub2 typ2 name2)
    | name1 == name2
    , typ1 == typ2   = True 
    | otherwise = False
       
similar (VarL v1) (VarL v2) = similarVar v1 v2
similar (Hole sub1) (Hole sub2) = True 

similar (Letrec env1 s1) (Letrec env2 s2) 
    | s1 `similar` s2 =
          let me1 = metaEnvVars env1
              me2 = metaEnvVars env2
              meCheck = (length me1 == length me2) &&  
                        (let 
                            issimilarEnvVar xs = and (map (\(x,y) -> x `similarEnvVar` y) xs)
                            similarEnvVar (Env sub1 e1) (Env sub2 e2) = e1 == e2 
                            sfn (Env _ e1) (Env _ e2) = compare e1 e2                            
                         in issimilarEnvVar (zip (sortBy sfn me1) (sortBy sfn me2)))                         

              ce1 = constantEnvVars env1
              ce2 = constantEnvVars env2
              ceCheck = (length ce1 == length ce2) &&  
                        (let 
                            issimilarEnvVar xs = and (map (\(x,y) -> x `similarEnvVar` y) xs)
                            similarEnvVar (Env sub1 e1) (Env sub2 e2) = e1 == e2 
                            sfn (Env _ e1) (Env _ e2) = compare e1 e2
                         in issimilarEnvVar (zip (sortBy sfn ce1) (sortBy sfn ce2)))    
              mc1 = metaChains env1
              mc2 = metaChains env2
              mcCheck = (length mc1 == length mc2) &&  
                        (let 
                            isSimChain xs = and (map (\(x,y) -> x `simChain` y) xs)
                            simChain (MetaChain sub1 typ1 name1 z1 s1) (MetaChain sub2 typ2 name2 z2 s2) = 
                                        typ1 == typ2 && name1 == name2 && z1 `similarVar` z2 && s1 `similar` s2
                            sfn (MetaChain sub1 typ1 name1 z1 s1)
                                (MetaChain sub2 typ2 name2 z2 s2) =
                                    case compare name1 name2 of
                                         EQ -> case compare s1 s2 of
                                                    EQ -> compare z1 z2
                                                    other -> other
                                         other -> other
                                        
                         in isSimChain (zip (sortBy sfn mc1) (sortBy sfn mc2))    )
              cc1 = constantChains env1
              cc2 = constantChains env2
              ccCheck = (length cc1 == length cc2) &&  
                        (let 
                            isSimChain xs = and (map (\(x,y) -> x `simChain` y) xs)
                            simChain (ConstantChain sub1 typ1 name1 z1 s1) (ConstantChain sub2 typ2 name2 z2 s2) = 
                                       typ1 == typ2 && name1 == name2 && z1 `similarVar` z2 && s1 `similar` s2
                            sfn (ConstantChain sub1 typ1 name1 z1 s1)
                                (ConstantChain sub2 typ2 name2 z2 s2) =
                                    case compare name1 name2 of
                                         EQ -> case compare s1 s2 of
                                                    EQ -> compare z1 z2
                                                    other -> other
                                         other -> other
                                         
                         in isSimChain (zip (sortBy sfn cc1) (sortBy sfn cc2))    )                   
              b1 = bindings env1
              b2 = bindings env2
              bCheck = (length b1 == length b2) &&  
                        (let 
                            isSimChain xs = and (map (\(x:=s,y:=t) -> (x `similarVar` y && s `similar` t)) xs)
                            sfn (x:=u1) (y:=u2) = case compare u1 u2 of
                                                   EQ -> compare x y
                                                   other -> other
                         in isSimChain (zip (sortBy sfn b1) (sortBy sfn b2)))    
                         
                         
          in 
              mytrace (show $ (meCheck , ceCheck , mcCheck , ccCheck , bCheck, (sort b1, sort b2)))  
              (meCheck && ceCheck && mcCheck && ccCheck && bCheck)
              
similar (Letrec env1 s1) (Letrec env2 s2) = False
similar _ _ = False          
               
               
similarArgpos (VarPos v1) (VarPos v2) = similarVar v1 v2
similarArgpos (Binder xs1 s1) (Binder xs2 s2) 
 = 
  (length xs1 == length xs2) 
  &&
  all  (== True) [similarVar x1 x2 | (x1,x2) <- zip xs1 xs2]
  &&
  similar s1 s2

similarArgpos _ _ = False  


similarVar (VarLift sub1 v1) (VarLift sub2 v2) = True
similarVar (MetaVarLift sub1 v1) (MetaVarLift sub2 v2) = True
similarVar (ConstantVarLift sub1 v1) (ConstantVarLift sub2 v2) = True
similarVar _ _ = False



-- ============================================
-- more functions for full renamings
applySingleFullRenamingAlphaBase ren alphabase = 
 [ (sc',item')
  | (sc,item) <- alphabase
  , sc' <- applySingleFullRenamingSC ren sc
  , item' <- applySingleFullRenamingSubstItem ren item
  ] 


applySingleFullRenamingNCC :: SingleFullRenaming -> NonCaptureConstraints -> NonCaptureConstraints
applySingleFullRenamingNCC ren nccs =
 let 
  res = 
   [(s',d')
    | (s,d) <- nccs
      ,s' <- applySingleFullRenamingEither ren s
      ,d' <- applySingleFullRenaming ren d
   ]
  in 
  {-
   trace 
     (   "\n DEBUG: apply full renamaing:\n"
      ++ "\n DEBUG:   renaming: " ++ show ren 
      ++ "\n DEBUG:   NCCs before: " ++ show nccs
      ++ "\n Debug:   NCCs after: " ++ show res
     ) 
   -}  
     res

applySingleFullRenamingEither ren (Left e) = 
 [Left e' | e' <- applySingleFullRenaming ren e]
applySingleFullRenamingEither ren (Right env) = 
 [Right env' | env' <- applySingleFullRenamingEnv ren env]

applySingleFullRenaming :: SingleFullRenaming -> Expression -> [Expression]
applySingleFullRenaming ren (MetaExpression sub s) =
  [MetaExpression sub' s | sub' <- (applySingleFullRenamingSubst ren sub)]   
applySingleFullRenaming ren (Constant sub s) =
  [Constant sub' s | sub' <- (applySingleFullRenamingSubst ren sub)]   
applySingleFullRenaming ren (MetaCtxt sub srt name) =
  [MetaCtxt sub' srt name | sub' <- (applySingleFullRenamingSubst ren sub)]   
applySingleFullRenaming ren (ConstantCtxt sub srt name) =
  [ConstantCtxt sub' srt name | sub' <- (applySingleFullRenamingSubst ren sub)]   
applySingleFullRenaming ren (MetaCtxtSubst sub srt name expr) =
  [MetaCtxtSubst sub' srt name expr'
   | sub' <- (applySingleFullRenamingSubst ren sub)
   , expr' <- (applySingleFullRenaming ren expr)
   ]
applySingleFullRenaming ren (ConstantCtxtSubst sub srt name expr) =
  [ConstantCtxtSubst sub' srt name expr'
   | sub' <- (applySingleFullRenamingSubst ren sub)
   , expr' <- (applySingleFullRenaming ren expr)
   ]
applySingleFullRenaming ren (VarL v) =
   [VarL v' | v' <- (applySingleFullRenamingVar ren v)]
   
applySingleFullRenaming ren (Letrec env expr) =
   [Letrec env' expr'
     | env' <- (applySingleFullRenamingEnv ren env)
     , expr' <- (applySingleFullRenaming ren expr)
     ]
     
applySingleFullRenaming ren (Hole sub) =
   [ Hole sub' | sub' <- (applySingleFullRenamingSubst ren sub)]
   
applySingleFullRenaming ren (Fn f args) =
  let args' = map (applySingleFullRenamingArgPos ren) args
      combs []       = [[]]
      combs (xs:xss) = [x:ys | x <- xs, ys <- combs xss ]
      argsCombs = combs args'
  in [Fn f newargs | newargs <- argsCombs]

applySingleFullRenamingArgPos ren (VarPos v) =
 [VarPos v' | v' <- applySingleFullRenamingVar ren v]
 
applySingleFullRenamingArgPos ren (Binder vs expr) =
 let vs' = map (applySingleFullRenamingVar ren) vs 
     combs []       = [[]]
     combs (xs:xss) = [x:ys | x <- xs, ys <- combs xss ]
     vsCombs = combs vs'
 in [Binder us expr' 
    |
    us <- vsCombs,
    expr' <- applySingleFullRenaming ren expr
    ]
    
applySingleFullRenamingEnv ren env =
    let 
     combs []       = [[]]
     combs (xs:xss) = [x:ys | x <- xs, ys <- combs xss ]
     applySingleFullRenamingBinding (x := e) = 
          [x' := e'
          | x' <- applySingleFullRenamingVar ren x  
          , e' <- applySingleFullRenaming ren e
          ]
     bindings' = combs (map applySingleFullRenamingBinding (bindings env))
     applySingleFullRenamingMetaChain (MetaChain sub typ name z s) = 
          [MetaChain sub' typ name z' s'
          | sub' <- applySingleFullRenamingSubst ren sub  
          , z' <- applySingleFullRenamingVar ren z
          , s' <- applySingleFullRenaming ren s
          ]
     metachains' = combs (map applySingleFullRenamingMetaChain (metaChains env))
     applySingleFullRenamingConstantChain (ConstantChain sub typ name z s) = 
          [ConstantChain sub' typ name z' s'
          | sub' <- applySingleFullRenamingSubst ren sub
          , z' <- applySingleFullRenamingVar ren z
          , s' <- applySingleFullRenaming ren s
          ]
     constantchains' = combs (map applySingleFullRenamingConstantChain (constantChains env))
     applySingleFullRenamingEnv (Env sub name) = 
          [Env sub' name
          | sub' <- applySingleFullRenamingSubst ren sub  
          ]
     metaenvs' = combs (map applySingleFullRenamingEnv (metaEnvVars env))
     constantenvs' = combs (map applySingleFullRenamingEnv (constantEnvVars env))
    in 
     [  empty{bindings = bs
             ,metaChains = ms
             ,constantChains = cs 
             ,metaEnvVars = me 
             ,constantEnvVars = ce
             }
        | bs <- bindings'
        , ms <- metachains'
        , cs <- constantchains'
        , me <- metaenvs'
        , ce <- constantenvs'
     ]
    
applySingleFullRenamingSubst :: SingleFullRenaming -> Substitution -> [Substitution]    
applySingleFullRenamingSubst ren sub =
  let itemlist = substList sub 
  in [sub{substList=sub'} | sub' <- applySingleFullRenamingSubstList ren itemlist]
  
applySingleFullRenamingSubstList :: SingleFullRenaming -> [SubstItem] -> [[SubstItem]]
applySingleFullRenamingSubstList ren list =
    let 
     combs []       = [[]]
     combs (xs:xss) = [x:ys | x <- xs, ys <- combs xss ]
    in combs (map (applySingleFullRenamingSubstItem ren) list)

applySingleFullRenamingSubstItem :: SingleFullRenaming -> SubstItem -> [SubstItem]
applySingleFullRenamingSubstItem ren (SubstName (None comp) name)
 | Just names <- Map.lookup name ren
 = [SubstName (None comp') name' |
      name' <- names
     ,comp' <- applySingleFullRenamingSC ren comp
   ]
 | otherwise 
 = [SubstName (None comp') name |
     comp' <- applySingleFullRenamingSC ren comp
   ]
applySingleFullRenamingSubstItem ren (SubstName (CapVars comp) name)
 | Just names <- Map.lookup name ren
 = [SubstName (CapVars comp') name' |
      name' <- names
     ,comp' <- applySingleFullRenamingSC ren comp
   ]
 | otherwise 
 = [SubstName (CapVars comp') name |
     comp' <- applySingleFullRenamingSC ren comp
   ]
applySingleFullRenamingSubstItem ren (SubstName (LetVars comp) name)
 | Just names <- Map.lookup name ren
 = [SubstName (LetVars comp') name' |
      name' <- names
     ,comp' <- applySingleFullRenamingSC ren comp
   ]
 | otherwise 
 = [SubstName (LetVars comp') name |
     comp' <- applySingleFullRenamingSC ren comp
   ]
   
applySingleFullRenamingSubstItem ren (SubstVar v1 v2)
 = [SubstVar v1' v2' |
      v1' <- applySingleFullRenamingVar ren v1
     ,v2' <- applySingleFullRenamingVar ren v2
    ]
applySingleFullRenamingSubstItem ren (SubstSet list)
 = [SubstSet l | l <- applySingleFullRenamingSubstList ren list]

applySingleFullRenamingSC ren (SCVar v)
 = [SCVar v' | v' <- applySingleFullRenamingVar ren v]
applySingleFullRenamingSC ren (SCExpr e)
 = [SCExpr e' | e' <- applySingleFullRenaming ren e]
applySingleFullRenamingSC ren (SCCtxt e)
 = [SCCtxt e' | e' <- applySingleFullRenaming ren e]
applySingleFullRenamingSC ren (SCMetaEnv (Env sub e))
 = [SCMetaEnv (Env sub' e) 
   | sub' <- applySingleFullRenamingSubst ren sub 
   ]
applySingleFullRenamingSC ren (SCConstantEnv (Env sub e))
 = [SCConstantEnv (Env sub' e) 
   | sub' <- applySingleFullRenamingSubst ren sub 
   ]
applySingleFullRenamingSC ren (SCMetaChain sub typ name)
 = [SCMetaChain sub' typ name
   | sub' <- applySingleFullRenamingSubst ren sub 
   ]
applySingleFullRenamingSC ren (SCConstantChain sub typ name)
 = [SCConstantChain sub' typ name
   | sub' <- applySingleFullRenamingSubst ren sub 
   ]
   

applySingleFullRenamingVar ren (VarLift sub name)
 | Just  newnames <- Map.lookup name ren =  [(VarLift sub n) | n <- newnames]
 | otherwise  = [VarLift sub name]
applySingleFullRenamingVar ren (MetaVarLift sub name)
 | Just  newnames <- Map.lookup name ren =  [(MetaVarLift sub n) | n <- newnames]
 | otherwise  = [MetaVarLift sub name]
applySingleFullRenamingVar ren (ConstantVarLift sub name)
 | Just  newnames <- Map.lookup name ren =  [(ConstantVarLift sub n) | n <- newnames]
 | otherwise  = [ConstantVarLift sub name]
   




-- ===================================================================================
-- alphaRepair should repair an alpha renaming after some transformations
-- i.e. we have expression s which is already alpha-renamed s.t. s fulfills the strong DVC
-- and then s is transformed into s'.
-- There are cases were s' does no longer fulfill the strong DVC, since e.g. an abstraction
-- has been copied and there now two syntactically equal subexpressions. 
-- Instead of alpha-renaming s' again, we try to repair the alpha renaming
-- The trick is that we have the freshness constraints for s and can extend them
-- by new names for s'. A view is that if X |-> X' is a fresh renaming of X, then
-- we can also see this renaming as a renaming  X |-> vect(X') where vect(X') is a set of
-- fresh names. Analogously this works for meta-alpha-renamings.
-- If there are now two instances of X' at binders in s', then we choose two different instances
-- of vect(X') for the repairing the renaming of s'. However, the scope of X' has to be respected
-- 
-- We consider some examples:
--
-- C[\X.S,Y] is renamed into C[\X'.SalphaS[X |-> X'],Y] and then 
-- the transformation results in C[\X'.SalphaS[X |-> X'],\X'.SalphaS[X |-> X']]
-- repairing leads to C[\X'.SalphaS[X |-> X'],\X''.Salpha1S[X |-> X'']]#
--
-- C[\X.S,Y] is renamed into C[\X'.SalphaS[X |-> X'],Y] and then
-- the transformation results in C[\X'.SalphaS[X |-> X'],SalphaS]
-- repairing leads to C[\X'.SalphaS[X |-> X'],\X''.Salpha1S]

-- C[\X.S,.,.,] is renamed into C[\X'.SalphaS[X |-> X'],.,.] and then
-- the transformation results in C[\X'.SalphaS[X |-> X'],\X'.SalphaS[X |-> X'],X']
-- In this case repairing is impossible, since the free occurence of X' in the third hole
-- cannot be renamed

-- let Y=\X.Y ... is renamed into  let Y'=\X'.Y'[X|->X', Y|->Y'] and then
-- the transformation results in let Y'=\X'.(\X'.Y'[X|->X', Y|->Y'])[X|->X', Y|->Y']
-- what to do in this case? 
-- let Y'=\X'.(\X1.Y'[X|->X1, Y|->Y'])[X|->X', Y|->Y']
-- seems to be right?
--
-- This can also be a little bit more complex:
-- let Y=\X.C[Y] ---> let Y'=\X'.CalphaC[X|->X',Y|->Y'][Y']
--
-- 
--
--
-- Thus we have to check when re-renaming a binder \X' into \X'' 
-- whether there are other occurences of X' in the whole expression
-- and whether these occurrences are dangerous or not

-- 
--
--
--
-- Again the algorithm works inside out

-- DS: July 16: TODO mangle the cdelta constraints of the problem
--     to apply the new renamings to the old constraints, provided
--     there are subterms like S_alpha_1_S   which now would become
--         S_alpha_2_S, since it may happen that rules like (case)
--     add constraints! (see comments in ConstraintInclusion also)
-- is this done?

startAlphaRepair alphabase fresh e expr = 
    let (fresh',alphabase',expr') = alphaRepair alphabase fresh e expr 
        names = getUsedRenamingNamesExpr expr'
        alphabase'' = simplify names alphabase'
    in (fresh',alphabase'',expr')


simplify _ [] = []    
simplify  occuring ((sc,item):rest) 
    | Set.null ((Set.intersection) (proc item) occuring)
       = simplify occuring rest
    | otherwise = (sc,item):(simplify occuring rest) 
    
proc (SubstVar v1 v2) = getUsedRenamingNamesVar v2
proc (SubstName (None comp) n) =  (Set.fromList [n])
proc (SubstName (LetVars comp) n) =  (Set.fromList [n])
proc (SubstName (CapVars comp) n) =  (Set.fromList [n])



 
alphaRepair alphabase fresh expr (VarL v) =  (fresh,alphabase,VarL v)
    
 

alphaRepair alphabase fresh expr (MetaExpression (Substitution (item1:rest)) s)
    | Just (SCExpr s') <- lookup item1 [(b,a) | (a,b) <- alphabase]
    , (MetaExpression (Substitution []) s) == s'
    , (f:fresh') <- fresh
    = let newAlpha = "ALPHA" ++ f
          (SubstName _ oldAlpha) = item1
          newSubstItem = (SubstName (None (SCExpr s') ) newAlpha)
          rest' = replaceInSubstListBy (SCExpr s')  newAlpha oldAlpha rest 
          newAlphaBase = alphabase ++ [(SCExpr s',newSubstItem)]
      in  
           (fresh', newAlphaBase, MetaExpression (Substitution (newSubstItem:rest')) s)

alphaRepair alphabase fresh expr e@(MetaExpression (Substitution sub) s) =
    -- sub does not contain a fresh substitution of s, or sub is empty
    (fresh,alphabase,e)



alphaRepair alphabase fresh expr (Constant (Substitution (item1:rest)) s)
    | Just (SCExpr s') <- lookup item1 [(b,a) | (a,b) <- alphabase]
    , (Constant (Substitution []) s) == s'
    , (f:fresh') <- fresh
    = let newAlpha = "ALPHA" ++ f
          (SubstName _ oldAlpha) = item1
          newSubstItem = (SubstName (None (SCExpr s') ) newAlpha)
          rest' = replaceInSubstListBy (SCExpr s')  newAlpha oldAlpha rest 
          newAlphaBase = alphabase ++ [(SCExpr s',newSubstItem)]         
      in  (fresh', newAlphaBase, Constant (Substitution (newSubstItem:rest')) s)
      
alphaRepair alphabase fresh expr e@(Constant (Substitution sub) s) =
    -- sub does not contain a fresh substitution of s, or sub is empty
    (fresh,alphabase,e)
      
alphaRepair alphabase fresh expr (MetaCtxtSubst (Substitution (item1:rest)) srt name e)
    | Just (SCCtxt s') <- lookup item1 [(b,a) | (a,b) <- alphabase]
    , (MetaCtxt (Substitution []) srt name) == s'
    , (f:fresh') <- fresh
    = let (fresh'',a'',e') = alphaRepair alphabase fresh' expr e
          newAlpha = "ALPHA" ++ f
          (SubstName _ oldAlpha) = item1
          newSubstItem = (SubstName (None (SCCtxt s') ) newAlpha)
          rest' = replaceInSubstListBy (SCCtxt s')  newAlpha oldAlpha rest 
          newAlphaBase = a'' ++ [(SCCtxt s',newSubstItem)]
          newE = replaceInExprBy (SCCtxt s') newAlpha oldAlpha e'
      in  
           (fresh'', newAlphaBase, MetaCtxtSubst (Substitution (newSubstItem:rest')) srt name newE)
           
alphaRepair alphabase fresh expr (MetaCtxtSubst sub srt name e) =
    let (f',a',e') = alphaRepair alphabase fresh expr e
    in (f',a',MetaCtxtSubst sub srt name e')


alphaRepair alphabase fresh expr (ConstantCtxtSubst (Substitution (item1:rest)) srt name e)
    | Just (SCCtxt s') <- lookup item1 [(b,a) | (a,b) <- alphabase]
    , (ConstantCtxt (Substitution []) srt name) == s'
    , (f:fresh') <- fresh
    = let (fresh'',a'',e') = alphaRepair alphabase fresh' expr e
          newAlpha = "ALPHA" ++ f
          (SubstName _ oldAlpha) = item1
          newSubstItem = (SubstName (None (SCCtxt s') ) newAlpha)
          rest' = replaceInSubstListBy (SCCtxt s')  newAlpha oldAlpha rest 
          newAlphaBase = a'' ++ [(SCCtxt s',newSubstItem)]
          newE = replaceInExprBy (SCCtxt s') newAlpha oldAlpha e'
      in  
           (fresh'', newAlphaBase, ConstantCtxtSubst (Substitution (newSubstItem:rest')) srt name newE)
           
alphaRepair alphabase fresh expr (ConstantCtxtSubst sub srt name e) =
    let (f',a',e') = alphaRepair alphabase fresh expr e
    in (f',a',ConstantCtxtSubst sub srt name e')

           

    
alphaRepair alphabase fresh expr (Letrec env s) =
 let env' = env -- tbd
     (freshAfterS,a',s') = (alphaRepair alphabase fresh expr s)
     list = [(y,x) | (SCVar x, SubstVar x' y) <- a']
     -- ----------------------------------------------------
     -- process the bindings
     startRepairBindings1 =                       
          [ case lookup oldto list of
               Nothing ->  (Left oldto,s)                               -- do not rename
               (Just oldfrom) -> (Right (oldto,oldfrom,newto),s)    -- rename
           | (newto, oldto := s) <- zip freshAfterS (bindings env)
          ]
     freshAfterS2 = drop (length startRepairBindings1) freshAfterS
     (freshAfterBindings,a'',newRHS) = 
         let rhs= [e | (x:=e) <- bindings env]
             go f a (e:es) = 
                 let (fNeu,aNeu,eNeu) = alphaRepair a f expr e
                     (fNeu2,aNeu2,esNeu) = go fNeu aNeu es
                 in (fNeu2,aNeu2,(eNeu:esNeu))
             go f a [] = (f,a,[])
         in 
             go freshAfterS2 a' rhs
     startRepairBindings =                       
         zipWith (\x e -> case x of 
                           (Left z,_) -> (Left z,e)
                           (Right t,_) -> (Right t,e))
                 startRepairBindings1
                 newRHS
     
     triplesBVars = 
              [ case oldfrom of 
                   VarLift substold a -> (oldfrom,oldto,VarLift substold newto)
                   MetaVarLift substold a -> (oldfrom,oldto,MetaVarLift substold newto)
                   ConstantVarLift substold a -> (oldfrom,oldto,ConstantVarLift substold newto)
                | (Right (oldto,oldfrom,newto),_) <- startRepairBindings
              ]
                                                       
     varSubst = map (\(old,new) -> case old of 
                                    VarLift substold a -> (old,VarLift substold new)
                                    MetaVarLift substold a -> (old,MetaVarLift substold new)
                                    ConstantVarLift substold a -> (old,ConstantVarLift substold new)) 
                    [(oldfrom,newto) | (Right (oldto,oldfrom,newto),s) <- startRepairBindings] 
                    
     newVars = map snd varSubst
     alphaBaseAfterBindings = a'' ++ 
                         [ (SCVar oldfrom, SubstVar oldfrom 
                             (case oldfrom of 
                                VarLift substold a -> (VarLift substold newto)
                                MetaVarLift substold a -> (MetaVarLift substold newto)
                                ConstantVarLift substold a -> (ConstantVarLift substold newto)
                             )
                            )
                                | (Right (oldto,oldfrom,newto),s) <- startRepairBindings]
     envAfterBindings = foldr 
                            (\(oldfrom,oldto,newto) env -> replaceTriplesEnv oldfrom oldto newto env) 
                            env{bindings=[
                                  case x of 
                                        (Left  z,s) -> z := s
                                        (Right (oldto,oldfrom,newto),s) -> 
                                            (case oldfrom of 
                                                VarLift substold a -> (VarLift substold newto)
                                                MetaVarLift substold a -> (MetaVarLift substold newto)
                                                ConstantVarLift substold a -> (ConstantVarLift substold newto)
                                            ) := s  | x <- startRepairBindings]
                                 } 
                            triplesBVars
     sAfterBindings    =  foldr (\(oldfrom,oldto,newto) e -> replaceTriples oldfrom oldto newto e) s' triplesBVars
     -- ----------------------------------------------------
     -- process the metaChains
     startRepairMetaChains1 =                       
          [ case lookup oldto list of
               Nothing ->  (Left oldto,sub,typ,name,s)                               -- do not rename
               (Just oldfrom) -> (Right (oldto,oldfrom,newto),sub,typ,name,s)    -- rename
           | (newto, MetaChain sub typ name oldto s) <- zip freshAfterBindings (metaChains envAfterBindings)
          ]
     freshAfterMetaChains0 = drop (length startRepairMetaChains1) freshAfterBindings
     (freshAfterMetaChains,a''',newRHSMetaChains) = 
         let rhs= [s | MetaChain sub typ name oldto s <- metaChains envAfterBindings]
             go f a (e:es) = 
                 let (f',a',e') = alphaRepair a f expr e
                     (f'',a'',es') = go f' a' es
                 in (f'',a'',(e':es'))
             go f a [] = (f,a,[])
         in 
             go freshAfterMetaChains0 alphaBaseAfterBindings rhs
     startRepairMetaChains =                       
         zipWith (\x e -> case x of 
                           (Left z,sub,typ,name,s) -> (Left z,sub,typ,name,e)
                           (Right t,sub,typ,name,s) -> (Right t,sub,typ,name,e))
                 startRepairMetaChains1
                 newRHSMetaChains
     
     triplesMCVars = 
              [ case oldfrom of 
                   VarLift substold a -> (oldfrom,oldto,VarLift substold newto)
                   MetaVarLift substold a -> (oldfrom,oldto,MetaVarLift substold newto)
                   ConstantVarLift substold a -> (oldfrom,oldto,ConstantVarLift substold newto)
                | (Right (oldto,oldfrom,newto),_,_,_,_) <- startRepairMetaChains
              ]
     varSubstMC = map
                    (\(old,new) -> case old of 
                                    VarLift substold a -> (old,VarLift substold new)
                                    MetaVarLift substold a -> (old,MetaVarLift substold new)
                                    ConstantVarLift substold a -> (old,ConstantVarLift substold new))
                    [(oldfrom,newto) | (Right (oldto,oldfrom,newto),_,_,_,_) <- startRepairMetaChains]
     alphaBaseAfterMetaChains = a''' ++
                         [ (SCVar oldfrom, SubstVar oldfrom 
                             (case oldfrom of 
                                VarLift substold a -> (VarLift substold newto)
                                MetaVarLift substold a -> (MetaVarLift substold newto)
                                ConstantVarLift substold a -> (ConstantVarLift substold newto)
                             )
                            )
                                | (Right (oldto,oldfrom,newto),_,_,_,_) <- startRepairMetaChains]
     envAfterMetaChains = foldr 
                            (\(oldfrom,oldto,newto) env -> replaceTriplesEnv oldfrom oldto newto env) 
                            envAfterBindings{metaChains=[
                                  case x of 
                                        (Left  z,sub,typ,name,s) -> MetaChain sub typ name z s
                                        (Right (oldto,oldfrom,newto),sub,typ,name,s) -> 
                                          MetaChain sub typ name 
                                            (case oldfrom of 
                                                VarLift substold a -> (VarLift substold newto)
                                                MetaVarLift substold a -> (MetaVarLift substold newto)
                                                ConstantVarLift substold a -> (ConstantVarLift substold newto)
                                            ) s  | x <- startRepairMetaChains]
                                 } 
                            triplesMCVars
                            
     sAfterMetaChains    =  foldr (\(oldfrom,oldto,newto) e -> replaceTriples oldfrom oldto newto e) sAfterBindings triplesMCVars
     -- repair alphas of metachains
     metaChains2 =
        [
        case sub of
           [] -> Left (sub,typ,name,z,s)
           (item:rest) -> 
              case lookup item [(b,a) | (a,b) <- alphaBaseAfterMetaChains] of
                Just m@(SCMetaChain sub0 typ0 name0) ->
                   if (SCMetaChain (Substitution []) typ name) == m then
                        let newAlpha = "ALPHA" ++ f
                            (SubstName _ oldAlpha) = item
                            newSubstItem = (SubstName (None m) newAlpha)
                        in Right (oldAlpha,newAlpha,newSubstItem,(newSubstItem):rest,typ,name,z,s)
                   else
                      Left (sub,typ,name,z,s)
       |          
        (MetaChain (Substitution sub) typ name z s,f) <-  zip (metaChains envAfterMetaChains) freshAfterMetaChains
       ]
     freshAfterMetaChains2 = drop (length metaChains2) freshAfterMetaChains
     alphaBaseAfterMetaChains2 = alphaBaseAfterMetaChains ++
        [ (SCMetaChain (Substitution []) typ name, newSubstItem)
            | Right (oldAlpha,newAlpha,newSubstItem,_,typ,name,z,s) <- metaChains2
        ]
     envAfterMetaChains2 = foldr 
                            (\(sc,oldAlpha,newAlpha) env -> replaceInEnvBy sc newAlpha oldAlpha env)
                            envAfterMetaChains{
                                metaChains = [  case x of 
                                                Right (oldAlpha,newAlpha,newSubstItem,sub,typ,name,z,s)
                                                    -> MetaChain (Substitution sub) typ name z s
                                                Left (sub,typ,name,z,s) 
                                                    -> MetaChain (Substitution sub) typ name z s
                                            | x <- metaChains2
                                            ]}
                            
                           [ (SCMetaChain (Substitution []) typ name,oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,sub,typ,name,z,s) <- metaChains2]
     sAfterMetaChains2 =  foldr 
                            (\(sc,oldAlpha,newAlpha) e -> replaceInExprBy sc newAlpha oldAlpha e)
                            sAfterMetaChains
                            [ (SCMetaChain (Substitution []) typ name,oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,_,typ,name,z,s) <- metaChains2]  
     -- ----------------------------------------------------
     -- process the constantChains
     startRepairConstantChains1 =                       
          [ case lookup oldto list of
               Nothing ->  (Left oldto,sub,typ,name,s)                               -- do not rename
               (Just oldfrom) -> (Right (oldto,oldfrom,newto),sub,typ,name,s)    -- rename
           | (newto, ConstantChain sub typ name oldto s) <- zip freshAfterMetaChains2 (constantChains envAfterMetaChains2)
          ]
     freshAfterConstaintChains0 = drop (length startRepairConstantChains1) freshAfterMetaChains2
     (freshAfterConstantChains,a4,newRHSConstantChains) = 
         let rhs= [s | ConstantChain sub typ name oldto s <- constantChains envAfterMetaChains2]
             go f a (e:es) = 
                 let (f',a',e') = alphaRepair a f expr e
                     (f'',a'',es') = go f' a' es
                 in (f'',a'',(e':es'))
             go f a [] = (f,a,[])
         in 
             go freshAfterConstaintChains0 alphaBaseAfterMetaChains2 rhs
     startRepairConstantChains =                       
         zipWith (\x e -> case x of 
                           (Left z,sub,typ,name,s) -> (Left z,sub,typ,name,e)
                           (Right t,sub,typ,name,s) -> (Right t,sub,typ,name,e))
                 startRepairConstantChains1
                 newRHSConstantChains
     
     triplesCCVars = 
              [ case oldfrom of 
                   VarLift substold a -> (oldfrom,oldto,VarLift substold newto)
                   MetaVarLift substold a -> (oldfrom,oldto,MetaVarLift substold newto)
                   ConstantVarLift substold a -> (oldfrom,oldto,ConstantVarLift substold newto)
                | (Right (oldto,oldfrom,newto),_,_,_,_) <- startRepairConstantChains
              ]
     varSubstCC = map
                    (\(old,new) -> case old of 
                                    VarLift substold a -> (old,VarLift substold new)
                                    MetaVarLift substold a -> (old,MetaVarLift substold new)
                                    ConstantVarLift substold a -> (old,ConstantVarLift substold new))
                    [(oldfrom,newto) | (Right (oldto,oldfrom,newto),_,_,_,_) <- startRepairConstantChains]
     alphaBaseAfterConstantChains = a4 ++
                         [ (SCVar oldfrom, SubstVar oldfrom 
                             (case oldfrom of 
                                VarLift substold a -> (VarLift substold newto)
                                MetaVarLift substold a -> (MetaVarLift substold newto)
                                ConstantVarLift substold a -> (ConstantVarLift substold newto)
                             )
                            )
                                | (Right (oldto,oldfrom,newto),_,_,_,_) <- startRepairConstantChains]
     envAfterConstantChains = foldr 
                            (\(oldfrom,oldto,newto) env -> replaceTriplesEnv oldfrom oldto newto env) 
                            envAfterMetaChains2{constantChains=[
                                  case x of 
                                        (Left  z,sub,typ,name,s) -> ConstantChain sub typ name z s
                                        (Right (oldto,oldfrom,newto),sub,typ,name,s) -> 
                                          ConstantChain sub typ name 
                                            (case oldfrom of 
                                                VarLift substold a -> (VarLift substold newto)
                                                MetaVarLift substold a -> (MetaVarLift substold newto)
                                                ConstantVarLift substold a -> (ConstantVarLift substold newto)
                                            ) s  | x <- startRepairConstantChains]
                                 } 
                            triplesCCVars
                            
     sAfterConstantChains    =  foldr (\(oldfrom,oldto,newto) e -> replaceTriples oldfrom oldto newto e) sAfterMetaChains2 triplesCCVars
     -- repair alphas of constantchains
     constantChains2 =
        [
        case sub of
           [] -> Left (sub,typ,name,z,s)
           (item:rest) -> 
              case lookup item [(b,a) | (a,b) <- alphaBaseAfterConstantChains] of
                Just m@(SCConstantChain sub0 typ0 name0) ->
                   if (SCConstantChain (Substitution []) typ name) == m then
                        let newAlpha = "ALPHA" ++ f
                            (SubstName _ oldAlpha) = item
                            newSubstItem = (SubstName (None m) newAlpha)
                        in Right (oldAlpha,newAlpha,newSubstItem,(newSubstItem):rest,typ,name,z,s)
                   else
                      Left (sub,typ,name,z,s)
       |          
        (ConstantChain (Substitution sub) typ name z s,f) <-  zip (constantChains envAfterConstantChains) freshAfterConstantChains
       ]
     freshAfterConstantChains2 = drop (length constantChains2) freshAfterConstantChains
     alphaBaseAfterConstantChains2 = alphaBaseAfterConstantChains ++
        [ (SCConstantChain (Substitution []) typ name, newSubstItem)
            | Right (oldAlpha,newAlpha,newSubstItem,_,typ,name,z,s) <- constantChains2
        ]
     envAfterConstantChains2 = foldr 
                            (\(sc,oldAlpha,newAlpha) env -> replaceInEnvBy sc newAlpha oldAlpha env)
                            envAfterConstantChains{
                                constantChains = [  case x of 
                                                Right (oldAlpha,newAlpha,newSubstItem,sub,typ,name,z,s)
                                                    -> ConstantChain (Substitution sub) typ name z s
                                                Left (sub,typ,name,z,s) 
                                                    -> ConstantChain (Substitution sub) typ name z s
                                            | x <- constantChains2
                                            ]}
                            
                           [ (SCConstantChain (Substitution []) typ name,oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,sub,typ,name,z,s) <- constantChains2]
     sAfterConstantChains2 =  foldr 
                            (\(sc,oldAlpha,newAlpha) e -> replaceInExprBy sc newAlpha oldAlpha e)
                            sAfterConstantChains
                            [ (SCConstantChain (Substitution []) typ name,oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,_,typ,name,z,s) <- constantChains2]         
                            
                            
     
     -- repair alphas of MetaEnvVars
     repairMetaEnvVars =
        [ case sub of
           [] -> Left (sub,e)
           (item:rest) -> 
              case lookup item [(b,a) | (a,b) <- alphaBaseAfterConstantChains2] of
                Just m@(SCMetaEnv (Env sub0 name0)) ->
                   if (SCMetaEnv (Env (Substitution []) e)) == m then
                        let newAlpha = "ALPHA" ++ f
                            (SubstName _ oldAlpha) = item
                            newSubstItem = (SubstName (None m) newAlpha)
                        in Right (oldAlpha,newAlpha,newSubstItem,(newSubstItem):rest,e)
                   else
                      Left (sub,e)
        |          
          (Env (Substitution sub) e,f) <- zip (metaEnvVars envAfterConstantChains2) freshAfterConstantChains2
       ]
     freshAfterMetaEnvVars = drop (length repairMetaEnvVars) freshAfterConstantChains2
     alphaBaseAfterMetaEnvVars = alphaBaseAfterConstantChains2 ++
        [ (SCMetaEnv (Env (Substitution []) name), newSubstItem)
            | Right (oldAlpha,newAlpha,newSubstItem,sub,name) <- repairMetaEnvVars
        ]
     envAfterMetaEnvVars = foldr 
                            (\(sc,oldAlpha,newAlpha) env -> replaceInEnvBy sc newAlpha oldAlpha env)
                            envAfterConstantChains2{
                                metaEnvVars = [case x of 
                                                Right (oldAlpha,newAlpha,newSubstItem,sub,e)
                                                    -> Env (Substitution sub) e
                                                Left (sub,e) 
                                                    -> Env (Substitution sub) e
                                              | x <- repairMetaEnvVars
                                            ]}
                            
                           [ (SCMetaEnv (Env (Substitution []) e),oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,sub,e) <- repairMetaEnvVars]
     sAfterMetaEnvVars =  foldr 
                            (\(sc,oldAlpha,newAlpha) e -> replaceInExprBy sc newAlpha oldAlpha e)
                            sAfterConstantChains2         
                           [ (SCMetaEnv (Env (Substitution []) e),oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,sub,e) <- repairMetaEnvVars]

     -- repair alphas of ConstantEnvVars
     repairConstantEnvVars =
        [ case sub of
           [] -> Left (sub,e)
           (item:rest) -> 
              case lookup item [(b,a) | (a,b) <- alphaBaseAfterMetaEnvVars] of
                Just m@(SCConstantEnv (Env sub0 name0)) ->
                   if (SCConstantEnv (Env (Substitution []) e)) == m then
                        let newAlpha = "ALPHA" ++ f
                            (SubstName _ oldAlpha) = item
                            newSubstItem = (SubstName (None m) newAlpha)
                        in Right (oldAlpha,newAlpha,newSubstItem,(newSubstItem):rest,e)
                   else
                      Left (sub,e)
        |          
          (Env (Substitution sub) e,f) <- zip (constantEnvVars envAfterMetaEnvVars) freshAfterMetaEnvVars
       ]
     freshAfterConstantEnvVars = drop (length repairConstantEnvVars) freshAfterMetaEnvVars
     alphaBaseAfterConstantEnvVars = alphaBaseAfterMetaEnvVars ++
        [ (SCConstantEnv (Env (Substitution []) name), newSubstItem)
            | Right (oldAlpha,newAlpha,newSubstItem,sub,name) <- repairConstantEnvVars
        ]
     envAfterConstantEnvVars = foldr 
                            (\(sc,oldAlpha,newAlpha) env -> replaceInEnvBy sc newAlpha oldAlpha env)
                            envAfterMetaEnvVars{
                                constantEnvVars = [case x of 
                                                Right (oldAlpha,newAlpha,newSubstItem,sub,e)
                                                    -> Env (Substitution sub) e
                                                Left (sub,e) 
                                                    -> Env (Substitution sub) e
                                              | x <- repairConstantEnvVars
                                            ]}
                            
                           [ (SCConstantEnv (Env (Substitution []) e),oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,sub,e) <- repairConstantEnvVars]
     sAfterConstantEnvVars =  foldr 
                            (\(sc,oldAlpha,newAlpha) e -> replaceInExprBy sc newAlpha oldAlpha e)
                            sAfterMetaEnvVars
                           [ (SCConstantEnv (Env (Substitution []) e),oldAlpha,newAlpha) | Right (oldAlpha,newAlpha,newSubstItem,sub,e) <- repairConstantEnvVars]                           
 in  (freshAfterConstantEnvVars,alphaBaseAfterConstantEnvVars,Letrec envAfterConstantEnvVars sAfterConstantEnvVars)
     
     
     
alphaRepair alphabase fresh expr (Fn f args) =    
    let go fr a (x:xs) = 
          let (f',a',x') = alphaRepairArgPos a fr expr x
              (f'',a'',xs') = go f'  a' xs
          in (f'',a'',x':xs')
        go fr a [] = (fr,a,[])
        (fresh',alphabase',args') = go fresh alphabase args
    in 
        (fresh',alphabase',Fn f args')
        
        
alphaRepair alphabase fresh expr e = (fresh,alphabase,e)


alphaRepairArgPos alphabase fresh expr (VarPos v) = (fresh,alphabase,VarPos v)
alphaRepairArgPos alphabase fresh expr (Binder [] s) =
    let (fresh',a',s') = alphaRepair alphabase fresh expr s
    in (fresh',a',Binder [] s')

alphaRepairArgPos alphabase fresh expr (Binder vs s) 
    | us@(u:_) <-
       let list = [lookup v [(y,x) | (SCVar x, SubstVar x' y) <- alphabase] | v <- vs] 
       in if all isJust list then map fromJust list else [] =


    let (fresh',a',s') = alphaRepair alphabase fresh expr s
        (newnames,fresh'') = splitAt (length us) fresh'
        newVarNames = [(varPrefix y)  ++ name | (name,y) <-  (zip newnames us)]
        varSubst = zipWith (\old new -> case old of 
                                    VarLift substold a -> (old,VarLift substold new)
                                    MetaVarLift substold a -> (old,MetaVarLift substold new)
                                    ConstantVarLift substold a -> (old,ConstantVarLift substold new)) us newVarNames
        triples = zipWith3 (\oldfrom oldto newto -> case oldfrom  of 
                                    VarLift substold a -> (oldfrom,oldto,VarLift substold newto)
                                    MetaVarLift substold a -> (oldfrom,oldto,MetaVarLift substold newto)
                                    ConstantVarLift substold a -> (oldfrom,oldto,ConstantVarLift substold newto)) us vs newVarNames 
        newVars = map snd varSubst
        newAlphaBase = a' ++ (zipWith (\x y ->  (SCVar x, SubstVar x y)) us newVars)
        s'' = foldr (\(oldfrom,oldto,newto) e -> replaceTriples oldfrom oldto newto e) s' triples
    in 
        (fresh'', newAlphaBase, Binder newVars s'')
        
alphaRepairArgPos alphabase fresh expr (Binder (v:vs) s)  
 = let (f',a',Binder vs' s') = alphaRepairArgPos alphabase fresh expr (Binder vs s)   
   in (f',a',Binder (v:vs') s')

   
   
getUsedRenamingNamesExpr (MetaExpression (Substitution sub) n) =
    getUsedRenamingNamesSubst sub
    
getUsedRenamingNamesExpr (Constant (Substitution sub) n) =
    getUsedRenamingNamesSubst sub

getUsedRenamingNamesExpr (MetaCtxtSubst (Substitution sub) srt name e) =
    Set.union (getUsedRenamingNamesSubst sub) (getUsedRenamingNamesExpr e)
    
getUsedRenamingNamesExpr (ConstantCtxtSubst (Substitution sub) srt name e) =
    Set.union (getUsedRenamingNamesSubst sub) (getUsedRenamingNamesExpr e)
    
getUsedRenamingNamesExpr (MetaCtxt (Substitution sub) srt name) =
    (getUsedRenamingNamesSubst sub) 
    
getUsedRenamingNamesExpr (ConstantCtxt (Substitution sub) srt name) =
    (getUsedRenamingNamesSubst sub) 
    
getUsedRenamingNamesExpr (Fn f args) =
   Set.unions (map getUsedRenamingNamesArgPos args) 

getUsedRenamingNamesExpr (VarL var) =
   getUsedRenamingNamesVar var
   
getUsedRenamingNamesExpr (Letrec env expr) =
   Set.union (getUsedRenamingNamesEnv env) (getUsedRenamingNamesExpr expr)
   
getUsedRenamingNamesEnv env =
    Set.unions
        [(Set.unions [Set.union (getUsedRenamingNamesVar x) (getUsedRenamingNamesExpr e)  | (x := e) <- bindings env])
        ,(Set.unions [Set.unions [(getUsedRenamingNamesVar z)
                                ,(getUsedRenamingNamesExpr e)
                                ,(getUsedRenamingNamesSubst sub)]
                                | (MetaChain (Substitution sub) srt n z e) <- metaChains env])
        ,(Set.unions [Set.unions [(getUsedRenamingNamesVar z)
                                ,(getUsedRenamingNamesExpr e)
                                ,(getUsedRenamingNamesSubst sub)]
                                | (ConstantChain (Substitution sub) srt n z e) <- constantChains env])
        ,(Set.unions [(getUsedRenamingNamesSubst sub)  | (Env (Substitution sub) e) <- metaEnvVars env])
        ,(Set.unions [(getUsedRenamingNamesSubst sub)  | (Env (Substitution sub) e) <- constantEnvVars env])    
        ]
getUsedRenamingNamesVar (MetaVarLift (Substitution sub) n) =
    Set.union (getUsedRenamingNamesSubst sub) (Set.fromList [n])
getUsedRenamingNamesVar (ConstantVarLift (Substitution sub) n) =
    Set.union (getUsedRenamingNamesSubst sub) (Set.fromList [n])
getUsedRenamingNamesVar (VarLift (Substitution sub) n) =
    Set.union (getUsedRenamingNamesSubst sub) (Set.fromList [n])
    
getUsedRenamingNamesSubst sub =
    Set.unions (map getUsedRenamingNamesSubstItem sub)
getUsedRenamingNamesSubstItem (SubstName (None comp) n) =
    Set.union (getUsedRenamingNamesSCComp comp ) (Set.fromList [n])
getUsedRenamingNamesSubstItem (SubstName (LetVars comp) n) =
    Set.union (getUsedRenamingNamesSCComp comp ) (Set.fromList [n])
getUsedRenamingNamesSubstItem (SubstName (CapVars comp) n) =
    Set.union (getUsedRenamingNamesSCComp comp ) (Set.fromList [n])
getUsedRenamingNamesSubstItem (SubstVar v1 v2) = 
    Set.unions (map getUsedRenamingNamesVar  [v1,v2])
getUsedRenamingNamesSubstItem (SubstSet s) = 
    getUsedRenamingNamesSubst s

getUsedRenamingNamesArgPos (VarPos v) = getUsedRenamingNamesVar v
getUsedRenamingNamesArgPos (Binder vs e) = 
    Set.unions ((getUsedRenamingNamesExpr e):(map getUsedRenamingNamesVar vs))
    
getUsedRenamingNamesSCComp (SCExpr e) = getUsedRenamingNamesExpr e    
getUsedRenamingNamesSCComp (SCVar v) = getUsedRenamingNamesVar v  
getUsedRenamingNamesSCComp (SCCtxt e) = getUsedRenamingNamesExpr e    
getUsedRenamingNamesSCComp (SCMetaEnv (Env (Substitution sub) e)) = getUsedRenamingNamesSubst sub
getUsedRenamingNamesSCComp (SCConstantEnv (Env (Substitution sub) e)) = getUsedRenamingNamesSubst sub
getUsedRenamingNamesSCComp (SCMetaChain (Substitution sub) _ _ ) = getUsedRenamingNamesSubst sub
getUsedRenamingNamesSCComp (SCConstantChain (Substitution sub) _ _ ) = getUsedRenamingNamesSubst sub



    
    
    
   
   
   
replaceTriplesSubst :: VarLift -> VarLift -> VarLift -> [SubstItem] -> [SubstItem]    
replaceTriplesSubst oldfrom oldto newto [] = []
replaceTriplesSubst oldfrom oldto newto (item@(SubstVar from to):rest)
    | oldfrom == from && oldto == to = (SubstVar from newto):(replaceTriplesSubst oldfrom oldto newto rest)
    | otherwise = item:(replaceTriplesSubst oldfrom oldto newto rest)
    
replaceTriplesSubst oldfrom oldto newto ((SubstSet set):rest)
 = (SubstSet (replaceTriplesSubst oldfrom oldto newto set)):(replaceTriplesSubst oldfrom oldto newto rest)

replaceTriplesSubst oldfrom oldto newto ((SubstName (None comp) n:rest))
 = (SubstName (None $ replaceTriplesComp oldfrom oldto newto comp) n):(replaceTriplesSubst oldfrom oldto newto rest)
     
replaceTriplesSubst oldfrom oldto newto ((SubstName (LetVars comp) n:rest))
 = (SubstName (LetVars $ replaceTriplesComp oldfrom oldto newto comp) n):(replaceTriplesSubst oldfrom oldto newto rest)

replaceTriplesSubst oldfrom oldto newto ((SubstName (CapVars comp) n:rest))
 = (SubstName (CapVars $ replaceTriplesComp oldfrom oldto newto comp) n):(replaceTriplesSubst oldfrom oldto newto rest)
 

replaceTriplesVar oldfrom oldto newto v@(MetaVarLift (Substitution sub) var) 
    | varname oldto == var = MetaVarLift (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) (varname newto)
    | otherwise = MetaVarLift (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) var
    where varname (MetaVarLift _ v) = v
          varname (ConstantVarLift _ v) = v
          varname (VarLift _ v) = v
          
 
replaceTriplesVar oldfrom oldto newto v@(VarLift (Substitution sub) var) 
    | varname oldto == var = VarLift (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) (varname newto)
    | otherwise = VarLift (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) var
    where varname (MetaVarLift _ v) = v
          varname (ConstantVarLift _ v) = v
          varname (VarLift _ v) = v
          
replaceTriplesVar oldfrom oldto newto v@(ConstantVarLift (Substitution sub) var) 
    | varname oldto == var = ConstantVarLift (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) (varname newto)
    | otherwise = ConstantVarLift (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) var
    where varname (MetaVarLift _ v) = v
          varname (ConstantVarLift _ v) = v
          varname (VarLift _ v) =  v

    
replaceTriplesComp :: VarLift -> VarLift -> VarLift -> SyntaxComponent -> SyntaxComponent    
replaceTriplesComp oldfrom oldto newto (SCVar var) =
    SCVar (replaceTriplesVar oldfrom oldto newto var)
    
replaceTriplesComp oldfrom oldto newto (SCExpr e) =
    SCExpr (replaceTriples oldfrom oldto newto e)
replaceTriplesComp oldfrom oldto newto (SCCtxt e) =
    SCCtxt (replaceTriples oldfrom oldto newto e)
    
replaceTriplesComp oldfrom oldto newto (SCMetaEnv (Env (Substitution sub) e)) =
    SCMetaEnv (Env (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) e)
replaceTriplesComp oldfrom oldto newto (SCConstantEnv (Env (Substitution sub) e)) =
    SCConstantEnv (Env (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) e)
replaceTriplesComp oldfrom oldto newto (SCMetaChain (Substitution sub) typ e) =
    SCMetaChain (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) typ e
replaceTriplesComp oldfrom oldto newto (SCConstantChain (Substitution sub) typ e) =
    SCConstantChain (Substitution $ replaceTriplesSubst oldfrom oldto newto sub) typ e
   

replaceTriples :: VarLift -> VarLift -> VarLift -> Expression -> Expression
replaceTriples oldfrom oldto newto (MetaExpression (Substitution subs) s) =
    MetaExpression (Substitution (replaceTriplesSubst oldfrom oldto newto subs)) s
    
replaceTriples  oldfrom oldto newto (Constant (Substitution subs) s) =
    Constant (Substitution (replaceTriplesSubst oldfrom oldto newto subs)) s

replaceTriples  oldfrom oldto newto (VarL v) =
    VarL (replaceTriplesVar oldfrom oldto newto v) 
    
replaceTriples  oldfrom oldto newto (MetaCtxtSubst (Substitution sub) srt name e) =
  MetaCtxtSubst 
        (Substitution (replaceTriplesSubst  oldfrom oldto newto sub)) 
        srt 
        name 
        (replaceTriples  oldfrom oldto newto e)
    
replaceTriples  oldfrom oldto newto (ConstantCtxtSubst (Substitution sub) srt name e) =
  ConstantCtxtSubst 
        (Substitution (replaceTriplesSubst  oldfrom oldto newto sub)) 
        srt 
        name 
        (replaceTriples  oldfrom oldto newto e)    
      
replaceTriples  oldfrom oldto newto (Fn f args)  =
  Fn f (map (replaceTriplesArgPos oldfrom oldto newto) args)


replaceTriples  oldfrom oldto newto (Letrec env e) =
   Letrec (replaceTriplesEnv oldfrom oldto newto env) (replaceTriples oldfrom oldto newto e)
        
replaceTriples  oldfrom oldto newto (MetaCtxt (Substitution sub) srt name) =
  MetaCtxt
        (Substitution (replaceTriplesSubst  oldfrom oldto newto sub)) 
        srt 
        name 

    
replaceTriples  oldfrom oldto newto (ConstantCtxt (Substitution sub) srt name) =
  ConstantCtxt
        (Substitution (replaceTriplesSubst  oldfrom oldto newto sub)) 
        srt 
        name 
        
replaceTriples  oldfrom oldto newto (Hole (Substitution sub)) =
  Hole
        (Substitution (replaceTriplesSubst  oldfrom oldto newto sub)) 
      
replaceTriples _ _ _ other = error $ "in replaceTriples " ++ show other
        
    
replaceTriplesEnv  oldfrom oldto newto env =

  let  
        newBindings = [(replaceTriplesVar oldfrom oldto newto x) := (replaceTriples oldfrom oldto newto e) |  (x:=e) <- bindings env]
        newMetaEnvVars = [(Env (Substitution (replaceTriplesSubst oldfrom oldto newto sub)) e) | (Env (Substitution sub) e) <- metaEnvVars env]
        newConstantEnvVars = [(Env (Substitution (replaceTriplesSubst oldfrom oldto newto sub)) e) | (Env (Substitution sub) e) <- constantEnvVars env]
        newMetaChains = [MetaChain 
                            (Substitution (replaceTriplesSubst oldfrom oldto newto sub))
                            typ 
                            name
                            (replaceTriplesVar oldfrom oldto newto z)
                            (replaceTriples oldfrom oldto newto s)
                        |   (MetaChain (Substitution sub) typ name z s) <- metaChains env
                        ]
        newConstantChains = 
                             [ConstantChain 
                              (Substitution (replaceTriplesSubst oldfrom oldto newto sub))
                              typ 
                              name
                              (replaceTriplesVar oldfrom oldto newto z)
                              (replaceTriples oldfrom oldto newto s)
                        |     (ConstantChain (Substitution sub) typ name z s) <- constantChains env
                        ]
       in  env{bindings=newBindings
                    ,metaEnvVars=newMetaEnvVars
                    ,constantEnvVars = newConstantEnvVars
                    ,metaChains=newMetaChains
                    ,constantChains=newConstantChains}
       
        
   
replaceTriplesArgPos oldfrom oldto newto (Binder vs e)
    | null [v | v <- vs, oldto == v] = 
         Binder (map (replaceTriplesVar oldfrom oldto newto) vs) (replaceTriples oldfrom oldto newto e)
    | otherwise = 
         Binder vs e

    
replaceInSubstListBy :: SyntaxComponent -> String -> String -> [SubstItem] -> [SubstItem]    
replaceInSubstListBy syntaxcomponent new old ((SubstName (None comp) name):rest)
    | name == old && comp == syntaxcomponent = ((SubstName (None comp) new):(replaceInSubstListBy syntaxcomponent new old rest))
    
replaceInSubstListBy syntaxcomponent new old ((SubstName (LetVars comp) name):rest)
    | name == old && comp == syntaxcomponent = ((SubstName (LetVars comp) new):(replaceInSubstListBy syntaxcomponent new old rest))
    
replaceInSubstListBy syntaxcomponent new old ((SubstName (CapVars comp) name):rest)
    | name == old && comp == syntaxcomponent = ((SubstName (CapVars comp) new):(replaceInSubstListBy syntaxcomponent new old rest))
    
replaceInSubstListBy syntaxcomponent new old ((SubstSet set):rest) =
    let set' = replaceInSubstListBy syntaxcomponent new old set 
        
    in (SubstSet set'):(replaceInSubstListBy syntaxcomponent new old rest)
replaceInSubstListBy syntaxcomponent new old (item:rest) =
        item:(replaceInSubstListBy syntaxcomponent new old rest)
replaceInSubstListBy _ _ _ [] = []        
    
    
replaceInEnvBy :: SyntaxComponent -> String -> String -> Environment -> Environment    
replaceInEnvBy sc newAlpha oldAlpha env = 
 let 
  metaChains' = 
    [
    (MetaChain 
        (Substitution (replaceInSubstListBy sc newAlpha oldAlpha sub))
        typ
        name
        (replaceInVarBy sc newAlpha oldAlpha z)
        (replaceInExprBy sc newAlpha oldAlpha s))
            |  (MetaChain (Substitution sub) typ name z s) <- metaChains env]
  constantChains' = 
    [
    (ConstantChain 
        (Substitution (replaceInSubstListBy sc newAlpha oldAlpha sub))
        typ
        name
        (replaceInVarBy sc newAlpha oldAlpha z)
        (replaceInExprBy sc newAlpha oldAlpha s))
            |  (ConstantChain (Substitution sub) typ name z s) <- constantChains env]
  metaEnvVars' = [ Env (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) e | (Env (Substitution sub) e) <- metaEnvVars env]
  constantEnvVars' = [ Env (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) e | (Env (Substitution sub) e) <- constantEnvVars env]  
  bindings' = [ 
               (replaceInVarBy sc newAlpha oldAlpha x)
               :=
               (replaceInExprBy sc newAlpha oldAlpha e)
               | (x:=e) <- bindings env
               ]
 in env{metaChains=metaChains', constantChains=constantChains', metaEnvVars=metaEnvVars', constantEnvVars= constantEnvVars', bindings=bindings'}

replaceInVarBy :: SyntaxComponent -> String -> String -> VarLift -> VarLift
replaceInVarBy sc newAlpha oldAlpha (VarLift (Substitution sub) v) =
    (VarLift (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) v)
     
replaceInVarBy sc newAlpha oldAlpha (MetaVarLift (Substitution sub) v) =
    (MetaVarLift (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) v)
    
replaceInVarBy sc newAlpha oldAlpha (ConstantVarLift (Substitution sub) v) =
    (ConstantVarLift (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) v)

replaceInExprBy :: SyntaxComponent -> String -> String -> Expression -> Expression

replaceInExprBy sc newAlpha oldAlpha  (MetaExpression (Substitution sub) s) =
    MetaExpression (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) s

replaceInExprBy sc newAlpha oldAlpha  (Constant (Substitution sub) s) 
    | otherwise = Constant (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub) s

replaceInExprBy sc newAlpha oldAlpha  (VarL v) =
    VarL (replaceInVarBy sc newAlpha oldAlpha v) 

replaceInExprBy sc newAlpha oldAlpha  (Letrec env s) =
    Letrec 
        (replaceInEnvBy sc newAlpha oldAlpha env) 
        (replaceInExprBy sc newAlpha oldAlpha s)
                                  
replaceInExprBy sc newAlpha oldAlpha  (MetaCtxtSubst (Substitution sub) srt name e) =
    MetaCtxtSubst
     (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub)
     srt
     name
     (replaceInExprBy sc newAlpha oldAlpha e)
     
replaceInExprBy sc newAlpha oldAlpha  (MetaCtxt (Substitution sub) srt name) =
    MetaCtxt
     (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub)
     srt
     name
replaceInExprBy sc newAlpha oldAlpha  (ConstantCtxt (Substitution sub) srt name) =
    ConstantCtxt
     (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub)
     srt
     name
replaceInExprBy sc newAlpha oldAlpha  (ConstantCtxtSubst (Substitution sub) srt name e) =
    ConstantCtxtSubst
     (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub)
     srt
     name
     (replaceInExprBy sc newAlpha oldAlpha e)
     
replaceInExprBy sc newAlpha oldAlpha  (Hole (Substitution sub)) =
    Hole
     (Substitution $ replaceInSubstListBy sc newAlpha oldAlpha sub)
     
replaceInExprBy sc newAlpha oldAlpha  (Fn f args) =
  Fn f (map (replaceInArgPosBy sc newAlpha oldAlpha) args)

replaceInArgPosBy sc newAlpha oldAlpha (VarPos v)
 = VarPos (replaceInVarBy sc newAlpha oldAlpha v)
 
replaceInArgPosBy sc newAlpha oldAlpha (Binder vs e)
 = Binder (map (replaceInVarBy sc newAlpha oldAlpha) vs)
          (replaceInExprBy sc newAlpha oldAlpha e)
 
 
-- ********************************************************************************
-- The main function to check alpha-equivalence of two expressions 
alphaEquiv fresh expr1 expr2 =
    let (res,_,fullScope) = alphaEquivWithScopeExpr fresh (Map.empty,Map.empty) expr1 expr2 
    in 
      -- if res then 
        -- trace 
        -- (show(expr1,expr2))
         -- (trace (show fullScope) (res,fullScope))
       -- else 
       (res,fullScope)
-- types and helper functions for renamings
type ScopeRenaming = (Map.Map String String, Map.Map String String)
type SingleFullRenaming = Map.Map String [String]
type FullRenaming  = (SingleFullRenaming, SingleFullRenaming)  -- two renamings for the two expressions shown to be alpha-equivalent

insertIntoMultiSet :: String -> String -> Map.Map String [String] -> Map.Map String [String]
insertIntoMultiSet key value mp 
 | Just entry <- Map.lookup key mp = Map.insert key (value:entry) mp
 | otherwise                       = Map.insert key [value] mp 
 
emptyFullRenaming :: FullRenaming
emptyFullRenaming = (Map.empty,Map.empty)

unionMultiSets :: Map.Map String [String] 
                -> Map.Map String [String]
                 -> Map.Map String [String]
                 
unionMultiSets mp1 mp2 = foldr 
                             (\(k,v) m -> insertIntoMultiSet k v m) 
                             mp2 
                            [(k,v) | (k,vlist) <- (Map.assocs mp1),v <- vlist]

unionListOfMultiSets :: [Map.Map String [String]] -> Map.Map String [String]
unionListOfMultiSets list =  foldr (\m1 m2 -> unionMultiSets m1 m2) Map.empty list 

unionListOfFullScopes :: [FullRenaming] -> FullRenaming
unionListOfFullScopes list = 
 let (leftComps,rightComps) = unzip list 
 in (unionListOfMultiSets leftComps,unionListOfMultiSets rightComps)     
--
  

alphaEquivWithScopeExpr :: [String]          -- ^ fresh names
                           -> ScopeRenaming 
                           -> Expression 
                           -> Expression 
                           -> (Bool
                              ,[String]
                              ,FullRenaming) 
alphaEquivWithScopeExpr fresh scope@(scope1,scope2) expr1@(MetaExpression sub1 e1) expr2@(MetaExpression sub2 e2) 
 | (e1 == e2) = 
     let 
      alpha1 = case sub1 of
                (Substitution ((SubstName (None (SCExpr (MetaExpression _ e1'))) alpha1):_)) -> if e1 == e1' then [alpha1] else error "in alphaEquiv"
                (Substitution []) -> []
      alpha2 = case sub2  of
                (Substitution ((SubstName (None (SCExpr (MetaExpression _ e2'))) alpha2):_)) -> if e2 == e2' then [alpha2] else error "in alphaEquiv"
                (Substitution []) -> []
       in
        case () of 
         _ | null alpha1 && null alpha2 ->  (alphaEquivWithScopeSubst scope sub1 sub2,fresh,emptyFullRenaming)
           | null alpha1 || null alpha2  -> (False,fresh,emptyFullRenaming) 
           | otherwise ->
              let 
               (newAlpha:fresh')= fresh
               alpha1' = head alpha1
               alpha2' = head alpha2
               scope' = (Map.insert alpha1' newAlpha scope1,Map.insert alpha2' newAlpha scope2)
               (m1,m2) = emptyFullRenaming
              in (alphaEquivWithScopeSubst scope' sub1 sub2
                 ,fresh'
                 ,(insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))           
              
alphaEquivWithScopeExpr fresh scope@(scope1,scope2) expr1@(Constant sub1 e1) expr2@(Constant sub2 e2) 
 | (e1 == e2) = 
     let 
      alpha1 = case sub1 of
                (Substitution ((SubstName (None (SCExpr (Constant _ e1'))) alpha1):_)) -> if e1 == e1' then [alpha1] else error "in alphaEquiv"
                (Substitution []) -> []
      alpha2 = case sub2  of
                (Substitution ((SubstName (None (SCExpr (Constant _ e2'))) alpha2):_)) -> if e2 == e2' then [alpha2] else error "in alphaEquiv"
                (Substitution []) -> []
       in
        case () of 
         _ | null alpha1 && null alpha2 ->  (alphaEquivWithScopeSubst scope sub1 sub2,fresh,emptyFullRenaming)
           | null alpha1 || null alpha2  -> (False,fresh,emptyFullRenaming) 
           | otherwise ->
              let
               (newAlpha:fresh')= fresh
               alpha1' = head alpha1
               alpha2' = head alpha2
               scope' = (Map.insert alpha1' newAlpha scope1,Map.insert alpha2' newAlpha scope2)
               (m1,m2) = emptyFullRenaming
              in (alphaEquivWithScopeSubst scope' sub1 sub2
                 ,fresh'
                 ,(insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))           

alphaEquivWithScopeExpr fresh scope   (Fn f1 args1) (Fn f2 args2) 
    | f1 == f2  = 
       let 
         go fr [] []  = (True,fr,emptyFullRenaming)
         go fr (a:as) (b:bs) =
          let (isEquiv, fr', fullscopeLocal) = alphaEquivWithScopeArgPos  fr scope a b
              (restEquiv, fr'', fullscopeRest) = go fr' as bs
          in  (isEquiv && restEquiv, fr'', unionListOfFullScopes [fullscopeLocal,fullscopeRest]) 
       in 
         go fresh args1 args2
    | otherwise = (False,fresh,emptyFullRenaming)
    
alphaEquivWithScopeExpr fresh scope (VarL v1) (VarL v2) =
    (alphaEquivWithScopeVar scope v1 v2,fresh,emptyFullRenaming)

alphaEquivWithScopeExpr fresh scope@(scope1,scope2)  c1@(MetaCtxtSubst sub1 srt1 name1 e1) c2@(MetaCtxtSubst sub2 srt2 name2 e2)  
    | srt1 == srt2 
    , name1 == name2 
    =   
        let 
            alpha1 =
               case sub1 of
                 (Substitution ((SubstName (None (SCCtxt (MetaCtxt _ _ name1'))) a1):_))
                   -> if name1 == name1' then [a1] else error "in alphaEquiv"
                 (Substitution []) -> []   
            alpha2 =
               case sub2  of
                 (Substitution ((SubstName (None (SCCtxt (MetaCtxt _ _ name2'))) a2):_))
                   -> if name2 == name2' then [a2] else error "in alphaEquiv"
                 (Substitution []) -> []   
        in case () of
            _ | (null alpha1 && null alpha2) -> 
                let b1 = alphaEquivWithScopeSubst scope sub1 sub2
                    (b2,fresh1,sc2) = alphaEquivWithScopeExpr fresh scope e1 e2
                in (b1 && b2,fresh1,sc2)
            _ | null alpha1 || null alpha2 -> (False,fresh,emptyFullRenaming) 
            _ | otherwise ->
                  let 
                    (newAlpha:fresh')= fresh
                    alpha1' = head alpha1
                    alpha2' = head alpha2
                    scope' = (Map.insert alpha1' newAlpha scope1,Map.insert alpha2' newAlpha scope2)
                    b1 = alphaEquivWithScopeSubst scope' sub1 sub2
                    (b2,fresh2,sc2) = alphaEquivWithScopeExpr  fresh' scope' e1 e2
                    (m1,m2) = sc2
                   in (b1 && b2, fresh2, (insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))

alphaEquivWithScopeExpr fresh scope@(scope1,scope2)  c1@(ConstantCtxtSubst sub1 srt1 name1 e1) c2@(ConstantCtxtSubst sub2 srt2 name2 e2)  
    | srt1 == srt2 
    , name1 == name2 
    =   
        let 
            alpha1 =
               case sub1 of
                 (Substitution ((SubstName (None (SCCtxt (ConstantCtxt _ _ name1'))) a1):_))
                   -> if name1 == name1' then [a1] else error "in alphaEquiv"
                 (Substitution []) -> []   
            alpha2 =
               case sub2  of
                 (Substitution ((SubstName (None (SCCtxt (ConstantCtxt _ _ name2'))) a2):_))
                   -> if name2 == name2' then [a2] else error "in alphaEquiv"
                 (Substitution []) -> []   
        in case () of
            _ | (null alpha1 && null alpha2) -> 
                let b1 = alphaEquivWithScopeSubst scope sub1 sub2
                    (b2,fresh1,sc2) = alphaEquivWithScopeExpr fresh scope e1 e2
                in (b1 && b2,fresh1,sc2)
            _ | null alpha1 || null alpha2 -> (False,fresh,emptyFullRenaming) 
            _ | otherwise ->
                  let 
                    (newAlpha:fresh')= fresh
                    alpha1' = head alpha1
                    alpha2' = head alpha2
                    scope' = (Map.insert alpha1' newAlpha scope1,Map.insert alpha2' newAlpha scope2)
                    b1 = alphaEquivWithScopeSubst scope' sub1 sub2
                    (b2,fresh2,sc2) = alphaEquivWithScopeExpr fresh' scope' e1 e2
                    (m1,m2) = sc2
                   in (b1 && b2, fresh2, (insertIntoMultiSet alpha1' newAlpha m1, insertIntoMultiSet alpha2' newAlpha m2))

alphaEquivWithScopeExpr fresh scope  (Letrec env1 inexpr1) (Letrec env2 inexpr2) =   
 let 
   -- first  assign components to components            
    bindings1 = bindings env1
    bindings2 = bindings env2
    -- bindingsAssignments  = [zip bindings1 bs2 | bs2 <- permutations bindings2] 
    bindingsAssignments =
     let
      mergeGroups :: [([Binding],[Binding])] -> [[Binding]] -> [[Binding]] -> [([Binding],[Binding])] 
      mergeGroups res [] []  = res
      mergeGroups res (g1@((_ := b1):_):gs1) (g2@((_ := b2):_):gs2) 
       | fingerprint b1 == fingerprint b2 = (mergeGroups ((g1,g2):res) gs1 gs2)
      mergeGroups _ _ _ = []
      assignGroups [] = [[]]
      assignGroups ((g1,g2):rest)
       | length g1 == length g2 = 
          let rek = assignGroups rest
              now = [zip g1 g2' | g2' <- permutations g2]
          in  [n ++ r | n <- now, r <- rek]
       | otherwise = []
     in 
       assignGroups $
         mergeGroups []  
          (groupBy (\(_ := b1) (_ := b2) -> (fingerprint b1) == (fingerprint b2))
             (sortBy (\(_ := b1) (_ := b2) -> compare (fingerprint b1) (fingerprint b2)) bindings1))
          (groupBy (\(_ := b1) (_ := b2) -> (fingerprint b1) == (fingerprint b2))
             (sortBy (\(_ := b1) (_ := b2) -> compare (fingerprint b1) (fingerprint b2)) bindings2))


    mc1       = metaChains env1
    mc2       = metaChains env2            
    -- metaChainAssignments = [zip mc1 mcs2 | mcs2 <- permutations mc2]            
    metaChainAssignments =             
     let
      mergeGroups res [] []  = res
      mergeGroups res (g1@((MetaChain _ typ1 name1 _ s1):_):gs1) (g2@((MetaChain _ typ2 name2 _ s2):_):gs2) 
       | name1 == name2
       , typ1 == typ2
       , fingerprint s1 == fingerprint s2  = (mergeGroups ((g1,g2):res) gs1 gs2)
      mergeGroups _ _ _ = []
      assignGroups [] = [[]]
      assignGroups ((g1,g2):rest)
       | length g1 == length g2 = 
          let rek = assignGroups rest
              now = [zip g1 g2' | g2' <- permutations g2]
          in  [n ++ r | n <- now, r <- rek]
       | otherwise = []
      sortFn (MetaChain _ typ1 name1 _ s1) (MetaChain _ typ2 name2 _ s2) =
        case compare typ1 typ2 of
          EQ -> case compare name1 name2 of
                   EQ -> compare (fingerprint s1) (fingerprint s2)
                   other -> other
          other -> other
      groupFn m1 m2 = sortFn m1 m2 == EQ
     in 
       assignGroups $
         mergeGroups []  
          (groupBy groupFn
             (sortBy sortFn mc1))
          (groupBy groupFn
             (sortBy sortFn mc2))

    cc1       = constantChains env1
    cc2       = constantChains env2            
    -- constantChainAssignments = [zip cc1 ccs2 | ccs2 <- permutations cc2]            
    constantChainAssignments = 
     let
      mergeGroups res [] []  = res
      mergeGroups res (g1@((ConstantChain _ typ1 name1 _ s1):_):gs1) (g2@((ConstantChain _ typ2 name2 _ s2):_):gs2) 
       | name1 == name2
       , typ1 == typ2
       , fingerprint s1 == fingerprint s2  = (mergeGroups ((g1,g2):res) gs1 gs2)
      mergeGroups _ _ _ = []
      assignGroups [] = [[]]
      assignGroups ((g1,g2):rest)
       | length g1 == length g2 = 
          let rek = assignGroups rest
              now = [zip g1 g2' | g2' <- permutations g2]
          in  [n ++ r | n <- now, r <- rek]
       | otherwise = []
      sortFn (ConstantChain _ typ1 name1 _ s1) (ConstantChain _ typ2 name2 _ s2) =
        case compare typ1 typ2 of
          EQ -> case compare name1 name2 of
                   EQ -> compare (fingerprint s1) (fingerprint s2)
                   other -> other
          other -> other
      groupFn m1 m2 = sortFn m1 m2 == EQ
     in 
       assignGroups $
         mergeGroups []  
          (groupBy groupFn
             (sortBy sortFn cc1))
          (groupBy groupFn
             (sortBy sortFn cc2))
    me1       = metaEnvVars env1
    me2       = metaEnvVars env2            
    -- metaEnvsAssignments = [zip me1 mes2 | mes2 <- permutations me2]                        
    metaEnvsAssignments = 
     let
      mergeGroups res [] []  = res
      mergeGroups res (g1@((Env _ n1):_):gs1) (g2@((Env _ n2):_):gs2) 
       | n1 == n2 = (mergeGroups ((g1,g2):res) gs1 gs2)
      mergeGroups _ _ _ = []
      assignGroups [] = [[]]
      assignGroups ((g1,g2):rest)
       | length g1 == length g2 = 
          let rek = assignGroups rest
              now = [zip g1 g2' | g2' <- permutations g2]
          in  [n ++ r | n <- now, r <- rek]
       | otherwise = []
      sortFn (Env _ n1) (Env _ n2) = compare n1 n2
      groupFn m1 m2 = sortFn m1 m2 == EQ
     in 
       assignGroups $
         mergeGroups []  
          (groupBy groupFn
             (sortBy sortFn me1))
          (groupBy groupFn
             (sortBy sortFn me2))
    ce1       = constantEnvVars env1
    ce2       = constantEnvVars env2
    -- constantEnvsAssignments = [zip ce1 ces2 | ces2 <- permutations ce2]                                    
    constantEnvsAssignments =
     let
      mergeGroups res [] []  = res
      mergeGroups res (g1@((Env _ n1):_):gs1) (g2@((Env _ n2):_):gs2) 
       | n1 == n2 =(mergeGroups ((g1,g2):res) gs1 gs2)
      mergeGroups _ _ _ = []
      assignGroups [] = [[]]
      assignGroups ((g1,g2):rest)
       | length g1 == length g2 = 
          let rek = assignGroups rest
              now = [zip g1 g2' | g2' <- permutations g2]
          in  [n ++ r | n <- now, r <- rek]
       | otherwise = []
      sortFn (Env _ n1) (Env _ n2) = compare n1 n2
      groupFn m1 m2 = sortFn m1 m2 == EQ
     in 
       assignGroups $
         mergeGroups []  
          (groupBy groupFn
             (sortBy sortFn ce1))
          (groupBy groupFn
             (sortBy sortFn ce2))
    allAssignments = [(aBS,aMC,aCC,aME,aCE) 
                        |  aBS <- bindingsAssignments
                        ,  aMC <- metaChainAssignments
                        ,  aCC <- constantChainAssignments
                        ,  aME <- metaEnvsAssignments
                        ,  aCE <- constantEnvsAssignments]
    -- now use the assignments to proceed further, first all binders must be renamed
    -- and the scopes have to be adjusted
    allChecksForBinders = 
     [   let
           (ok1,fresh1,scope1,fullScope1) =
             let 
               go fr [] scope  = (True,fr,scope,emptyFullRenaming) 
               go fr ((x1:=_,x2:=_):rest) scope = 
                  let 
                    (ok',fr',scope',fsc)      = alphaRenameBoundVar fr scope x1 x2  
                    (ok'',fr'',scope'',fsc'') = go fr' rest scope'  
                  in (ok' && ok'',fr'', scope'', unionListOfFullScopes [fsc,fsc''])                    
              in go fresh aBS scope 
              
           (ok2,fresh2,scope2,fullScope2) =
             let 
               go fr [] scope = (True,fr,scope,emptyFullRenaming) 
               go fr ((MetaChain sub1 typ1 name1 z1 s1,MetaChain sub2 typ2 name2 z2 s2):rest) scope@(scope1,scope2) = 
                  let 
                    (ok',fr',scope',fsc)      = alphaRenameBoundVar fr scope z1 z2  
                    alpha1 = case sub1 of 
                               (Substitution ((SubstName (None (SCMetaChain sub1' typ1' name1')) a1):_) )
                                  -> if typ1 == typ1' && name1 == name1' then [a1] else error "in alphaEquiv"
                               (Substitution []) -> []
                    alpha2 = case sub2 of 
                               (Substitution ((SubstName (None (SCMetaChain sub2' typ2' name2')) a2):_) )
                                  -> if typ2 == typ2' && name2 == name2' then [a2] else error "in alphaEquiv"
                               (Substitution []) -> []
                    (okR,frR,scR,fscR) =
                       case () of 
                         _ | (null alpha1 && null alpha2)  -> (True,fr',scope',fsc)
                           |  null alpha1 || null alpha2   -> (False,fr',scope',fsc)
                           | otherwise ->
                                let 
                                  (scope1',scope2') = scope'
                                  (newAlpha:fr'')= fr'
                                  alpha1' = head alpha1
                                  alpha2' = head alpha2
                                  scope'' = (Map.insert alpha1' newAlpha scope1',Map.insert alpha2' newAlpha scope2')
                                  (m1,m2) = fsc
                                in (ok'&&typ1 == typ2&&name1 == name2,fr'',scope'',(insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))
                    (okRest,frRest,scopeRest,fscRest) = (go frR rest scR) 
                  in (okR && okRest, frRest, scopeRest, unionListOfFullScopes [fscR,fscRest])
              in go fresh1  aMC scope1
              
              
           (ok3,fresh3,scope3,fullScope3) =
             let 
               go fr [] scope = (True,fr,scope,emptyFullRenaming) 
               go fr ((ConstantChain sub1 typ1 name1 z1 s1,ConstantChain sub2 typ2 name2 z2 s2):rest) scope@(scope1,scope2) = 
                  let 
                    (ok',fr',scope',fsc)      = alphaRenameBoundVar fr scope z1 z2  
                    alpha1 = case sub1 of 
                               (Substitution ((SubstName (None (SCConstantChain sub1' typ1' name1')) a1):_) )
                                  -> if typ1 == typ1' && name1 == name1' then [a1] else error "in alphaEquiv"
                               (Substitution []) -> []
                    alpha2 = case sub2 of 
                               (Substitution ((SubstName (None (SCConstantChain sub2' typ2' name2')) a2):_) )
                                  -> if typ2 == typ2' && name2 == name2' then [a2] else error "in alphaEquiv"
                               (Substitution []) -> []
                    (okR,frR,scR,fscR) =
                       case () of 
                         _ | (null alpha1 && null alpha2)  -> (True,fr',scope',fsc)
                           |  null alpha1 || null alpha2   -> (False,fr',scope',fsc)
                           | otherwise ->
                                let 
                                  (scope1',scope2') = scope'
                                  (newAlpha:fr'')= fr'
                                  alpha1' = head alpha1
                                  alpha2' = head alpha2
                                  scope'' = (Map.insert alpha1' newAlpha scope1',Map.insert alpha2' newAlpha scope2')
                                  (m1,m2) = fsc
                                in (ok'&&typ1 == typ2&&name1 == name2,fr'',scope'',(insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))
                    (okRest,frRest,scopeRest,fscRest) = (go frR  rest scR)
                  in (okR && okRest, frRest, scopeRest, unionListOfFullScopes [fscR,fscRest])
              in go fresh2 aCC scope2  
                
           (ok4,fresh4,scope4,fullScope4) =
             let 
               go fr [] scope = (True,fr,scope,emptyFullRenaming) 
               go fr ((Env sub1 name1,Env sub2 name2):rest) scope@(scope1,scope2) = 
                  let 
                    alpha1 = case sub1 of 
                                          (Substitution ((SubstName (None (SCMetaEnv (Env sub1' name1'))) a1):_) )
                                             -> if  name1 == name1' then [a1] else error "in alphaEquiv"
                                          (Substitution []) -> []
                    alpha2 = case sub2 of 
                                          (Substitution ((SubstName (None (SCMetaEnv (Env sub2' name2'))) a2):_) )
                                             -> if  name2 == name2' then [a2] else error "in alphaEquiv"
                                          (Substitution []) -> []
                    (okR,frR,scR,fscR) =
                       case () of 
                         _ | (null alpha1 && null alpha2)  -> (True,fr,scope,emptyFullRenaming)
                           |  null alpha1 || null alpha2   -> (False,fr,scope,emptyFullRenaming)
                           | otherwise ->
                                let 
                                  (newAlpha:fr')= fr
                                  alpha1' = head alpha1
                                  alpha2' = head alpha2
                                  scope' = (Map.insert alpha1' newAlpha scope1,Map.insert alpha2' newAlpha scope2)
                                  (m1,m2) = emptyFullRenaming
                                in (name1 == name2,fr',scope',(insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))
                    (okRest,frRest,scopeRest,fscRest) = (go frR  rest scR)
                  in (okR && okRest, frRest, scopeRest, unionListOfFullScopes [fscR,fscRest])
              in  (go fresh3 aME scope3)  
              
                
           (ok5,fresh5,scope5,fullScope5) =
             let 
               go fr [] scope =(True,fr,scope,emptyFullRenaming) 
               go fr ((Env sub1 name1,Env sub2 name2):rest) scope@(scope1,scope2) = 
                  let 
                    alpha1 = case sub1 of 
                                          (Substitution ((SubstName (None (SCConstantEnv (Env sub1' name1'))) a1):_) )
                                             -> if  name1 == name1' then [a1] else error "in alphaEquiv"
                                          (Substitution []) -> []
                    alpha2 = case sub2 of 
                                          (Substitution ((SubstName (None (SCConstantEnv (Env sub2' name2'))) a2):_) )
                                             -> if  name2 == name2' then [a2] else error "in alphaEquiv"
                                          (Substitution []) -> []
                    (okR,frR,scR,fscR) =
                       case () of 
                         _ | (null alpha1 && null alpha2)  -> (True,fr,scope,emptyFullRenaming)
                           |  null alpha1 || null alpha2   -> (False,fr,scope,emptyFullRenaming)
                           | otherwise ->
                                let 
                                  (newAlpha:fr')= fr
                                  alpha1' = head alpha1
                                  alpha2' = head alpha2
                                  scope' = (Map.insert alpha1' newAlpha scope1,Map.insert alpha2' newAlpha scope2)
                                  (m1,m2) = emptyFullRenaming
                                in (name1 == name2,fr',scope',(insertIntoMultiSet alpha1' newAlpha m1,insertIntoMultiSet alpha2' newAlpha m2))
                    (okRest,frRest,scopeRest,fscRest) = (go frR rest scR)
                  in (okR && okRest, frRest, scopeRest, unionListOfFullScopes [fscR,fscRest])
              in go fresh4   aCE scope4
         in (ok1 && ok2 && ok3 && ok4 && ok5,fresh5,scope5,unionListOfFullScopes [fullScope1,fullScope2,fullScope3,fullScope4,fullScope5],aBS,aMC,aCC,aME,aCE)
                
     |  (aBS,aMC,aCC,aME,aCE) <- allAssignments  
     ]
    validAllChecksForBinders = [(freshNew,scopeNew,fullScopeNew,aBS,aMC,aCC,aME,aCE) | (ok,freshNew,scopeNew,fullScopeNew,aBS,aMC,aCC,aME,aCE) <- allChecksForBinders, ok]
    checkRHSAndInExprOfValids =
     [
      let (okRes1,freshRes1,fullScopeNew1) = 
              let       
                go fr []                     = (True,fr,emptyFullRenaming)
                go fr ((x1:=e1,x2:=e2):rest) =
                     let (ok,fr',fsc)    = alphaEquivWithScopeExpr fr scopeNew e1 e2
                         (ok',fr'',fsc') = (go fr' rest)
                      in (ok && ok',fr'', unionListOfFullScopes [fsc,fsc'])
              in go freshNew aBS                      
             
          (okRes2,freshRes2,fullScopeNew2) =     
              let       
                go fr []     = (True,fr,emptyFullRenaming)
                go fr ((MetaChain sub1 typ1 name1 z1 s1,MetaChain sub2 typ2 name2 z2 s2):rest) =
                     let ok1 = alphaEquivWithScopeVar   scopeNew z1 z2
                         ok2 = alphaEquivWithScopeSubst scopeNew sub1 sub2
                         (ok3,fr',fsc) = alphaEquivWithScopeExpr fr scopeNew s1 s2
                         (ok4,fr'',fsc') = go fr' rest
                      in (ok1 && ok2 && ok3 && ok4,fr'', unionListOfFullScopes [fsc,fsc'])
              in go freshRes1 aMC 
              
          (okRes3,freshRes3,fullScopeNew3) =     
              let       
                go fr []     = (True,fr,emptyFullRenaming)
                go fr ((ConstantChain sub1 typ1 name1 z1 s1,ConstantChain sub2 typ2 name2 z2 s2):rest) =
                     let ok1 = alphaEquivWithScopeVar   scopeNew z1 z2
                         ok2 = alphaEquivWithScopeSubst scopeNew sub1 sub2
                         (ok3,fr',fsc) = alphaEquivWithScopeExpr fr scopeNew s1 s2
                         (ok4,fr'',fsc') = go fr' rest
                      in (ok1 && ok2 && ok3 && ok4,fr'', unionListOfFullScopes [fsc,fsc'])
              in go freshRes2 aCC 
          okRes4 =           
              let       
                go []     = True 
                go ((Env sub1 name1,Env sub2 name2):rest) =
                     let ok = alphaEquivWithScopeSubst scopeNew sub1 sub2
                         ok' = go rest 
                     in (ok && ok')
              in go aME 
              
          okRes5 =           
              let       
                go []     = True 
                go ((Env sub1 name1,Env sub2 name2):rest) =
                     let ok = alphaEquivWithScopeSubst scopeNew sub1 sub2
                         ok' = go rest 
                     in (ok && ok')
              in go aCE 
          (okRek,freshRes4,fullScopeNew4) = alphaEquivWithScopeExpr freshRes3 scopeNew inexpr1 inexpr2
      in
        (okRes1 && okRes2 && okRes3 && okRes4 && okRes5 && okRek
        ,freshRes4
        ,unionListOfFullScopes [fullScopeNew1,fullScopeNew2,fullScopeNew3,fullScopeNew4,fullScopeBefore]
        )
     |
       (freshNew,scopeNew,fullScopeBefore,aBS,aMC,aCC,aME,aCE) <- validAllChecksForBinders
     ]
    valid = [(o,f,ffsc) | (o,f,ffsc) <- checkRHSAndInExprOfValids, o]
 in
  case () of 
   _ | and [length bindings1 == length bindings2
            ,length mc1 == length mc2
            ,length cc1 == length cc2
            ,length me1 == length me2
            ,length ce1 == length ce2
            ] 
          -> if null valid then mytrace ("\n noVALID" 
                                   ++
                                   "\ncheckRHSAndInExprOfValids:\n"
                                   ++
                                   show [ok | (ok,_,_) <- checkRHSAndInExprOfValids]
                                   ++
                                   "\nallChecksForBinders:\n"
                                   ++
                                   show [ok5 | (ok5,fresh5,scope5,fsc,aBS,aMC,aCC,aME,aCE) <- allChecksForBinders]
                                 )
                                  (False,fresh,emptyFullRenaming) 
                       else head valid 
                             
     | otherwise -> mytrace "\nlength not ok" (False,fresh,emptyFullRenaming)
            
            
alphaEquivWithScopeExpr f sc e1 e2 = mytrace ("forgotten case?: " ++ show (e1,e2)) (False,f,emptyFullRenaming)            
            
unionsScopes xs = 
  let (a,b) = unzip xs                
  in (Map.unions a,Map.unions b)
    
    
-- alphaEquivWithScopeVar:    
-- check whether two variables are equivalent respecting the current scope
-- which represents the alpha renaming 
    
alphaEquivWithScopeVar :: ScopeRenaming 
                          -> VarLift 
                          -> VarLift 
                          -> Bool    
alphaEquivWithScopeVar  scope@(scope1,scope2)   (MetaVarLift sub1 v1) (MetaVarLift sub2 v2) 
    | Just k1 <- Map.lookup v1 scope1
    , Just k2 <- Map.lookup v2 scope2
    , k1 == k2
    = alphaEquivWithScopeSubst scope sub1 sub2
    | Nothing <- Map.lookup v1 scope1
    , Nothing <- Map.lookup v2 scope2
    , v1 == v2
    = alphaEquivWithScopeSubst scope sub1 sub2    
alphaEquivWithScopeVar scope@(scope1,scope2)   (ConstantVarLift sub1 v1) (ConstantVarLift sub2 v2) 
    | Just k1 <- Map.lookup v1 scope1
    , Just k2 <- Map.lookup v2 scope2
    , k1 == k2
    = alphaEquivWithScopeSubst scope sub1 sub2
    | Nothing <- Map.lookup v1 scope1
    , Nothing <- Map.lookup v2 scope2
    , v1 == v2
    = alphaEquivWithScopeSubst scope sub1 sub2    
alphaEquivWithScopeVar scope@(scope1,scope2)   (VarLift sub1 v1) (VarLift sub2 v2) 
    | Just k1 <- Map.lookup v1 scope1
    , Just k2 <- Map.lookup v2 scope2
    , k1 == k2
    = alphaEquivWithScopeSubst scope sub1 sub2
    | Nothing <- Map.lookup v1 scope1
    , Nothing <- Map.lookup v2 scope2
    , v1 == v2
    = alphaEquivWithScopeSubst  scope sub1 sub2    
alphaEquivWithScopeVar scope@(scope1,scope2)   x y = 
    mytrace ("\nInVarCase:" ++ show (scope,x,y)) False

  
    
alphaRenameBoundVar :: [String] -- fresh names 
                       -> ScopeRenaming -- current scope 
                       -> VarLift
                       -> VarLift 
                       -> (Bool
                          ,[String]
                          ,ScopeRenaming
                          ,FullRenaming
                          )                       

alphaRenameBoundVar fresh scope@(scope1,scope2)   v1@(MetaVarLift sub1 v1') v2@(MetaVarLift sub2 v2') =
    let (f:fresh')   = fresh
        scope'       = (Map.insert v1' f scope1,Map.insert v2' f scope2)
        (m1,m2)       = emptyFullRenaming
        fullRenaming  = (insertIntoMultiSet v1' f m1,insertIntoMultiSet v2' f m2)
    in  (alphaEquivWithScopeSubst scope' sub1 sub2, fresh', scope',fullRenaming)
    
alphaRenameBoundVar fresh scope@(scope1,scope2)   v1@(VarLift sub1 v1') v2@(VarLift sub2 v2') =
    let (f:fresh')   = fresh
        scope'       = (Map.insert v1' f scope1,Map.insert v2' f scope2)
        (m1,m2)       = emptyFullRenaming
        fullRenaming  = (insertIntoMultiSet v1' f m1,insertIntoMultiSet v2' f m2)
    in  (alphaEquivWithScopeSubst scope' sub1 sub2, fresh', scope',fullRenaming)
    
alphaRenameBoundVar fresh scope@(scope1,scope2)   v1@(ConstantVarLift sub1 v1') v2@(ConstantVarLift sub2 v2') =
    let (f:fresh')   = fresh
        scope'       = (Map.insert v1' f scope1,Map.insert v2' f scope2)
        (m1,m2)       = emptyFullRenaming
        fullRenaming  = (insertIntoMultiSet v1' f m1,insertIntoMultiSet v2' f m2)
    in  (alphaEquivWithScopeSubst scope' sub1 sub2, fresh', scope',fullRenaming)

--     
    
alphaEquivWithScopeArgPos fresh scope (VarPos v1) (VarPos v2) =
    (alphaEquivWithScopeVar scope v1 v2,fresh,emptyFullRenaming)

alphaEquivWithScopeArgPos fresh scope@(scope1,scope2) (Binder vs1 e1) (Binder vs2 e2) =     
 let
    go fresh scope [] [] = (True,fresh,scope,emptyFullRenaming)
    go fresh scope (v1:vs1) (v2:vs2) = 
        let
            (b,fresh',scope',fsc) = alphaRenameBoundVar fresh scope  v1 v2
            (b',fresh'',scope'',fsc') = go fresh' scope'  vs1 vs2
        in (b && b',fresh'',scope'',unionListOfFullScopes [fsc,fsc'])
    (b1,fresh1,scope1,fullScope) = go fresh scope vs1 vs2
    (b2,fresh2,fullScope2)       = alphaEquivWithScopeExpr fresh1 scope1 e1 e2
 in (b1 && b2, fresh2, unionListOfFullScopes [fullScope,fullScope2])
       
-- alphaEquivWithScopeSubst:       
--   checks if the current scope makes two renaming substitutions similar
--   no fresh names, and no adjustment of the scope needed
alphaEquivWithScopeSubst :: ScopeRenaming
                            -> Substitution 
                            -> Substitution 
                            -> Bool 
                            
alphaEquivWithScopeSubst scope@(scope1,scope2) (Substitution sub1) (Substitution sub2) =
   let sub1' = applyScope scope1 sub1
       sub2' = applyScope scope2 sub2
   in similarListsOfListItems sub1' sub2'  

applyScope scope [] = [] 
applyScope scope ((SubstVar (MetaVarLift _ v1) (MetaVarLift _ v2)):sub) =
    let 
        v1' = case Map.lookup v1 scope of
                Just l -> l
                Nothing -> v1
        v2' = case Map.lookup v2 scope of
                Just l -> l
                Nothing -> v2
    in (PairItem (v1',v2')):(applyScope scope sub)
    
applyScope scope ((SubstVar (ConstantVarLift _ v1) (ConstantVarLift _ v2)):sub) =
    let 
        v1' = case Map.lookup v1 scope of
                Just l -> l
                Nothing -> v1
        v2' = case Map.lookup v2 scope of
                Just l -> l
                Nothing -> v2
    in (PairItem (v1',v2')):((applyScope scope sub))

applyScope scope ((SubstVar (VarLift _ v1) (VarLift _ v2)):sub) =
    let 
        v1' = case Map.lookup v1 scope of
                Just l -> l
                Nothing -> v1
        v2' = case Map.lookup v2 scope of
                Just l -> l
                Nothing -> v2
    in (PairItem (v1',v2')):((applyScope scope sub))

applyScope scope ((SubstName (None s1) n1):sub) =
    let 
        n1' = case Map.lookup n1 scope of
                Just l -> l
                Nothing -> n1
    in (PairItem (getNSC s1,n1')):((applyScope scope sub))
    

applyScope scope ((SubstName (CapVars s1) n1):sub) =
    let 
        n1' = case Map.lookup n1 scope of
                Just l -> l
                Nothing -> n1
    in (PairItem (getNSC s1,n1')):((applyScope scope sub))


applyScope scope ((SubstName (LetVars s1) n1):sub)=
    let 
        n1' = case Map.lookup n1 scope of
                Just l -> l
                Nothing -> n1
    in (PairItem (getNSC s1,n1')):((applyScope scope sub))    

applyScope scope ((SubstSet set):sub)=
    let set' = Set.fromList $ applyScope scope set
    in (SetItem set'):(applyScope scope sub)
    
getNSC (SCExpr (MetaExpression _ n1)) = n1 
getNSC (SCExpr (Constant _ n1)) = n1
getNSC (SCCtxt (MetaCtxt sub1 srt1 n1)) = n1
getNSC (SCCtxt (ConstantCtxt sub1 srt1 n1)) = n1
getNSC (SCMetaEnv (Env _ e1)) = e1
getNSC (SCConstantEnv (Env _ e1))  = e1
getNSC (SCMetaChain sub1 typ1 name1)  = name1
getNSC (SCConstantChain sub1 typ1 name1) = name1



toItemList [] = [] 
toItemList ((SubstVar (MetaVarLift _ v1) (MetaVarLift _ v2)):sub) =
    (PairItem (v1,v2)):(toItemList sub)
    
toItemList ((SubstVar (ConstantVarLift _ v1) (ConstantVarLift _ v2)):sub) =
    (PairItem (v1,v2)):(toItemList sub)

toItemList ((SubstVar (VarLift _ v1) (VarLift _ v2)):sub) =
    (PairItem (v1,v2)):(toItemList sub)

toItemList ((SubstName (None s1) n1):sub) =
    (PairItem (getNSC s1,n1)):((toItemList sub))
    

toItemList ((SubstName (CapVars s1) n1):sub) =
    (PairItem (getNSC s1,n1)):((toItemList sub))


toItemList ((SubstName (LetVars s1) n1):sub)=
    (PairItem (getNSC s1,n1)):((toItemList sub))

toItemList ((SubstSet set):sub)=
    (SetItem $  Set.fromList $ toItemList set):(toItemList sub)

 
data ListItem a = PairItem a | SetItem (Set.Set (ListItem a))
 deriving (Ord,Eq,Show)

-- similarListsOfListItems (x:xs) (y:ys)
--     | trace (show (x,y)) False = undefined
similarListsOfListItems [] [] = True

similarListsOfListItems ((PairItem a):xs) ((PairItem b):ys)
    | a == b = similarListsOfListItems xs ys
    | otherwise = False 
    
similarListsOfListItems ((PairItem a):xs) ((SetItem set):ys)
    | Set.null set = similarListsOfListItems ((PairItem a):xs) ys
    | Set.member (PairItem a) set =
         similarListsOfListItems xs ((SetItem $ Set.delete (PairItem a) set):ys)
similarListsOfListItems ((SetItem set):xs) ((PairItem a):ys)
    | Set.null set = similarListsOfListItems xs ((PairItem a):ys)
    | Set.member (PairItem a) set =
         similarListsOfListItems ((SetItem $ Set.delete (PairItem a) set):xs) ys
         
similarListsOfListItems ((SetItem set1):xs) ((SetItem set2):ys)
    | Set.null set1 = similarListsOfListItems xs ((SetItem set2):ys)
    | Set.null set2 = similarListsOfListItems ((SetItem set1):xs) ys
    | otherwise = 
        let common = Set.intersection set1 set2
        in if Set.null common then False
                              else 
                               let set1' = set1 Set.\\ common
                                   set2' = set2 Set.\\ common
                               in similarListsOfListItems ((SetItem set1'):xs) ((SetItem set2'):ys)
similarListsOfListItems [] ((SetItem set2):ys)
    | Set.null set2 = similarListsOfListItems [] ys
similarListsOfListItems ((SetItem set1):xs) []
    | Set.null set1 = similarListsOfListItems xs []
     
similarListsOfListItems _ _ = False     
-- similarListsOfListItems _ _ = trace ("FALSE LAST") $ trace "\n" False
