-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Matching.Matching
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- The matching algorithm for lambda-expressions with recursive bindings
--
-----------------------------------------------------------------------------

module LRSX.Matching.Matching where
import Data.Function
import Data.Maybe
import Data.List
--
import Debug.Trace
--
import LRSX.Language.Syntax
import LRSX.Util.Util
import LRSX.Unification.UnificationProblem
import LRSX.Matching.ConstraintInclusion 
import LRSX.Language.Alpha (simsubst)

-- ====================================================================================================================
-- Matching
-- ====================================================================================================================

-- assumption: Variables are on the left-hand side, all variables on the right hand side are treated as constants
-- constantDelta contains the constraints for the right hand side
-- and Delta contains the constraints for the left hand side


-- ********************************************************************************************************************
-- The matching function performs the whole matching and builds a tableau
-- ********************************************************************************************************************

matching :: Problem -> Tableau

#ifdef DEBUG
matching problem 
 | trace (("\n\nMATCHING STARTS WITH PROBLEM\n" ++  show (gamma problem) ++ "\n")) False = undefined
#endif   

matching problem = 
  -- ---------------------------------------------------------------------------------------------------------------------------------------
  -- Initial checks on the input problem
  --
  --   check whether the weak DVC holds for the input equations
  case checkWeakDVCGamma (gamma problem) of 
    Nothing ->  
       -- check the occurence restrictions for the left hand sides of the equations: we use the same function as for 
       -- the unification algorithm, but replace the right hand sides of the equations by a constant (since only the lhs have to be checked)
       case check problem{gamma=let gam  = (gamma problem)
                                    env' = [l :=?=: empty |  l :=?=: r <- environmentEq gam]
                                    ex'  = [l :=?=: Fn "__REPLACED_RHS" [] | l :=?=: r <- expressionEq gam]
                                    var' = [l :=?=: ConstantVarLift emptySubstitution "_X" | l :=?=: r <- varLiftEq gam]
                                in gam{environmentEq=env',expressionEq=ex',varLiftEq=var'}
                         }  of 
                  Nothing -> expand [1] problem
                  Just err -> Node problem [Fail $ "Occurence restrictions in input:" ++ err]
    Just err ->  Node problem [Fail $ "Weak DVC failure in input:" ++ err]
  -- ---------------------------------------------------------------------------------------------------------------------------------------    
 where
  -- the expand function expands the problem until it is solved or Fails. 
  -- the parameter i is used to number the steps (only for nicer output of the derivation tree)
  expand i pb  
   -- problem is solved (gamma is empty) and no further step is applicable
   | solved (gamma pb)             
   , Empty <- matchStep  heur (pb) = 
       let -- compose the solution
           sol= composeAll (solution pb)
           -- apply the matcher to the equations
           instExpr = applySolution sol (gamma problem)
           -- apply the solution to the NCCs in delta3
           instDelta = applySolution sol (delta3 $ delta pb)    
           -- check the noncapture constraints 
           capcon  = checkNCCs (delta2 $ delta pb) instDelta  
           -- 
           nccDVC = splitFlatToNCC (concatMap getNCCsEnvDVC $ concatMap  getEnvironments (concat [ [s,t] | s :=?=: t <- expressionEq instExpr]))
                 -- wrong     ++  concat [getNCCsExpr s ++ getNCCsExpr t | s :=?=: t <- expressionEq instExpr ]
           
--           redVars = nub $ concat [metaVars s ++ metaVars t| s :=?=: t <- expressionEq instExpr ]
--           (groundInst,fresh') = mkStandardSubst (fresh pb) (constantDelta pb) redVars
--          groundInstExpr = applySolution groundInst instExpr
--          checkRhoPrime mapping = checkWeakDVCGamma (applySolution (mapping) instExpr)
           
           -- check whether the given NCC (or the alphaBase) implies the NCCs in Delta3
           (str,capConImplication) = nccImplies (delta2 $ delta pb) (nccDVC ++ (delta3 $ constantDelta  pb))  instDelta (alphaBase problem) 
          -- check the Weak DVC of the instantiated expressions
       in 
           case checkWeakDVCGamma instExpr of     
             Nothing -> 
               -- check whether the capture constraints are implied by cdelta3
                case capConImplication of
                  Nothing -> -- check non-capture constraints for Delta3
                     case capcon of 
                        Nothing -> -- solution found
                                   -- trace str $                        
                                             Solved (pb{solution=sol
                                                    ,constantDelta=(constantDelta pb){delta3=( (delta3 (constantDelta pb)) ++ (nccDVC ) )}
                                                    ,mpLogging = str ++
                                                                        ("\nMATCHINGLOG : START MATCHING" ++
                                                                          "\nMATCHINGLOG : Matching ended successfully" ++
                                                                          "\nMATCHINGLOG :  Gamma:" ++ show (gamma problem) ++
                                                                          "\nMATCHINGLOG :  Delta:" ++ show (delta problem) ++
                                                                          "\nMATCHINGLOG :  CDelta:" ++ show (constantDelta problem) ++
                                                                          "\nMATCHINGLOG :  OUTPUT:" ++
                                                                          "\nMATCHINGLOG :  INSTEXPR:" ++ show instExpr ++
                                                                          "\nMATCHINGLOG :  Gamma:" ++ show (gamma pb) ++
                                                                          "\nMATCHINGLOG :  Delta:" ++ show (delta pb) ++
                                                                          "\nMATCHINGLOG :  CDelta:" ++ show (constantDelta pb) ++
                                                                          "\nMATCHINGLOG :  NCCs:      " ++ showSplitted (splitCapCon (delta3 (delta pb))) ++
                                                                          "\nMATCHINGLOG :  IMPLIED by " ++ showSplitted (splitCapCon (delta3 (constantDelta pb))) ++
                                                                          "\nMATCHINGLOG : STOP MATCHING\n"
                                                                          )
                                                    }
                                                 )
                        Just err -> -- capture constraints in Delta3 are violated 
#ifdef DEBUG
                                     trace ("Matching fails for capture constraint fail")
#endif                        
                                     error ( "does this ever happen???" -- (Node pb [Fail $ "capture constraint fail" ++ (show err)])
                                      ++ (err))
                                                 
                  Just err -> -- capture constraints are not implied 
#ifdef DEBUG
                                                    trace ("\nMATCHING:" ++
                                                           "\n Gamma:" ++ show (gamma problem) ++
                                                           "\n Delta:" ++ show (delta problem) ++
                                                           "\n CDelta:" ++ show (constantDelta problem) ++
                                                           "\n OUTPUT:" ++
                                                           "\n INSTEXPR:" ++ show instExpr ++
                                                           "\n Gamma:" ++ show (gamma pb) ++
                                                           "\n Delta:" ++ show (delta pb) ++
                                                           "\n CDelta:" ++ show (constantDelta pb)++
                                                           "\n NCCs:      " ++ showSplitted (splitCapCon (delta3 (delta pb))) ++
                                                           "\n NOT IMPLIED by " ++ showSplitted (splitCapCon (delta3 (constantDelta pb))) ++
                                                           "\n INCLUSION ENGINE START\n"      ++ str ++
                                                           "\n INCLUSION ENGINE STOP\n" 
                                                          )
#endif                                                   
                                                         (Node pb [Fail $ "capture constraint implication fails: " ++ err])
             Just err -> -- weak DVC does not hold
#ifdef DEBUG
                         trace ("Matching fails for WEAKDVC")
#endif                 
                         Node pb [Fail $ "Weak DVC failure " ++ err ++ " " ++ (show (expressionEq instExpr))]
  expand (i:is) pb =
   -- problem is not solved 
   case matchStep heur pb of 
     Some problems -> case problems of
                        xs ->  Node pb (map (\(j,p) -> expand (j:i:is) p) (zip [1..] xs))
     MatchFail err  -> Node pb [Fail err]
     Empty         -> error $ "in expand " ++ (take 1000 $ show (pb)) -- show $ Node (solution pb, gamma pb, delta pb) []
     
  heur = [applyFailRules
         ,applyFirstOrderRules
         ,applyContextRules
         ,applyGuessRules
         ,applyEnvRules
         ]
     
  matchStep [] pb = Empty
  matchStep (matchfn:xs) pb =
    case matchfn pb of 
      Empty -> matchStep xs pb
      Some [] -> matchStep xs pb
      Some ps -> Some ps
      MatchFail err -> MatchFail err

-- ********************************************************************************************************************
-- The tableau of the matching algorithm

data Tableau = Node Problem [Tableau] 
             | Fail String
             | Solved Problem 

dt (Node _ xs) = 1 + (maximum $ map dt xs)
dt x = 1
sz (Node _ xs) = 1+ (sum $ map sz xs)
sz (Fail _) = 1
sz (Solved _) = 1
instance Show Tableau where
 show t = showTableau [1] t
  where
   showTableau ind (Node p xs) = 
     indent ind ++ "Open: "  ++ "\n"
     ++ "   Gamma:\n" 
     ++ "     " ++ show (gamma p) ++ "\n"
     ++ "   Sol:\n"
     ++ "     " ++ (intercalate "\n     " (map show (solution p))) ++ "\n" 
     ++ "   Delta:\n"
     ++ "     " ++ (show (delta p)) ++ "\n"
     ++ "   ConstanstDelta:\n"
     ++ "     " ++ (show (constantDelta p)) ++ "\n"
     ++ "\n" ++ (concat [showTableau (if length xs > 1 then ind ++ [i] else ind) x | (i,x) <- zip [1..] xs])
   showTableau ind (Fail a)   = indent ind ++ "FAIL: " ++  a ++ "\n"
   showTableau ind (Solved pb) = 
            let sol = solution pb
                gam = applySolution sol (inputGamma pb)
                del = delta pb
            in indent ind ++ "Solved: Unified Equations:" ++ show gam ++ "\n  Solution:" ++ show sol ++ "\n"
   indent ind = let px = (concat $ intersperse "." (map show ind)) 
                in  px ++ (replicate (mx - (length px)) ' '     ) ++ ": "
   mx = 10 -- length (show $ branchDepth [1] t) 
   branchDepth i (Node _ [x]) = branchDepth i x
   branchDepth i (Node _ xs) =  let a = [branchDepth (i ++ [v]) x  | (v,x) <- zip [1..] xs]
                                in if null a then i else maximumBy (compare `on` (length . show)) a   
   branchDepth i _ = i 
    
-- helper functions for the tableau:

-- compute solved and open leaves 
solvedAndOpenLeaves (Node a []) = [Node a []]
solvedAndOpenLeaves (Node _ cs) = concatMap solvedAndOpenLeaves cs
solvedAndOpenLeaves (Fail x)    = []
solvedAndOpenLeaves (Solved x)  = [Solved x] 

-- compute the solved leaves only    
solvedLeaves (Node _ cs) = concatMap solvedLeaves cs
solvedLeaves (Fail _)    = []
solvedLeaves (Solved x)  = [x] 

-- compute all leaves 
allLeaves (Node a []) = [Node a []]
allLeaves (Node _ cs) = concatMap allLeaves cs
allLeaves (Fail x)    = [Fail x]
allLeaves (Solved x)  = [Solved x] 
-- ********************************************************************************************************************
   
   

-- ********************************************************************************************************************
-- Result of one step of the matching algorithm 
 
data MatchResult = MatchFail String  -- matching fails
                  | Empty            -- no matching rule applicable
                  | Some [Problem]   -- new problems after one step
 deriving(Show)
-- ********************************************************************************************************************

 
 
-- ********************************************************************************************************************
-- Fail rules
-- ********************************************************************************************************************
applyFailRules :: Problem -> MatchResult
applyFailRules problem@(Problem{solution=sol,gamma=gam,delta=del,constantDelta=cdel})
-- **********************************
-- Fail for variable equations:
  -- \n{x} =?= \n{y}  or \n{x}alpha = \n{x}alpha' and alpha /= alpha'
  |  (eq:_) <- [eq | eq@(VarLift substx x :=?=: VarLift substy y) <- (varLiftEq gam), x /= y || not (substx `simsubst` substy)]
    =  MatchFail $ show eq 
  -- \n{x} =?= Y* 
  |  (eq:_) <- [eq | eq@((VarLift substx x) :=?=: (ConstantVarLift  substy y)) <-  (varLiftEq gam)] 
  =  MatchFail $ show eq 
  -- X* =?= n{x}
  |  (eq:_) <- [eq | eq@((ConstantVarLift  substx x) :=?=: (VarLift  substy y)) <-  (varLiftEq gam)]
    =  MatchFail $ show eq 
  -- X* =?= Y*  or X*alpha =?= X*alpha' and alpha /= alpha'
  |  (eq:_) <- [eq | eq@((ConstantVarLift  substx x) :=?=: (ConstantVarLift  substy y)) <-  (varLiftEq  gam), x /= y || not (substx `simsubst` substy)] 
    =  MatchFail $ show eq 
-- **********************************
-- Fail for expression equations
  -- var x =?= t  
  --  with t = f t1 ... tn or t = letrec ... or t = S*  or t = D*[t']
  | (eq:_) <- [eq | eq@(VarL _ :=?=: rhs) <- (expressionEq gam)
                  , case rhs of 
                        (Fn _ _) -> True
                        (Letrec _ _) -> True
                        (Constant _ _) -> True
                        (ConstantCtxtSubst _ _ _ _) -> True
                        _ -> False
               ] = MatchFail $ show eq
               
               
  -- f t1 ... tn = t 
  --  with t = f' s1 .. sn, or t = letrec ..., or t = S*, or t= var y, or t=D*[s]  
  | (eq:_) <- [eq | eq@(Fn f args :=?=: rhs) <- (expressionEq gam)
                  , case rhs of 
                        (Fn f' _) -> f /= f' 
                        (Letrec _ _) -> True
                        (Constant _ _) -> True
                        (VarL _) -> True
                        (ConstantCtxtSubst _ _ _ _) -> True
                        _ -> False
               ] = MatchFail $ show eq

  -- letrec ... = t 
  --  with t = f s1 .. sn, or t = S*, or t= var y, or t=D*[s]  
  | (eq:_) <- [eq | eq@(Letrec _ _ :=?=: rhs) <- (expressionEq gam)
                  , case rhs of 
                        (Fn _ _) -> True
                        (Constant _ _) -> True
                        (VarL _) -> True
                        (ConstantCtxtSubst _ _ _ _) -> True
                        _ -> False
               ] = MatchFail $ show eq
  -- S* = t 
  --  with t = f s1 .. sn, t=letrec..., or t = S'*, or t= var y, or t=D*[s]  
  -- or S* alpha =?= S* alpha' with alpha /= alpha'
  | (eq:_) <- [eq | eq@(Constant sub name :=?=: rhs) <- (expressionEq gam)
                  , case rhs of 
                        (Fn _ _) -> True
                        (Constant sub1 name1) -> name /= name1 || (not $ sub `simsubst` sub1)
                        (Letrec _ _) -> True
                        (VarL _) -> True
                        (ConstantCtxtSubst _ _ _ _) -> True
                        _ -> False
               ] = MatchFail $ show eq
  -- D*[s] =?= t and t /= D*[s']  
  | (eq:_) <- [eq | eq@(ConstantCtxtSubst sub srt name s :=?=: rhs) <- (expressionEq gam)
                  , case rhs of 
                        (ConstantCtxtSubst sub1 srt1 name1 s1) -> not (srt1 == srt && sub `simsubst` sub1 && name == name1)
                        _ -> True -- Fail in any other case 
               ] = MatchFail $ show eq
               

   -- D[s] =?= t and D in Delta1 and
   --   t = letrec .... and sort(D) = A
   --   t = var x
   --   t = f with ar(f)=0
   --   t = S*
   --   t = f s1 ... sn and oari(f)(i) = V for all i
   --   t = f s1 ... sn and oari(f) = (m1,...,mn) and all mi /= 0 and srt(D) < C
   --   t = D1*[t'] and D1* not in CDelta1 or srt(D1) > srt(D)
  | (eq:_) <- [eq | eq@(MetaCtxtSubst sub srt name s :=?=: rhs) <- (expressionEq gam)
                  , ((MetaCtxt sub srt name) `elem` (delta1 del) || (MetaCtxt emptySubstitution srt name) `elem` (delta1 del))
                  , case rhs of 
                        (Letrec _ _) ->  srt == ACtxt
                        (VarL _)     ->  True
                        (Fn f [])    ->  True
                        (Constant _ _) -> True
                        (Fn f args) -> let bari:_ = [bindarity | (name,arity,strictpos,bindarity) <- functions problem, f == name]
                                           res1 = (all (== (-1)) bari && not (null bari))
                                           res2 = srt < Ctxt && all (/= 0) bari && not (null bari)
                                       in res1 || res2
                        (ConstantCtxtSubst sub1 srt1 name1 s1) ->  not ((ConstantCtxt sub1 srt1 name1 `elem` (delta1 cdel))
                                                                        || (ConstantCtxt emptySubstitution srt1 name1 `elem` (delta1 cdel)))
                                                                ||  srt1 > srt  
                        _ -> False
               ]
     = MatchFail $ show eq
  --   D[s] =?= = f s1 ... sn , SPOS(f) = emptyset, srt(D)=A, D in delta1
  | Just eq@((ctxt@(MetaCtxtSubst sub ACtxt cname s) :=?=: (Fn fname arguments)),rest)
           <- splitBy (\(l:=?=:r) -> case (l,r) of 
                                      (MetaCtxtSubst sub ACtxt cname s, Fn fname _) ->  
                                        ((MetaCtxt sub ACtxt cname) `elem` (delta1 del) || (MetaCtxt emptySubstitution ACtxt cname) `elem` (delta1 del))
                                      _ -> False) (expressionEq gam)
  ,Just [] <- (getStrictPos problem fname)  
   = MatchFail $ show eq
-- ******************************
-- Rules for environment equations
-- Fail: env = {} and env is not empty
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
               ,isEmpty right 
               ,  (not (null (bindings left )) 
                  || not (null (metaChains left)) 
                  || not (null (constantChains left))
                  || not (null [v | v <- constantEnvVars left])
                  || not (null [ v | v <- metaEnvVars left, v `elem` (delta2 del)]))]
    = MatchFail $ show eq
-- Fail: {} = env and env is not empty
--
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , isEmpty left 
                   , not (null (bindings right )) 
                     || not (null (constantChains right))
                     || not (null [v | v <- constantEnvVars right])
                ]
       = MatchFail $ show eq
-- Fail: b;env = env' and env' has no bindings
--
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , not (null (bindings left))
                   , null (bindings right)
                ]
       = MatchFail $ show eq
-- Fail: EE^K1;env = env' and env' has no bindings, and env' has no EE'^K2* with  sort K1 >= sort K2
--
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , not (null (metaChains left))
                   , null (bindings right)
                   , or [ null  [ () |  ConstantChain sub1 typ1 name1 z1 s1 <- constantChains right
                              , typ1 == typ || typ == AChain ]
                     | (MetaChain sub typ name z s) <- metaChains left]
                ]
       = MatchFail $ show eq
-- Fail: env = EE*^K2 and env /= E;env' and env /= EE'^K1 with   sort K1 >= sort K2
--
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , not (null (constantChains right))
                   , null (metaEnvVars left)
                   , or [null [() | MetaChain sub1 typ1 name1 z1 s1 <- metaChains left
                                  , typ1 == typ || typ == VarChain
                                  ]
                         && null [() | (ConstantChain sub1 typ1 name1 z1 s1) <- constantChains left, typ1 == typ, name1 == name, sub1 `simsubst` sub]
                        | (ConstantChain sub typ name z s) <- constantChains right
                        ]
                     
                ]
       = MatchFail $ show eq
--  EE*[z,s];env =?= env' and env' /= EE*[z',s'];env''       
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , not (null (constantChains left))
                   , or [null [() | ConstantChain sub1 typ1 name1 z1 s1 <- constantChains right
                                  , name == name1]
                        | (ConstantChain sub typ name z s) <- constantChains left
                        ]
                ]
       = MatchFail $ show eq
--  E*;env =?= env' and env' /= E*;env''       
--  E*alpha1;env =?= env' and env' /= E*alpha2;env'' with alpha2 = alpha1
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , not (null (constantEnvVars left))
                   , or [null [() | (Env sub1 name1) <- constantEnvVars right
                                  , name == name1 && sub `simsubst` sub1]
                        | (Env sub name) <- constantEnvVars left
                        ]
                ]
       = MatchFail $ show eq
--  env =?= E*;env' and env /= E*;env'' and env /= E1;env''   
--  env =?= E*alpha1;env and env /= E*alpha2 with alpha1 /= alpha2 and env /= E1;env''
   | (eq:_) <- [eq | eq@(left :=?=: right) <- (environmentEq gam)
                   , not (null (constantEnvVars right))
                   , null (metaEnvVars left)
                   , or [null [() | (Env sub1 name1) <- constantEnvVars left
                                  , name == name1 && sub `simsubst` sub1]
                        | (Env sub name) <- constantEnvVars right
                        ]
                ]
       = MatchFail $ show eq
--  E1;....;En =?= E1*;....;Em* all Ei in Delta2, and all E*j are not in CDelta2
  | (eq@(left :=?=: right,rest):_) 
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (hasBinding left) && not (hasBinding right)
         , null (metaChains left)
         , null (constantChains left)
         , not (null (metaEnvVars left))
         , not (null (constantEnvVars right))
         , null (constantEnvVars left)
         , all (\(v@(Env sub e1)) -> v `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del)) (metaEnvVars left)
         , all (\(v@(Env sub e1)) -> not ((v `elem` (delta2 cdel)) || ((Env emptySubstitution e1) `elem` (delta2 cdel)))) (constantEnvVars right)
         ]
   = MatchFail $ "all envvars in rhs are more general than envvars on right" ++ (show eq)
 | otherwise = Empty


 
-- ********************************************************************************************************************
-- First-Order Cases
-- ********************************************************************************************************************
applyFirstOrderRules :: Problem -> MatchResult
applyFirstOrderRules problem@(Problem{solution=sol,gamma=gam,delta=del,constantDelta=cdel})
-- 
-- RULE: Sol,Gamma cup {X = Y}, Delta --> Sol circ {X |-> Y}, Gamma[Y/X], Delta [Y/X] and X = ...  and X has an empty substitution
--
 | Just ((varx@(MetaVarLift sub x) :=?=: vary),rest) 
     <-  splitBy (\(l:=?=:r) -> case l of 
                                 MetaVarLift s x -> True
                                 _                 -> False) (varLiftEq gam)
 , isEmptySubstitution sub                                 
    = let sol' = (MapVarLift $ varx :|-> vary):sol
          gam' = subst vary varx (gam{varLiftEq = rest}) 
          del' = subst vary varx del
      in Some [problem{solution=sol',gamma=gam',delta=del'}]
--
--  RULE: Sol,Gamma cup {x = x}, Delta --> Sol, Gamma, Delta and the same substitution is applied to both occurences of x
--
  |  Just (eq@((VarLift subx x) :=?=: (VarLift suby y)),rest) 
       <-  splitBy (\(l :=?=: r) -> case (l,r) of 
                                     (VarLift subx x,VarLift suby y) -> x == y && subx `simsubst` suby
                                     _                             -> False) (varLiftEq gam)
  , subx `simsubst` suby                                    
    = Some [problem{gamma=gam{varLiftEq = rest}}]
--
-- RULE: Sol,Gamma cup {x* = x*}, Delta --> Sol, Gamma, Delta and the same substitution is applied to both occurences of x*
--
  |  Just (eq@((ConstantVarLift subx x) :=?=: (ConstantVarLift suby y)),rest) 
       <-  splitBy (\(l :=?=: r) -> case (l,r) of 
                                     (ConstantVarLift subx x,ConstantVarLift suby y) -> x == y && subx `simsubst` suby
                                     _                             -> False) (varLiftEq gam)
  , subx `simsubst` suby                                     
    = Some [problem{gamma=gam{varLiftEq = rest}}]
    
-- 
--  RULE: Sol,Gamma cup {S = s}, Delta --> Sol circ {S |-> s}, Gamma[s/S], Delta [s/S] if S is not contained in s
--
 | Just ((vars@(MetaExpression sub s) :=?=: t),rest) 
     <- splitBy (\(l:=?=:r) -> case l of 
                                MetaExpression sub s -> if (properSubTermOf l r) then error "BUG!:MetaExpression occurs on rhs" else isEmptySubstitution sub
                                _                    -> False                      ) (expressionEq gam)
 , isEmptySubstitution sub                                
    = let sol' = (MapExpr $ vars :|-> t):sol
          gam' = subst t vars (gam{expressionEq = rest}) 
          del' = subst t vars  del
      in Some [problem{solution=sol',gamma=gam',delta=del'} ]
-- 
--  RULE: Sol,Gamma cup {S* = S*} ---> Sol,Gamma
--
 | Just ((vars@(Constant sub s) :=?=: t),rest) 
     <- splitBy (\(l:=?=:r) -> case (l,r) of 
                                (Constant sub1 n1,Constant sub2 n2) -> n1 == n2 && sub1 `simsubst` sub2
                                _                    -> False                      ) (expressionEq gam)
    = let gam' = (gam{expressionEq = rest}) 
      in Some [problem{gamma=gam'} ]
-- 
--  RULE: Sol,Gamma cup {f s1...sn = f t1..tn}, Delta --> Sol, Gamma cup {si=ti}, Delta
-- 
 | Just (((Fn name args) :=?=: Fn name' args'),rest) 
     <-   splitBy (\(l:=?=:r) -> case (l,r) of 
                                  (Fn name _, Fn name' _) -> name == name'
                                  _                       -> False         ) (expressionEq gam)
    = let sol' = sol
          (vs,es) = computeAll args args'
          gam' = (gam{expressionEq = es ++ rest, varLiftEq = vs ++ (varLiftEq gam)}) 
          del' = del
          compute (VarPos v1) (VarPos v2)
            = let vareq = v1 :=?=: v2
              in ([vareq],[])
          compute b1@(Binder v1 e1) b2@(Binder v2 e2)
            | length v1 == length v2 = let 
                                          vareq = zipWith (:=?=:) v1 v2
                                          expreq = [e1 :=?=: e2]
                                       in (vareq,expreq)
            | otherwise = error ("found binders with different arity: "  ++ show b1 ++ " and " ++ show b2)
          compute b1 b2 =
            error ("Binder found at variable position with: " ++ show b1 ++ " and " ++ show b2)            
          computeAll args1 args2 = 
            let (vareqs,expreqs) = unzip (zipWith compute args1 args2) -- (,),(,)
            in (concat vareqs,concat expreqs)
      in Some [problem{solution=sol',gamma=gam',delta=del'} ]
      
--  
--  RULE: Sol,Gamma cup {var X = var Y}, Delta --> Sol, Gamma cup {X=Y}, Delta 
--  (for all cases of X,Y (Meta or concrete)
-- 
  | Just ((VarL varx :=?=: VarL vary),rest) 
       <-  splitBy (\(l :=?=: r) -> case (l,r) of 
                                     (VarL _, VarL _) -> True
                                     _                    -> False) (expressionEq gam)
     = let sol' = sol
           gam' = gam{ varLiftEq = (varx:=?=:vary):(varLiftEq gam)
                      , expressionEq = rest} 
           del' = del
      in Some [problem{solution=sol',gamma=gam',delta=del'} ]
-- 
--  RULE: Sol,Gamma cup {x.s = y.t}, Delta --> Sol, Gamma cup {x=y,s=t}, Delta 
--  (for all cases of X,Y (Meta or concrete)
-- 
  | ((var1 := e1) :=?=: (var2 := e2):rest) <-  (bindingEq gam)
     = let sol' = sol
           gam' = gam{ varLiftEq = (var1:=?=:var2):(varLiftEq gam)
                      , expressionEq = (e1 :=?=: e2):(expressionEq gam)
                      , bindingEq = rest} 
           del' = del
      in Some [problem{solution=sol',gamma=gam',delta=del'} ]
-- 
--  RULE: Sol,Gamma cup {letrec env1 in s1 = letrec env2 in s2}, Delta --> Sol, Gamma cup {env1=env2,s1=s2}, Delta 
-- 
  | Just ((Letrec env1 s1 :=?=: Letrec env2 s2),rest) 
      <- splitBy (\(l:=?=:r) -> case (l,r) of 
                                 (Letrec _ _, Letrec _ _) -> True
                                 _                       -> False) (expressionEq gam)
     = let sol' = sol
           gam' = gam{  environmentEq = (env1:=?=:env2):(environmentEq gam)
                      , expressionEq = (s1 :=?=: s2):rest} 
           del' = del
      in Some [problem{solution=sol',gamma=gam',delta=del'}] 


--
-- Rule: Gamma cup C*subs[s] = C*subs[t] --> Gamma cup s = t
--

        
   | Just ((ConstantCtxtSubst sub1 srt1 n1 s1 :=?=: ConstantCtxtSubst sub2 srt2 n2 s2),rest) 
            <- splitBy (\(l :=?=: r) -> case (l,r) of
                                         (ConstantCtxtSubst sub1 srt1 n1 s1, ConstantCtxtSubst sub2 srt2 n2 s2) -> n1 == n2 && srt1 == srt2 && sub1 == sub2
                                         _                  -> False) (expressionEq gam)
     = let sol' = sol
           gam' = gam{expressionEq = (s1 :=?=: s2): rest} 
           del' = del
        in Some [problem{gamma=gam',delta=del'}]
        
-- Rule: Gamma cup {} = {} --> Gamma
   | Just (env1 :=?=: env2,rest) <- splitBy (\(l :=?=: r) -> isEmpty l && isEmpty r) (environmentEq gam)
     = let sol' = sol
           gam' = gam{environmentEq = rest}
        in Some [problem{gamma=gam'} ]
-- no first-order rule applicable:
  | otherwise = Empty
-- ********************************************************************************************************************

      
      

-- RULE: Gamma cup {D_subst[s] = t} --> Gamma cup s=t [[._subst]/D_subst]  |  Gamma cup D_subst[s]=t  Delta1 cup D_subst
-- only if t /= D[t] where t notin CDelta1       
applyGuessRules problem@(Problem{solution=sol,gamma=gam,delta=del,constantDelta=cdel})
 | Just (((MetaCtxtSubst sub srt name s) :=?=: t),_)
        <- splitBy 
            (\(l :=?=: r) -> case (l,r) of
                            (MetaCtxtSubst sub srt name s,r) -> case r of 
                                                                    ConstantCtxtSubst sub0 srt0 name0 t0 -> 
                                                                         (ConstantCtxt sub0 srt0 name0) `elem` (delta1 cdel)
                                                                      || (ConstantCtxt emptySubstitution srt0 name0) `elem` (delta1 cdel)
                                                                    other -> not ((MetaCtxt sub srt name) `elem` (delta1 del))
                            
                            _ -> False)
            (expressionEq gam)
     
    = let problem1 = 
             let oldDel_1 = delta1 del
                 newDel_1 = (MetaCtxt sub srt name):oldDel_1
             in  problem{delta = del{delta1 = newDel_1}}
          problem2 = 
             let sol' = (MapExpr $ MetaCtxt sub srt name :|-> (Hole sub) ):sol
                 gam' = subst (Hole sub) (MetaCtxt sub srt name) gam
                 del' = subst (Hole sub) (MetaCtxt sub srt name) del
             in problem{solution=sol',gamma=gam', delta=del'}
      in Some [problem1,problem2]
-- no guess rule applicable
 | otherwise = Empty
-- ********************************************************************************************************************
      
      
-- ********************************************************************************************************************
-- Rules for Contexts 
-- ********************************************************************************************************************
      
applyContextRules problem@(Problem{solution=sol,gamma=gam,delta=del,constantDelta=cdel})
-- D[s] = D'[s], sort(D) < sort(D') and D not in Delta1 ===> D must be emtpy

 | Just (((MetaCtxtSubst sub srt name s) :=?=: t),_)
        <- splitBy 
            (\(l :=?=: r) -> case (l,r) of
                            (MetaCtxtSubst sub srt name s,ConstantCtxtSubst sub0 srt0 name0 t) -> 
                                                           (not ((MetaCtxt sub srt name) `elem` (delta1 del)))
                                                             && srt0 > srt 
                                                             
                            _ -> False)
            (expressionEq gam)
    = let problem2 = 
             let sol' = (MapExpr $ MetaCtxt sub srt name :|-> (Hole sub) ):sol
                 gam' = subst (Hole sub) (MetaCtxt sub srt name) gam
                 del' = subst (Hole sub) (MetaCtxt sub srt name) del
             in problem{solution=sol',gamma=gam', delta=del'}
      in Some [problem2]
--
-- RULE A_subst[s] = f s1 ... sn
--  ===> {sol = sol cup {A_subst |- f s1 ... A'_subst ... sn}, si =?= A'[s]} for each strict position of f_i 
--
--
 | Just ((ctxt@(MetaCtxtSubst sub ACtxt cname s) :=?=: (Fn fname arguments)),rest)
     <- splitBy (\(l:=?=:r) -> case (l,r) of 
                                (MetaCtxtSubst sub ACtxt cname s, Fn fname _) ->  (MetaCtxt sub ACtxt cname) `elem` (delta1 del)
                                _ -> False) (expressionEq gam)
 , Just (positions@(_:_)) <- (getStrictPos problem fname)  
    = Some
      [let
       (name_A':fresh') = fresh problem
       exprAprimeS      = MetaCtxtSubst emptySubstitution ACtxt ("A"++name_A') s 
       contextA'        = MetaCtxt emptySubstitution ACtxt ("A"++name_A')
       contextA         = MetaCtxt sub ACtxt cname 
       gam'             = gam{expressionEq = (exprAprimeS :=?=: expr):rest}  
       del'             = subst rhs contextA del 
       rhs              = Fn fname (before ++ [Binder [] contextA'] ++ after)
       sol'             = (MapExpr $ contextA :|-> rhs):sol
      in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
       | (pos,(before,(Binder _ expr),after)) <- zip [1..] (splits2 arguments)
        , pos `elem` positions
      ]
      
--
-- D[s] = f s1 .. sn , D of sort T
--
 | Just ((ctxt@(MetaCtxtSubst sub TCtxt cname s) :=?=: (Fn fname (args))),rest)
     <- splitBy 
         (\(l:=?=:r) -> case (l,r) of
                          (MetaCtxtSubst sub srt cname s, Fn fname args) -> (MetaCtxt sub srt cname) `elem` (delta1 del)
                                                                            && srt `elem` [TCtxt]
                          _ -> False)                                    
         (expressionEq gam)
    = Some
      [let
         (name_d':fresh') = fresh problem
         d'_s = MetaCtxtSubst emptySubstitution TCtxt  ("T"++name_d') s 
         d' = MetaCtxt emptySubstitution TCtxt ("T"++name_d')  
         d = MetaCtxt sub TCtxt cname
         gam' = gam{expressionEq = (d'_s :=?=: si):rest}
         del' = subst rhs d del 
         rhs = Fn fname (before ++ [Binder [] d'] ++ after)
         sol' = (MapExpr $ d :|-> rhs):sol
       in 
         problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
       | (before,(Binder v si),after) <- splits2 args, null v -- only positions without binders!
       ]
--
-- D[s] = f s1 .. sn , D of sort  C, 
--
 | Just ((ctxt@(MetaCtxtSubst sub Ctxt cname s) :=?=: (Fn fname (args))),rest)
     <- splitBy (\(l:=?=:r) -> case (l,r) of
                                 (MetaCtxtSubst sub srt cname s, Fn fname args) -> (MetaCtxt sub srt cname) `elem` (delta1 del)
                                                                               && srt `elem` [Ctxt]
                                 _ -> False)                                   (expressionEq gam)
    = Some
      [let
         (name_d':fresh') = fresh problem
         d'_s = MetaCtxtSubst emptySubstitution Ctxt ("C"++name_d') s  -- todo
         d' = MetaCtxt emptySubstitution Ctxt ("C"++name_d')  
         d = MetaCtxt sub Ctxt cname
         gam' = gam{expressionEq = (d'_s :=?=: si):rest}
         del' = subst rhs d del 
         rhs = Fn fname (before ++ [Binder v d'] ++ after)
         sol' = (MapExpr $ d :|-> rhs):sol
       in 
         problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
       | (before,(Binder v si),after) <- splits2 args       -- hole may be below binders!
       ]

--
-- D[s] = letrec ...
-- not okay for matching, add an alternative where the hole is in an EE[]-chain
-- 

 | Just ((ctxt@(MetaCtxtSubst sub srt cname s) :=?=: (Letrec env s')),rest) 
     <- splitBy (\(l:=?=:r) -> case (l,r) of
                                (MetaCtxtSubst sub srt cname s, Letrec env s') -> (MetaCtxt sub srt cname) `elem` (delta1 del)
                                                                              && (srt `elem` [TCtxt,Ctxt])
                                _                                          -> False) (expressionEq gam)
    = Some $
      [
      let
       (name_d':fresh') = fresh problem
       d'_s = MetaCtxtSubst emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') s  
       d'   = MetaCtxt emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') 
       d    = MetaCtxt sub srt cname
       gam' = gam{expressionEq = (d'_s :=?=: s'):rest}
       del' = subst rhs d $ del
       rhs = Letrec env d'
       sol' = (MapExpr $ d :|-> rhs ):sol
      in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
     ]
      
      ++ 
      [
      let
       (name_d':name_ee':name_x':name_e:fresh') = fresh problem
       x' = MetaVarLift emptySubstitution $ "X" ++ name_x' 
       ee' =  case typ of 
               VarChain -> "VV" ++ name_ee'
               AChain -> "EE" ++ name_ee'
       d'_s = MetaCtxtSubst emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') s 
       d' = MetaCtxt emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') 
       d = MetaCtxt sub srt cname
       e = Env emptySubstitution ("E" ++ name_e)
       gam' = gam{expressionEq = rest, environmentEq =  (empty{metaEnvVars=[e],metaChains=[MetaChain emptySubstitution typ ee' x' d'_s]} :=?=: env):(environmentEq gam)}
       del' = subst rhs d del
       rhs = Letrec (empty{metaEnvVars=[e],metaChains=[MetaChain emptySubstitution typ ee' x' d']}) s'
       sol' = (MapExpr $ d :|-> rhs):sol
      in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
          
   | typ <- [AChain]] -- AChain should be enough, since the VarChains are subsumed by AChains
  
--
-- D'[s']  = D[s], -- prefix rule
-- requires:
--   sort(d) <= sort(d')
--   D /= [] and D' /= []
--   or both D and D' are unconstrained
 
 | Just (((MetaCtxtSubst sub srtd' d' s') :=?=: ctxt@(ConstantCtxtSubst sub0 srtd d s)),rest)
     <- splitBy (\(l:=?=:r) -> case (l,r) of 
                                (MetaCtxtSubst sub srtd' d' s',ConstantCtxtSubst sub0 srtd d s ) ->  
                                                                         ((     (MetaCtxt sub srtd' d') `elem` (delta1 del)
                                                                           && ((ConstantCtxt sub0 srtd d) `elem` (delta1 cdel) 
                                                                                || ((ConstantCtxt emptySubstitution srtd d)   `elem` (delta1 cdel))
                                                                                ))
                                                                           || ((MetaCtxt sub srtd' d') `notElem` (delta1 del)
                                                                           && ((ConstantCtxt sub0 srtd d) `notElem` (delta1 cdel)
                                                                                &&
                                                                                (ConstantCtxt emptySubstitution srtd d) `notElem` (delta1 cdel)
                                                                                )))
                                                                           && srtd <= srtd'
                                _                                                 -> False) (expressionEq gam)
    = Some $
      {- [

      let
       (name_d'':fresh') = fresh problem
       d'' = (if srtd == Ctxt then "C" else if srtd == TCtxt then "T" else "A") ++ name_d''
       srtd'' = srtd
       rhs = ConstantCtxtSubst sub0 srtd d (MetaCtxt emptySubstitution srtd d'') 
       sol' = (MapExpr $ MetaCtxt sub srtd' d' :|-> rhs):sol
       gam' = gam{expressionEq = (d''_s' :=?=: s):rest}
       d''_s' = MetaCtxtSubst emptySubstitution srtd'' d'' s'
       del' = subst rhs (MetaCtxt sub srtd' d') del
      in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'} 
     ]
     ++ -} 
     [
{- old
-- D'[s'] :=?=: D[s]
-- ========================
-- split s s.t. ctxt[s1] = s  
-- D' |-> D[ctxt]
-- s' =?= s1
     let
       rhs  = ConstantCtxtSubst sub0 srtd d ctxt
       sol' = (MapExpr $ MetaCtxt sub srtd' d' :|-> rhs):sol
       gam' = gam{expressionEq = (s' :=?=: s1):rest}
       del' = subst rhs (MetaCtxt sub srtd' d') del
      in problem{delta=del',gamma=gam',solution=sol'} 
       
       | (ctxt,s1) <- (splitInContexts (\f -> getStrictPos problem f) srtd' s)
     ]
-}
--    new
--       
-- D'[s'] =?= D*[s]   srt(D') >= srt(D*)
-- --->
-- D' |-> D* D''
-- D''[s'] =?= s
-- srt(D'') = srt(D')
     let
       (f:fr) = fresh problem
       srtd'' = srtd'
       d''_name = case srtd'' of 
                    ACtxt -> "A" ++ f
                    TCtxt -> "T" ++ f
                    Ctxt  -> "C" ++ f
       d''_ctxt = MetaCtxt emptySubstitution srtd'' d''_name             
       d''_ctxtsubst = MetaCtxtSubst emptySubstitution srtd'' d''_name             
       rhs  = ConstantCtxtSubst sub0 srtd d d''_ctxt
       sol' = (MapExpr $ MetaCtxt sub srtd' d' :|-> rhs):sol
       gam' = gam{expressionEq = ((d''_ctxtsubst s') :=?=: s):rest}
       del' = subst rhs (MetaCtxt sub srtd' d') del
     in problem{delta=del',gamma=gam',solution=sol',fresh=fr} 
    ]
   
 -- no context rule applicable:
 | otherwise = Empty
-- ********************************************************************************************************************




-- ********************************************************************************************************************
-- Rules for Environments
-- ********************************************************************************************************************
applyEnvRules :: Problem -> MatchResult
applyEnvRules problem@(Problem{solution=sol,gamma=gam,delta=del,constantDelta=cdel})
-- E*;env1 = E*;env2 ==> env1 = env2
  | (eq@(left :=?=: right),rest):_ 
      <-  [splitting | splitting@(left :=?=: right,rest) <- splits (environmentEq gam)
                     ,  not (null 
                               [(e1,e2) | (e1@(Env sub1 v1)) <- constantEnvVars left
                                        , (e2@(Env sub2 v2)) <- constantEnvVars right                                     
                                        , v1 == v2 
                                        , (sub1 `simsubst` sub2)])
                     ]
  , ((env1,env2):_) <-   [(left',right') | (e1@(Env sub1 v1),left') <- splits (constantEnvVars left)
                              , (e2@(Env sub2 v2),right') <- splits (constantEnvVars right)
                              , v1 == v2 
                              , (sub1 `simsubst` sub2)]
                                          
     = 
       let 
           gam' = gam{environmentEq = (left{constantEnvVars=env1} :=?=: right{constantEnvVars=env2}):rest}
       in Some [problem{gamma=gam'}] 
{-
-- EE*[z1,s1];...;EE*[zn,sn];env1 = EE*[z',s'];env2 ==> for all i: env1 = env2, z= z', s=s'
  | (eq@(left :=?=: right),rest):_ 
      <-  [splitting | splitting@(left :=?=: right,rest) <- splits (environmentEq gam)
                     ,  not (null 
                               [(e1,e2) | (e1@(ConstantChain sub1 typee1 ee1 z1 s1)) <- constantChains left
                                        , (e2@(ConstantChain sub2 typee2 ee2 z2 s2)) <- constantChains right                                     
                                        , ee1 == ee2
                                        , typee1 == typee2
                                        , (sub1 `simsubst` sub2)])
                     ]
  , ((list,env2,z2,s2):_) <- [ ([(left',z1,s1) | (e1@(ConstantChain sub1 typee1 ee1 z1 s1),left') <- splits (constantChains left)
                                                                  , ee1 == ee2
                                                                  , typee1 == typee2
                                                                  , (sub1 `simsubst` sub2)
                                ]
                                ,right'
                                ,z2
                                ,s2)
                              |
                                (e2@(ConstantChain sub2 typee2 ee2 z2 s2),right') <- splits (constantChains right)
                              ]
                                          
     = 
       let 
           gam' env1 z1 s1 = gam{environmentEq = (left{constantChains=env1} :=?=: right{constantChains=env2}):rest,
                                 expressionEq = (s1 :=?=: s2):(expressionEq gam),
                                 varLiftEq     = (z1 :=?=: z2):(varLiftEq gam)
                     }
       in Some [problem{gamma=gam' env1 z1 s1} | (env1,z1,s1) <- list] 
-}
--  ----
--  RULE: Sol,Gamma cup {E = env}, Delta --> Sol cup {E |-> env}, Gamma,... if env is nonempty 
--  ----
--  ok for matching
  | ((left :=?=: r),rest):_ 
      <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
                               ,isMetaEnv left
                               ,isSingleton  (metaEnvVars left)
                               , null (constantEnvVars left)
                               , null (constantChains left)
                               ,let [e@(Env sub v)] = metaEnvVars left
                               ,e `notElem` (delta2 del) 
                                 ||
                                (e `elem` (delta2 del) &&  not (isEmpty r))
                                 ||
                                (e `elem` (delta2 del) &&  not (null [e' | e' <- (constantEnvVars r), e' `elem` delta2 (constantDelta problem)]))                                  
                               ]
    ,[e@(Env sub v)] <- (metaEnvVars left)
     = 
       let sol' = (MapEnv $ e :|-> r): sol
           gam' = gam{environmentEq = rest}
           del' = substEnv r v del 
                   
       in Some [problem{solution=sol',gamma=gam',delta=del'}] 
--  ----
--  RULE: Sol,Gamma cup {E1,...,En = env}}, Delta --> Sol, Gamma cup {E1,...,En=env}, Ei in Delta2 if for all i: Ei not in Delta2 and env is non-empty
--  ----
--   | ((left :=?=: r),rest):_ 
--       <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
--                                ,null (bindings left)
--                                ,null (metaChains left)
--                                ,null (constantChains left)                               
--                                ,not (null (metaEnvVars left))
--                                , (not (null [() |  (Env sub e) <- (constantEnvVars  r), (Env sub e) `elem` (delta2 cdel) || (Env emptySubstitution e) `elem` (delta2 cdel)]))
--                                  || 
--                                  (not (null (bindings r)))
--                                  ||
--                                  (not (null (constantChains r)))
--                                ,let es = metaEnvVars left
--                                ,and [ei `notElem` (delta2 del) | ei <- es] 
--                                ]
--     , es <- (metaEnvVars left)
--      = 
--         Some [problem{delta = del{delta2 = ei:(delta2 del)}} | ei <- es] 
--  ----
--  RULE: Sol,Gamma cup {E1...En = {}}, Delta --> Sol cup {Ei |-> {}}, Gamma,... if for all i: Ei not in Delta2 
--  ----
  | ((left :=?=: r),rest):_ 
      <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
                               ,null (bindings left)
                               ,null (metaChains left)
                               ,isEmpty r
                               ,not (null $ metaEnvVars left)
                               ,and [e `notElem` (delta2 del) | e <- metaEnvVars left] 
                               ]
    ,es <- (metaEnvVars left)
    = 
       let sol' = [(MapEnv $  ei :|-> empty) | ei <- es ] ++ sol
           gam' = gam{environmentEq = rest}
           del' = foldr (\e d -> substEmpty e d) del [v | Env s v <- es ]
       in Some [problem{solution=sol',gamma=gam',delta=del'}] 
--  ----
--  RULES: Sol,Gamma cup {...=b1;env1}, Delta --> ...
--  ----
  | (((left :=?=: right),rest):_) <- [(left :=?=: right,rest) | splitting@(left :=?=: right,rest) <- splits (environmentEq gam)
                                                              , hasBinding right
                                                              , not (null (bindings left) && null (metaEnvVars left) && null (metaChains left))]
  , (b1:bindings_right') <- (bindings right)
     =
      Some $ ---------------
--          concat $
--           [ 
              [ -- for each binding b1 from right
                let gam' = gam { bindingEq = (b2 :=?=: b1):bindingEq gam
                               ,  environmentEq = (left' :=?=: right'):rest}
                    left' = left{bindings = bindings_left'}
                    right' = right{bindings = bindings_right'}
                in problem{gamma=gam'}                           
                | (b2,bindings_left') <- splits (bindings left)
              ]
              ++
              [ -- for each metaEnv E from left
                let (name_e':fresh') = fresh problem
                    e' = Env emptySubstitution ("E" ++ name_e')
                    b1e' = empty{bindings=[b1],metaChains=[],metaEnvVars=[e']}
                    sol' = (MapEnv $  ex :|-> b1e'):sol
                    gam' = gam {environmentEq = (left' :=?=: right'):rest}
                    left' = left{metaEnvVars = e':metaEnvVars_left'}
                    right' = right{bindings = bindings_right'}
                    del' = substEnv b1e' e del
                    
                in problem{solution=sol',fresh=fresh',gamma=gam',delta=del'}                           
                         
                | (ex@(Env sub e),metaEnvVars_left') <-  splits (metaEnvVars left) -- , e `elem` (delta2 del) -- not empty e
              ]
              ++
              (concat [ -- for each metaChain EE from left
                -- 4 rules!
                  
                 [let 
                      
                      -- EE[z,s]; env =?= b1; env2
                      -- case
                      -- EE[z,s] -> b1
                      sol' = (MapCV (MetaChain sub typ ee z s :|-> sigma_left)):sol
                      (name_ctxtname:fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      z_A_s = \z s ->
                                case typ of 
                                 AChain -> (z := MetaCtxtSubst emptySubstitution ACtxt ctxtname s)
                                 VarChain -> (z := s)
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (z_A_s z s :=?=: b1):(bindingEq gam)
                                }
                      sigma_left = (\z s -> binding (z_A_s z s))
                      right' = right{bindings=bindings_right'}
                      left' =  left{metaChains = metaChains_left'}
                      del'  = substChainVar sigma_left (MetaChain sub typ ee z s) del
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                  ,
                  let 
                      -- EE[z,s]; env =?= b1; env2
                      -- case
                      -- EE[z,s] -> z=A[X']; EE'[X',s] and z=A[X'] =?= b1 ...
                      sol' = (MapCV $ MetaChain sub typ ee z s :|-> sigma_left):sol
                      sigma_left = \z s -> (binding (z_A_vletX' z)){metaChains = [ee' s]}
                      (name_ctxtname:name_x':name_ee':fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      x' =  MetaVarLift emptySubstitution ("X"++name_x')
                      ee' = \s -> case typ of 
                              AChain -> MetaChain emptySubstitution AChain ("EE" ++ name_ee') x' s 
                              VarChain -> MetaChain emptySubstitution VarChain ("VV" ++ name_ee') x' s 
                      z_A_vletX' = \z -> case typ of 
                                       AChain -> (z := MetaCtxtSubst emptySubstitution ACtxt ctxtname (VarL x'))
                                       VarChain -> (z:= (VarL x'))
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (z_A_vletX' z :=?=: b1):(bindingEq gam)
                                }
                      left' = left{metaChains = (ee' s):metaChains_left'}
                      right' = right{bindings=bindings_right'}
                      del' = substChainVar sigma_left (MetaChain sub typ ee z s) del 
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                  ,
                  let 
                      -- EE[z,s]; env =?= b1; env2
                      -- case
                      -- EE[z,s] -> EE1[z,X1]; X1=A[X2]; EE2[x2,s] and X1=A[X2] =?= b1 ...
                      -- EE[z,s] -> EE1[z,A'[X1]]; X1=A[X2]; EE2[x2,s] and X1=A[X2] =?= b1 ...
                      sol' = (MapCV $ MetaChain sub typ ee z s :|-> sigma_left):sol
                      sigma_left = \z s -> (binding x1_A_vletX2){metaChains = [ee1 z,ee2 s]}
                      (name_ctxtname:name_ctxtnameA':name_x1:name_x2:name_ee1:name_ee2:fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      ctxtnameA' = "A" ++ name_ctxtnameA'
                      ctxtA' = \inhole -> case typ of 
                                 AChain -> (MetaCtxtSubst emptySubstitution ACtxt ctxtnameA' inhole)
                                 VarChain -> inhole
                      x1 =  MetaVarLift emptySubstitution ("X"++name_x1)
                      x2 =  MetaVarLift emptySubstitution ("X"++name_x2)
                      ee1 = \z -> 
                             case typ of
                               AChain -> MetaChain emptySubstitution AChain ("EE" ++ name_ee1) z (ctxtA' (VarL x1))
                               VarChain -> MetaChain emptySubstitution VarChain ("VV" ++ name_ee1) z (ctxtA' (VarL x1))
                      ee2 = \s ->
                             case typ of 
                               AChain -> MetaChain emptySubstitution AChain ("EE" ++ name_ee2) x2 s
                               VarChain -> MetaChain emptySubstitution VarChain ("VV" ++ name_ee2) x2 s
                      x1_A_vletX2 = case typ of 
                                       AChain -> (x1 := MetaCtxtSubst emptySubstitution ACtxt ctxtname (VarL x2))
                                       VarChain -> (x1 := VarL x2) 
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (x1_A_vletX2 :=?=: b1):(bindingEq gam)
                                }
                      right' = right{bindings=bindings_right'}
                      left' = left{metaChains = [ee1 z,ee2 s]++metaChains_left'}
                      del' = substChainVar sigma_left (MetaChain sub typ ee z s) del
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                 ,
                  let 
                      -- sol' = (MapEnv $ metaChain typ ee z s :|-> sigma_right):sol
                      -- EE[z,s]; env =?= b1; env2
                      -- case
                      -- EE[z,s] -> EE'[z,A'[X1]];X1=A[s] and  X1=A[s] =?= b1 ...
                      sol' = (MapCV $ MetaChain sub typ ee z s :|-> sigma_left):sol
                      sigma_left= \z s -> (binding (x1_A_s s)){metaChains = [ee' z]}
                      (name_ctxtname:name_ctxtnameA':name_x1:name_ee':fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      ctxtnameA' = "A" ++ name_ctxtnameA'
                      ctxtA' = \inhole -> case typ of 
                                 AChain -> (MetaCtxtSubst emptySubstitution ACtxt ctxtnameA' inhole)
                                 VarChain -> inhole

                      x1 =  MetaVarLift emptySubstitution ("X"++name_x1)
                      ee' = \z -> 
                             case typ of 
                              AChain -> MetaChain emptySubstitution AChain ("EE" ++ name_ee') z (ctxtA' (VarL x1))
                              VarChain -> MetaChain emptySubstitution VarChain ("VV" ++ name_ee') z (ctxtA' (VarL x1))
                      x1_A_s = \s -> 
                                case typ of 
                                 AChain -> (x1 := MetaCtxtSubst emptySubstitution ACtxt ctxtname s)
                                 VarChain -> (x1 := s)
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = ((x1_A_s s) :=?=: b1):(bindingEq gam)
                                }
                      right' = right{bindings=bindings_right'}
                      left' = left{metaChains = (ee' z):metaChains_left'}
                      del' = substChainVar sigma_left (MetaChain sub typ ee z s) del
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                 ]
                 
                -- already changed
                | ((MetaChain sub typ ee z s),metaChains_left') <- splits (metaChains left)
              ])
           -- for all b1 in the right
           -- already changed:
--            | (b1,bindings_right') <- splits (bindings right)
--           ]
          

--  RULE E1;env = E2;....  ---> E1 |-> E1' E2;   E1';env = ...  (all cases for E1 in parallel)
-- should be ok????:
  | ((left :=?=: right,rest):_) 
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (null (metaEnvVars left))
         , not (null (constantEnvVars right))
         , -- exists E2 s.t. E2 is not empty (in CDelta2)
           (or
            [(Env sub e2') `elem` (delta2 cdel) || (Env emptySubstitution e2') `elem` (delta2 cdel) 
              | (Env sub e2') <- (constantEnvVars right)
            ])
           -- or at least one E1 on the lhs is not in Delta1:
           || 
           ((any (\(v@(Env sub e1)) -> not (v `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del))) (metaEnvVars left)))
         ] 
       -- choose e2:
     , e2@(Env sub e2'):e2s <- if  -- if there exists at least one E1 on the lhs not in Delta2, then E2 can be choosen arbitrarily (here the first one)
                                   ((any (\(v@(Env sub e1)) -> not (v `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del))) (metaEnvVars left))) 
                                 then constantEnvVars right
                                 else -- filter only those E2 which are in delta2
                                      head 
                                        [(Env sub e2'):r | (Env sub e2',r) <- (splits (constantEnvVars right)), (Env sub e2') `elem` (delta2 cdel) || (Env emptySubstitution e2') `elem` (delta2 cdel)
                                        ]
   = 
    Some [
      let
         (name_e1':fresh') = fresh problem
         e1' = Env emptySubstitution ("E" ++ name_e1')
         sol' = (MapEnv (Env sub e1 :|-> empty{metaEnvVars=[e1'],constantEnvVars=[e2]})):sol
         gam' = gam {environmentEq = (left' :=?=: right'):rest}
         left' = left{metaEnvVars = (e1'):e1s}
         right' = right{constantEnvVars = e2s} 
         del' = substEnv empty{metaEnvVars=[e1'],constantEnvVars=[e2]} e1 del
                          
      in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'} 
       | (Env sub e1,e1s) <- splits (metaEnvVars left)
       , if -- e2 is non-empty
            (e2 `elem` (delta2 cdel) ||(Env emptySubstitution e2') `elem` (delta2 cdel))
            then True -- E1 may be in Delta2 or not in Delta2
            else -- e2 is empty or non-empty, then E1 must not be in Delta2
              not( (Env sub e1) `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del))
       ]
       
 {-  dead code    
  | ((left :=?=: right,rest):_)
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (null (metaEnvVars left))
         , not (null (constantEnvVars right))
         -- all E1 on the left are non-empty
         -- AUG16: not necessary? , (all (\(v@(Env sub e1)) -> (v `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del))) (metaEnvVars left))
         -- some E2 is nonempty
         , (any (\(v@(Env sub e2)) -> (v `elem` (delta2 cdel) || (Env emptySubstitution e2) `elem` (delta2 cdel))) (constantEnvVars right))

         ]
       -- E2 is non-empty
     , (e2@(Env sub e2'),e2s):_ <- [(v,r) | (v@(Env sub e2),r) <- splits (constantEnvVars right),  (v `elem` (delta2 cdel) || (Env emptySubstitution e2) `elem` (delta2 cdel))]
   = error "STOP HERE "
   -}
   {- Some [
      let
         (name_e1':fresh') = fresh problem
         e1' = Env emptySubstitution ("E" ++ name_e1')
         sol' = (MapEnv (Env sub e1 :|-> empty{metaEnvVars=[e1'],constantEnvVars=[e2]})):sol
         gam' = gam {environmentEq = (left' :=?=: right'):rest}
         left' = left{metaEnvVars = (e1'):e1s}
         right' = right{constantEnvVars = e2s}
         del' = substEnv empty{metaEnvVars=[e1'],constantEnvVars=[e2]} e1 del

      in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'}
       | (Env sub e1,e1s) <- splits (metaEnvVars left)
       -- all E1 can be used (irrelevant whether they are in Delta2 or not, since E2 is in Delta2
       ]
       
 -}       
{-
-- seems to be wrong:
  | ((left :=?=: right,rest):_) 
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (hasBinding left) && not (hasBinding right)
         ] 
     , e2:e2s <- constantEnvVars right
     , es <- metaEnvVars left
     , not (null es)
   = 
    Some [
      let
         (name_e1':fresh') = fresh problem
         e1' = Env emptySubstitution ("E" ++ name_e1')
         sol' = (MapEnv (Env sub e1 :|-> empty{metaEnvVars=[e1'],constantEnvVars=[e2]})):sol
         gam' = gam {environmentEq = (left' :=?=: right'):rest}
         left' = left{metaEnvVars = (e1'):e1s}
         right' = right{constantEnvVars = e2s} 
         del' = substEnv empty{metaEnvVars=[e1'],constantEnvVars=[e2]} e1 del
                          
      in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'} | (Env sub e1,e1s) <- splits (metaEnvVars left)
       ]
-}
-- new for matching: 
 
   --   EE1[z1,s1] =?= EE2[z2,s2]
   -- EE1[.,.] -> EE2[.,.]
   -- z1 = z2; s1 = s2 
-- Problem: this does not match EE1[z1,vlet x] against EE1[z1,app x y]

{- AUG16: case seems to be to special:
  | ((left :=?=: right,rest):_) 
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (hasBinding left) && not (hasBinding right) && null (metaEnvVars left) && null (constantEnvVars right)
         , length (metaChains left) == 1, length (constantChains right) == 1
         ] 
     , [ch@(MetaChain sub typee1 ee1 z1 s1)] <- metaChains left
     , [ConstantChain sub0 typee2 ee2 z2 s2] <- constantChains right
     , typee1 >= typee2 -- ??
   = Some $
     
 
     -- EE1[z1,s1] =?= EE2[z2,A[s2']]
     
--      ==> EE1[.,.] |-> EE2[.,A[.]]
--          z1 =?= z2
--          s1 ? s2'
     -- EE1[z1,s1] =?= VV2[z2,s2]
     
--      ==> EE1[.,.] |-> VV2[.,A[s2']]
--          z1 =?= z2
--          s1 ? A[s2]
     
      [
      let
         sigma_right = \parz pars -> empty{constantChains=[ConstantChain sub0 typee2 ee2 parz (subst pars (Hole emptySubstitution) ctxt)]}
         sol' = (MapCV (ch :|-> sigma_right):sol)
         gam' = gam {environmentEq = rest
                    ,varLiftEq = (z1 :=?=: z2):(varLiftEq gam)
                    ,expressionEq = (s1 :=?=: s2'):(expressionEq gam)
                    }
         del' = applySolution [(MapCV (ch :|-> sigma_right))] del
         
      in problem{delta=del',solution=sol',gamma=gam'}
         |(ctxt,s2') <- case typee1 of 
                         AChain -> (splitInContexts (\f -> getStrictPos problem f) ACtxt s2)
                         VarChain -> [(Hole emptySubstitution,s2)]
      ]
      -}
-- new for matching:
--  RULE EE1[z1,s1];E1,..env = EE2[z,s];....  
--   --->  (all cases for E1,EE1[z',s'] in parallel)
--   
--    possibilities
--      - E1 -> E1' EE2[z,s]

--      - EE1[z1,s1] -> EE2[z,s]  and z1 = z; s1 = s as equations!
--      ~ EE1[z1,s1] -> EE2[z,s]  and z1 = z; A[s1] = s as equations!

--      - EE1[z1,s1] -> EE1'[z1,vlet X]; EE2[z,s] and  X1 = z2  and EE1'[z',vlet X] remains in lhs 

--      - EE1[z1,s1] -> EE2[z,s];EE1'[X,s1] and z1 = z, vlet X = s and EE1'[X,s1] remain in lhs
--      ~ EE1[z1,s1] -> EE2[z,s];EE1'[X,s1] and z1 = z, A[vlet X] = s and EE1'[X,s1] remain in lhs          

--      - EE1[z1,s1] -> EE1'[z1,vlet X1]; EE2[z,s]; EE1''[X2; s1]
--         and X1 = z: vlet X2 = s as equation and EE1'[z1,vlet X1]; EE1'''[X2,s] remain in lhs
--      ~ EE1[z1,s1] -> EE1'[z1,vlet X1]; EE2[z,s]; EE1''[X2; s1]
--         and A[X1] = z: A'[vlet X2] = s as equation and EE1'[z1,vlet X1]; EE1'''[X2,s] remain in lhs

--   if they type of chains is the same then ok
--   there is an additional case where the type of EE1 is AChain and the type of EE2 is VarChain! 
  | ((left :=?=: right,rest):_) 
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         -- , --not (hasBinding left) && 
           -- not (hasBinding right) -- && (null $ constantEnvVars right)
         ] 
     , ee2@(ConstantChain sub0 typee2 nameee2 z s):ee2s <- constantChains right
     , ee1s <- metaChains left
     , e1s  <- metaEnvVars left
     , constantChainsLeft <- [cCL | cCL@(ConstantChain subCL typCL nameCL zCL sCL) <- constantChains left, nameCL == nameee2, subCL `simsubst` sub0]
     , not ((null $ e1s) && (null  ee1s) && null constantChainsLeft)
   = Some $
      -- EE*[z,s];env =?= EE*[z',s'];env' =
      -- =>
      -- z =?= z' 
      -- s =?= s'
      -- env =?= env'
     [
      let
         newVarEq  = (zCL :=?=: z):(varLiftEq gam)
         newExprEq = (sCL :=?=: s):(expressionEq gam)
         newLeft   = left{constantChains = restLeft} 
         newRight  = right{constantChains = ee2s}
         newEnvEq  = (newLeft :=?=: newRight):(rest)
         gam' = gam{varLiftEq = newVarEq, expressionEq = newExprEq, environmentEq = newEnvEq}
      in problem{gamma = gam'}
       |
        (ConstantChain subCL typCL nameCL zCL sCL,restLeft) <- splits (constantChains left), nameCL == nameee2, subCL `simsubst` sub0
      ]
      ++    
   
--      - E1 -> E1' EE2[z,s]
     [
      let
         (name_e1':fresh') = fresh problem
         e1' = Env emptySubstitution ("E" ++ name_e1')
         sol' = (MapEnv (Env sub e1 :|-> empty{metaEnvVars=[e1'], constantChains=[ee2]})):sol
         gam' = gam {environmentEq = (left' :=?=: right'):rest}
         left' = left{metaEnvVars = (e1'):e1s}
         right' = right{constantChains = ee2s} 
         del' = substEnv empty{metaEnvVars=[e1'], constantChains=[ee2]} e1 del
      in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'} | (Env sub e1,e1s) <- splits (metaEnvVars left)
      ]
      
      
      
      
      ++
      concat
       [ 
        --      - EE1[z1,s1] -> EE2[z,s]  and z1 = z; s1 = s as equations!
        --      ==> EE1[.,.] |-> EE2[.,A[.]]
--          z1 =?= z2
--          s1 ? s2'
     
         [let
          sol'   = (MapCV $ (ee1 :|-> sigma_right)):sol
          sigma_right = \z s -> empty{constantChains = [ConstantChain sub0 typee2 nameee2 z (subst s (Hole emptySubstitution)  ctxt)]}  
          gam'   = gam {environmentEq = (left' :=?=: right'):rest
                       , expressionEq = (s1:=?=:s2'):(expressionEq gam)
                       , varLiftEq     = (z1:=?=:z):(varLiftEq gam)}
          left'  = left{metaChains = ee1s}
          right' = right{constantChains = ee2s} 
          del'   = substChainVar sigma_right ee1 del
         in problem{delta=del',solution=sol',gamma=gam'} 
                  |(ctxt,s2') <- case typee1 of 
                         AChain -> (splitInContexts (\f -> getStrictPos problem f) ACtxt s)
                         VarChain -> [(Hole emptySubstitution,s)]

        ]
        
        
        ++
        --      - EE1[z1,s1] -> EE1'[z1,vlet X]; EE2[z,s] and  s1 = s 
        --    EE1[.,.] -> EE1'[.,A[vlet z]];EE2[z,A[.]]  
        --    s1 =?= s2' 
        --  where A is fresh
        
        [
         let
          sol'   = (MapCV $ (ee1 :|-> sigma_right)):sol
          sigma_right = \parz s -> (empty{metaChains=[ee1' parz],constantChains=[ConstantChain sub0 typee2 nameee2 z (subst s (Hole emptySubstitution)  ctxt)]})
          (name_ee1':name_x:name_a:fresh') = fresh problem
          -- x = "X" ++ name_x
          -- meta_x      = MetaVariableLet x
          actxt  = case typee1 of 
                    AChain -> MetaCtxtSubst emptySubstitution ACtxt ("A" ++ name_a)
                    VarChain -> id
          ee1'   = \parz -> MetaChain emptySubstitution typee1 ((prefixEE typee1) ++ name_ee1') parz (actxt (VarL z))
          gam'   = gam {environmentEq = (left' :=?=: right'):rest
                       ,expressionEq = (s1 :=?=: s2'):(expressionEq gam)}
          left'  = left{metaChains = (ee1' z1):ee1s}
          right' = right{constantChains = ee2s} 
          del'   = substChainVar sigma_right ee1 del
         in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'} 
           | (ctxt,s2') <- case typee1 of 
                         AChain -> (splitInContexts (\f -> getStrictPos problem f) ACtxt s)
                         VarChain -> [(Hole emptySubstitution,s)]
        ]
        ++
        --      - EE1[z1,s1] -> EE2[z,s];EE1'[X,s1] and EE1'[X,s1] remain in lhs  
--
--                  EE1[.,.] -> EE2[.,s];EE1'[X,.]
--                  z1 = z       
--                  A[vlet X] = s
                  
        -- special cases: 
        --        EE1[z1,s1] -> VV2[z,s];EE1'[X,s1] and z1 = z, vlet X = s
        --        VV1[z1,s1] -> VV2[z,s];VV1'[X,s1] and z1 = z, vlet X = s
        [
         let
          sol'   = (MapCV $ ee1:|-> sigma_right):sol
          sigma_right = \parz pars -> empty{metaChains = [ee1' pars], constantChains =[ConstantChain sub0 typee2 nameee2 parz s]}
          (name_ee1':name_x:name_a:fresh') = fresh problem
          a = "A" ++ name_a
          ctxtAroundx = case typee2 of 
                         AChain -> MetaCtxtSubst emptySubstitution ACtxt a
                         VarChain -> id
          x = "X" ++ name_x
          meta_x      = MetaVarLift emptySubstitution x
          ee1'   = \pars1 -> MetaChain emptySubstitution typee1 ((prefixEE typee1) ++ name_ee1') meta_x pars1
          gam'   = gam {environmentEq = (left' :=?=: right'):rest
                       , varLiftEq     = (z1 :=?=: z):(varLiftEq gam)
                       ,expressionEq  = (ctxtAroundx (VarL meta_x) :=?=: s):(expressionEq gam)
                       }
          left'  = left{metaChains = (ee1' s1):ee1s}
          right' = right{constantChains = ee2s} 
          del'   = substChainVar sigma_right ee1 del
         in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'} 
        ]
        ++
        --      - EE1[z1,s1] -> EE1'[z1,vlet X1]; EE2[z,s]; EE1''[X2; s1] 
        --         and X1 = z; A[vlet X2] = s as equation and EE1'[z1,vlet X1]; EE1'''[X2,s] remain in lhs
        
                
--
--                  EE1[.,.] -> EE1'[.,A0[vlet X1]]; EE2[z,s]; EE1''[X2,.]
--                  X1 = z       
--                  A[vlet X2] = s
--                  and EE1'[z1,vlet X1]; E1''[X2,s1] remain

        [
          let
              sol'   = (MapCV $ ee1 :|-> sigma_right):sol
              sigma_right = \parz pars -> (empty{metaChains=[meta_ee1' parz ,meta_ee1'' pars]
                                                ,constantChains =[ConstantChain sub0 typee2 nameee2 z s]})
              (name_ee1':name_x1:name_ee1'':name_x2:name_a:name_a0:fresh') = fresh problem
              a = "A" ++ name_a
              a0 = "A" ++ name_a0
              ctxtAroundx2 = case typee2 of 
                              AChain -> MetaCtxtSubst emptySubstitution ACtxt a
                              VarChain -> id
              ctxtAroundx1 = case typee1 of 
                              AChain -> MetaCtxtSubst emptySubstitution ACtxt a0
                              VarChain -> id
              x1 = "X" ++ name_x1
              x2 = "X" ++ name_x2
              ee1' = (prefixEE typee1) ++ name_ee1'
              ee1'' = (prefixEE typee1) ++ name_ee1''          
              meta_x1      = MetaVarLift emptySubstitution x1
              meta_x2      = MetaVarLift emptySubstitution x2
              meta_ee1'   = \parz -> MetaChain emptySubstitution typee1 ee1' parz (ctxtAroundx1 (VarL meta_x1))
              meta_ee1''  = \pars -> MetaChain emptySubstitution typee1 ee1'' meta_x2 pars
              gam'   = gam {environmentEq = (left' :=?=: right'):rest
                           ,varLiftEq      = (meta_x1 :=?=: z):(varLiftEq gam)
                           ,expressionEq  = (ctxtAroundx2 (VarL meta_x2) :=?=: s):(expressionEq gam)
                           }
              left'  = left{metaChains = (meta_ee1' z1):(meta_ee1'' s1):ee1s}
              right' = right{constantChains = ee2s} 
              del'   = substChainVar sigma_right ee1 del
          in problem{delta=del',solution=sol',gamma=gam',fresh=fresh'} 
        ]
        
         | (ee1@(MetaChain sub typee1 nameee1 z1 s1),ee1s) <- splits (metaChains left), typee2 <= typee1
       ]      
      
 -- no environment rule applicable:
 | otherwise = Empty
-- ********************************************************************************************************************
prefixEE (AChain) = "EE"
prefixEE (VarChain) = "VV"





-- =========================================================
-- TO BE DELETED

-- no longer used, only for debugging 
applyOldFailRules :: Problem -> MatchResult
applyOldFailRules problem@(Problem{solution=sol,gamma=gam,delta=del,constantDelta=cdel})
       
  
  --
--  Fail: Sol,Gamma cup {x = y}, Delta --> Fail if x/= y
--        or if x_{subst1} = x_{subst2} and subst1 /= subst2
--
  |  (eq:_) <- [eq | eq@(VarLift substx x :=?=: VarLift substy y) <- (varLiftEq gam), x /= y || not (substx `simsubst` substy)]
    =  MatchFail $ show eq 

--  n{x} = X*      
  |  (eq:_) <- [eq | eq@((VarLift substx x) :=?=: (ConstantVarLift  substy y)) <-  (varLiftEq gam)] 
  =  MatchFail $ show eq 
--
-- Fail for matching: x* = y 
--
  |  (eq:_) <- [eq | eq@((ConstantVarLift  substx x) :=?=: (VarLift  substy y)) <-  (varLiftEq gam)]
    =  MatchFail $ show eq 
--
-- Fail for matching x* = y* or x*_subst1 = x*_subst2 and subst1 /= subst2
--
  |  (eq:_) <- [eq | eq@((ConstantVarLift  substx x) :=?=: (ConstantVarLift  substy y)) <-  (varLiftEq  gam), x /= y || not (substx `simsubst` substy)] 
    =  MatchFail $ show eq 

    
  | (eq:_) <- (  
-- 
-- Fail: f s1 ... sn = var x
-- 
                [eq | eq@((Fn name args) :=?=: VarL _)                      <- (expressionEq gam)]
--
-- Fail: var x = f s1 ... sn
--
             ++ [eq | eq@((VarL _) :=?=: Fn name args)                      <- (expressionEq gam)]
--
-- Fail: f s1 ... sn = letrec ...
--
             ++ [eq | eq@((Fn _ _) :=?=: (Letrec _ _))                      <- (expressionEq gam)]
--
-- Fail: letrec ... = f s1 ... sn
--
             ++ [eq | eq@((Letrec _ _) :=?=: (Fn _ _))                      <- (expressionEq gam)]
--
-- Fail: (f1 s1...sn) = (f2 t1...tn)   and f1 /= f2
--
             ++ [eq | eq@((Fn name _) :=?=: (Fn name' _))                   <- (expressionEq gam), name /= name']
--
-- Fail: (var x) = letrec ...
--
             ++ [eq | eq@((VarL _) :=?=: (Letrec _ _))                      <- (expressionEq gam)]
--
-- Fail: letrec ... = (var x)
--
             ++ [eq | eq@((Letrec _ _) :=?=: VarL _ )                       <- (expressionEq gam)]
--
-- Fail: A[s] = letrec ... and A in Delta1 (i.e. A is non-empty)
--
             ++ [eq | eq@(vars@(MetaCtxtSubst s ACtxt d _) :=?=: Letrec _ _) <- (expressionEq gam), (MetaCtxt s ACtxt d) `elem` (delta1 del)]
--
--
-- Fail: D[s] = var x and D in Delta1 (i.e. D is non-empty)
--
             ++ [eq | eq@(vars@(MetaCtxtSubst s srt d _) :=?=: VarL _)       <- (expressionEq gam), (MetaCtxt s srt d) `elem` (delta1 del)]
--
-- Fail: D[s] = f, ar(f) = 0 and  D in Delta1 (i.e. D is non-empty)
--
             ++ [eq | eq@(vars@(MetaCtxtSubst s srt d _) :=?=: Fn f [])      <- (expressionEq gam), (MetaCtxt s srt d) `elem` (delta1 del)]
--
-- Fail: S1* = S2* or S* subst1 = S* subst2 and subst1 /= subst2
--
             ++ [eq | eq@((Constant sub1 c1) :=?=: Constant sub2 c2)               <- (expressionEq gam), (c1 /= c2 || sub1 /= sub2)]
--
-- Fail: S1* = C*[s]
--
             ++ [eq | eq@((Constant sub1 c1) :=?=: ConstantCtxtSubst _ _ _ _)               <- (expressionEq gam)]
--
-- Fail: letrec ... = S*
--
             ++ [eq | eq@((Letrec _ _) :=?=: Constant _ _)                   <- (expressionEq gam)]
--
-- Fail: f s1 ... sn = S*
--
             ++ [eq | eq@((Fn _ _) :=?=: Constant _ _)                       <- (expressionEq gam)]
--
-- Fail: var x = S*
--
             ++ [eq | eq@((VarL _) :=?=: Constant _ _)                       <- (expressionEq gam)]

--
-- Fail: S* = f s1 ... sn =
--
             ++ [eq | eq@((Constant _ _) :=?=: (Fn _ _))                       <- (expressionEq gam)]
--
-- Fail: S* = var x 
--
             ++ [eq | eq@((Constant _ _) :=?=: (VarL _)  )                     <- (expressionEq gam)]

-- Fail: S* = letrec ... 
--
             ++ [eq | eq@((Constant _ _) :=?=: (Letrec _ _))                   <- (expressionEq gam)]

--
-- Fail: letrec ... = C*
--
             ++ [eq | eq@((Letrec _ _) :=?=: ConstantCtxt _ _ _)             <- (expressionEq gam)]
--
-- Fail: f s1 ... sn = C*
--
             ++ [eq | eq@((Fn _ _) :=?=: ConstantCtxt _ _ _)                 <- (expressionEq gam)]
--
-- Fail: var x = C*
--
             ++ [eq | eq@((VarL _) :=?=: ConstantCtxt _ _ _)                 <- (expressionEq gam)]
--
-- Fail: letrec ... = C*[s]
--
             ++ [eq | eq@((Letrec _ _) :=?=: ConstantCtxtSubst _ _ _ _)      <- (expressionEq gam)]
--
-- Fail: f s1 ... sn = C*[s]
--
             ++ [eq | eq@((Fn _ _) :=?=: ConstantCtxtSubst _ _ _ _)          <- (expressionEq gam)]
--
-- Fail: var x = C*[s]
--
             ++ [eq | eq@((VarL _) :=?=: ConstantCtxtSubst _ _ _ _)          <- (expressionEq gam)]
--
-- Fail: D[s] = S* if D is non-empty
--
             ++ [eq | eq@(vars@(MetaCtxtSubst s srt d _) :=?=: Constant _ _) <- (expressionEq gam),
                                                                              (MetaCtxt s srt d) `elem` (delta1 del)
                 ]
--
-- Fail: D1[s] = D2*[t] and sort(D2) > sort(D1)
--
             ++ [eq | eq@(vars@(MetaCtxtSubst s srt1 d _) :=?=: ConstantCtxtSubst _ srt2 _ _)          
                                                                             <- (expressionEq gam), srt2 > srt1, 
                                                                                                     (MetaCtxt s srt1 d) `elem` (delta1 del)
                                                                             ]
--
-- Fail: D1*[s] = D2*[t] and sort(D1) /= sort(D2)  or D1*subst1[s] = D2*subst2[t] and subst1 /= subst2
--
             ++ [eq | eq@(vars@(ConstantCtxtSubst subst1 srt1 c1 _) :=?=: ConstantCtxtSubst subst2 srt2 c2 _)     
                                                                             <- (expressionEq gam), c1 /= c2 || (not (subst1 `simsubst` subst2))]
--
-- Fail: D1[s] = f s1 ... sn, D1 in Delta1, sort(D1) = A or T, operator arity of f = (n1,...,nm) with m > 0 and all ni > 0 or all ni = -1 (i.e. VarPos)
--
             ++ [eq | eq@(vars@(MetaCtxtSubst sub srt d _) :=?=: Fn f args)  <- (expressionEq gam)
                                                                                ,(MetaCtxt sub srt d) `elem` (delta1 del)
                                                                                , srt < Ctxt
                                                                                , let bari:_ = [bindarity | (name,arity,strictpos,bindarity) <- functions problem, f == name]
                                                                                -- , all (> 0) bari || all (== (-1)) bari  -- AUG16
                                                                                , all (/= 0) bari 
                                                                                , not (null bari)
                                                                                ]
             )
--
-- Fail D1[s] = D2*[t], D1 in Delta1 but D2* not in Delta1
--
    = MatchFail $ show eq
 | (eq:_) <- [eq | eq@(vars@(MetaCtxtSubst subst1 srt1 c1 _) :=?=: ConstantCtxtSubst subst2 srt2 c2 _)     
                                                                             <- (expressionEq gam)
                                                                                ,(MetaCtxt subst1 srt1 c1) `elem` (delta1 del)
                                                                                ,((ConstantCtxt subst2 srt2 c2) `notElem` (delta1 (constantDelta problem))
                                                                                && ((ConstantCtxt emptySubstitution srt2 c2) `notElem` (delta1 (constantDelta problem))))
                                                                                                                       ]              
    = MatchFail $ "Context on the lhs is more general than on the rhs" ++  show eq
--               
-- Fail: EE[z,s]; EE'[z,s']; ... = ... (same z)             
--
  | (eq:_) <- [eq | eq@(l :=?=: r)  <- environmentEq gam
                  , not (null [ ee | (MetaChain _ _ ee z s) <- metaChains l
                                   , (MetaChain _ _ ee' z' s') <- metaChains l
                                   , ee /= ee'
                                   , z == z'])
              ]   
    = MatchFail $ show eq              
--               
-- Fail: ... = EE*[z,s]; EE*'[z,s']; ... (same z)             
--
  | (eq:_) <- [eq | eq@(l :=?=: r)  <- environmentEq gam
                  , not (null [ ee | (ConstantChain _ _ ee z s) <- constantChains r
                                   , (ConstantChain _ _ ee' z' s') <- constantChains r
                                   , ee /= ee'
                                   , z == z'])
              ]   
    = MatchFail $ show eq              
--    
-- Fail: env = {} and env is not empty
--
   | (eq:_) <- [eq | eq@(left :=?=: right)                                   <- (environmentEq gam)
                                                                                ,isEmpty right && (not (null (bindings left )) 
                                                                                                || not (null (metaChains left)) 
                                                                                                || not (null (constantChains left))
                                                                                                || not (null [v | v <- constantEnvVars left])
                                                                                                || not (null [ v | v <- metaEnvVars left, v `elem` (delta2 del)]))]
       = MatchFail $ show eq
--    
-- Fail: {} = env and env is not empty
--
   | (eq:_) <- [eq | eq@(left :=?=: right)                                   <- (environmentEq gam),isEmpty left && (not $ isEmpty $ right)]
       = MatchFail $ show eq
       
--       
-- Fail: z=s;env = env'  where env' has no bindings
--
   | (eq:_) <- [eq | eq@(left :=?=: right)                                   <- (environmentEq gam)
                                                                                ,null (bindings right)
                                                                                ,not (null (bindings left))]
       = MatchFail $ show eq
--        
-- Fail: env = E  where env  contains a binding or an EE-Chain
--
   | (eq:_) <- [eq | eq@(left :=?=: right)                                   <- (environmentEq gam)
                                                                                ,isSingleton (constantEnvVars right)
                                                                                ,null (constantChains right)
                                                                                ,null (bindings right)
                                                                                ,not (null (bindings left) && null (metaChains left))]   
       = MatchFail $ "U1" ++ show eq
--       
-- Fail EE1[z,s] =?=  E;env' -- cannot match a chain against an arbitrary environemt
--
   | (eq:_) <- [eq | eq@(left :=?=: right)                                   <- (environmentEq gam)
                                                                                ,null (bindings left) && null (metaEnvVars left) && (not $ null $ metaChains left)
                                                                                ,null (bindings right) && (not $ null (constantEnvVars right))]
       = MatchFail $ "U2" ++ show eq
--       
-- Fail: A[s] = f s1 ... sn and f has no strict positions
--
  | Just eq@((ctxt@(MetaCtxtSubst sub ACtxt cname s) :=?=: (Fn fname arguments)),rest)
                                                                            <- splitBy (\(l:=?=:r) -> 
                                                                                            case (l,r) of 
                                                                                              (MetaCtxtSubst sub ACtxt cname s, Fn fname _) ->  (MetaCtxt sub ACtxt cname) `elem` (delta1 del)
                                                                                              _ -> False) (expressionEq gam)
  ,Just [] <- (getStrictPos problem fname)  
   = MatchFail $ show eq
--
-- Fail: D1[s] = D2*[t], D1 in Delta1, D2 not in Delta2
--
  | Just  (eq@((MetaCtxtSubst sub srtd' d' s') :=?=: ctxt@(ConstantCtxtSubst sub0 srtd d s)),rest)
                                                                            <- splitBy (\(l:=?=:r) -> 
                                                                                            case (l,r) of 
                                                                                              (MetaCtxtSubst sub srtd' d' s',ConstantCtxtSubst sub0 srtd d s ) -> 
                                                                                                                    ((MetaCtxt sub srtd' d') `elem` (delta1 del)
                                                                                                                        && (ConstantCtxt sub0 srtd d) `notElem` (delta1    cdel)
                                                                                                                        && (ConstantCtxt emptySubstitution srtd d) `notElem` (delta1    cdel)
                                                                                                                        )
                                                                                              _ -> False) (expressionEq gam)       
   = MatchFail $ show eq                                
--
-- Fail: E = {}  and E in Delta2
--
{-AUG16: should already be covered
  | (eq@(left :=?=: r),rest):_ 
      <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
                               ,isMetaEnv left
                               ,isSingleton  (metaEnvVars left)
                               ,let [e] = metaEnvVars left
                               ,null (constantChains r)
                               ,null (bindings r)
                               ,not (null (constantEnvVars r))
                               ,e `elem` delta2 del
                               ,null [e' | (Env _ e') <- (constantEnvVars r), e' `elem` [e'' | (Env _ e'') <- delta2 (constantDelta problem)]]                                  
                               ]
   = MatchFail $ show eq
-}
{- should be ok??? -}
  | (eq@(left :=?=: right,rest):_) 
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (hasBinding left) && not (hasBinding right)
         , null (metaChains left)
         , null (constantChains left)
         , not (null (metaEnvVars left))
         , not (null (constantEnvVars right))
         , all (\(v@(Env sub e1)) -> v `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del)) (metaEnvVars left)
         , all (\(v@(Env sub e1)) -> not ((v `elem` (delta2 cdel)) || ((Env emptySubstitution e1) `elem` (delta2 cdel)))) (constantEnvVars right)
         ]
   = MatchFail $ "all envvars in rhs are more general than envvars on right" ++ (show eq)
 {- -}
{-
| (eq@(left :=?=: right,rest):_)
      <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (hasBinding left) && not (hasBinding right)
         , null (metaChains left)
         , null (constantChains left)
         , length (metaEnvVars left) > length ((constantEnvVars right))
         , all (\(v@(Env sub e1)) -> v `elem` (delta2 del) || (Env emptySubstitution e1) `elem` (delta2 del)) (metaEnvVars left)
         , all (\(v@(Env sub e1)) -> ((v `elem` (delta2 cdel)) || ((Env emptySubstitution e1) `elem` (delta2 cdel)))) (constantEnvVars right)
         ]
   = MatchFail $ "all envvars in rhs are more general than envvars on right" ++ (show eq)
-}
--    
-- Better to check this only at the end
-- Fail: noncapture constraints (could also be checked later) TODO?
--
--  | Just c <- checkNonCaptureConstraints (delta3 del ++ (delta3 $ constantDelta problem))
--  =  MatchFail $ show c   
--

-- impossible cases:
--    
-- Fail for matching: x = X   ...    
--
  |  (eq:_) <- [eq | eq@((VarLift substx x) :=?=: (MetaVarLift  substy y)) <-  (varLiftEq gam)] 
  =  error $ "Matching impossible case:" ++ show eq
-- Fail: letrec ... = A[s] and A in Delta1 (i.e. A is non-empty)
--
 |  (eq:_) <- [eq | eq@(Letrec _ _ :=?=: vars@(MetaCtxtSubst s ACtxt d _)) <- (expressionEq gam), (MetaCtxt s ACtxt d) `elem` (delta1 del)]
  =  error $ "Matching impossible case:" ++ show eq
--
-- Fail: var x = D[s] = and D in Delta1 (i.e. D is non-empty)
--
 |  (eq:_) <- [eq | eq@(VarL _ :=?=: vars@(MetaCtxtSubst s srt d _))       <- (expressionEq gam), (MetaCtxt s srt d) `elem` (delta1 del)]
  =  error $ "Matching impossible case:" ++ show eq
--
-- Fail: f = D[s], ar(f) = 0 and  D in Delta1 (i.e. D is non-empty)
--
 |  (eq:_) <- [eq | eq@(Fn f [] :=?=: vars@(MetaCtxtSubst s srt d _))      <- (expressionEq gam), (MetaCtxt s  srt d) `elem` (delta1 del)]
  =  error $ "Matching impossible case:" ++ show eq
--
-- Fail: S = t, and S is a proper subterm of t
--
 |  (eq:_) <- [eq | eq@(vars@(MetaExpression _ s) :=?=: t)                <- (expressionEq gam), (properSubTermOf vars t)]
  =  error $ "Matching impossible case:" ++ show eq
 
  -- MatchFail $ show eq 

  
--
-- none of the Fail-Rules is applicable:   
--
 | otherwise = Empty 
-- ********************************************************************************************************************
