-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Matching.Reduction
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
--
-----------------------------------------------------------------------------
-- The Reduction Engine
--
-- main functions of the module:
--   checkAlphaEquivalence
--   alphaRenameProblem
--   alphaRepairProblem
--   tryReduce
--   tryReduceSetOfRules
--   applyConstraintEquation

module LRSX.Matching.Reduction where

import Data.List
import Data.Maybe
import Data.Either
import Data.Char
import Data.Function
import qualified Data.Map.Strict as Map  
import qualified Data.Set as Set
import Debug.Trace
import LRSX.Matching.Matching
import LRSX.Diagrams.Rules
import LRSX.Language.Syntax
import LRSX.Language.Constraints
import LRSX.Util.Util
import LRSX.Interface.Parser
import LRSX.Interface.Lexer
import LRSX.Unification.UnificationProblem
import LRSX.Matching.ConstraintInclusion
import qualified LRSX.Unification.Unification as U
import LRSX.Language.Alpha       
import Control.Parallel.Strategies

mytrace2 a b = b
 

shrinkNCC rp =
 let 
   expr = rpRedex rp
   occVar = getAllVarsFromExpr expr
   nabla_rp = rpConstraintBase rp
   shrink = 
                 splitToNCC $
                  Map.fromList 
                   [(s',dset') | (s',dset) <- [ (s,dset) | (s,dset) <- Map.assocs (splitCapCon $ delta3 $ nabla_rp), 
                     s `Set.member` occVar  || any (\scomp -> scomp `Set.member` occVar) (getAllSCVarsFromSubstitution (getSubstFromSC s))
                     ]  
                   , let dset' = Set.fromList [d | d <- Set.elems dset
                                 , d `Set.member` occVar 
                                   || any (\dcomp -> dcomp `Set.member` occVar) (getAllSCVarsFromSubstitution (getSubstFromSC d))]  
                   , not (Set.null dset')]
 in rp{rpConstraintBase = nabla_rp{delta3 = shrink}          }
-- ==========================================================================================
-- Data type for a reduction problem
data ReductionProblem = 
 ReductionProblem {
   rpRedex          :: Expression       -- the expression to be reduced / the result
  ,rpConstraintBase :: Delta            -- the constraints on the redex
  ,rpAlphaBase      :: [(SyntaxComponent,SubstItem)]         -- an alpha renaming with has been applied to redex
  ,rpRule           :: ConstrainedRule  -- the reduction rule to be applied
  ,rpFunInfo        :: FunInfo          -- info of function symbols
  ,rpFreshNames     :: [String]         -- fresh names
  ,rpLogging        :: String           -- Used for logging
 }
 deriving(Show)

-- ==========================================================================================
-- Checking AlphaEquivalence finally
-- the final check for alpha equivalence:       
fastCheckAlphaEquivalence fresh rp1 rp2 
        -- do a quick check whether the expressions (without alpha renamings) are simlet 
        -- if this does not hold, they cannot be alpha equivalent
        | not ((dropAllSubsExpr $ rpRedex rp1) `similar` (dropAllSubsExpr $ rpRedex rp2)) = ("",False)
        
fastCheckAlphaEquivalence fresh rp1 rp2 =
    case alphaEquiv (map ("AEQC_" ++) fresh)  (rpRedex rp1) (rpRedex rp2) of
        (b,fullrenaming) -> ("",b)
       
checkAlphaEquivalence :: [String]               -- ^ the fresh names
                         -> ReductionProblem    -- ^ reduction problem 1
                         -> ReductionProblem    -- ^ reduction problem 2
                         -> (String,Bool)       -- delivers true, if the rpRedexes of reduction problems 1 and 2 can be made alpha equivalent

checkAlphaEquivalence fresh rp1 rp2 
        -- do a quick check whether the expressions (without alpha renamings) are simlet 
        -- if this does not hold, they cannot be alpha equivalent
        | not ((dropAllSubsExpr $ rpRedex rp1) `similar` (dropAllSubsExpr $ rpRedex rp2)) = ("",False)
        
checkAlphaEquivalence fresh rp1 rp2 
 -- | trace ( show (rpRedex rp1) ) False = undefined
 -- | trace ( show (rpConstraintBase rp1) ) False = undefined
 -- | trace ( show (rpRedex rp2) ) False = undefined
 -- | trace ( show (rpConstraintBase rp2) ) False = undefined
 | otherwise =
    case alphaEquiv (map ("AEQC_" ++) fresh)  (rpRedex rp1) (rpRedex rp2) of
        (b,fullrenaming) -> 
          if b && not ((dropAllSubsExpr $ rpRedex rp1) `similar` (dropAllSubsExpr $ rpRedex rp2))
           then error $ show (dropAllSubsExpr $ rpRedex rp1,dropAllSubsExpr $ rpRedex rp2)
           else  
            if b then 
             -- **************************************************************************************
             -- stuff for check equivalence of NCCs of both Reduction Problems
             let 
              (fullrenaming1orig,fullrenaming2orig) = fullrenaming
              fullrenaming1 = fullrenaming1orig 
              
                           -- `Map.union`
                           --     Map.fromList [(old,news) | (old,news) <- Map.assocs fullrenaming2orig , isNothing $ Map.lookup old fullrenaming1orig]
              fullrenaming2 = fullrenaming2orig 
                           -- `Map.union`
                           --     Map.fromList [(old,news) | (old,news) <- Map.assocs fullrenaming1orig , isNothing  $ Map.lookup old fullrenaming2orig]
                                
              nabla_rp1 = rpConstraintBase rp1
              nabla_rp2 = rpConstraintBase rp2
              expr1 = rpRedex rp1
              expr2 = rpRedex rp2
              occVar1 = getAllVarsFromExpr expr1
              occVar2 = getAllVarsFromExpr expr2
              shrink1 = 
                 splitToNCC $
                  Map.fromList 
                   [(s',dset') | (s',dset) <- [ (s,dset) | (s,dset) <- Map.assocs (splitCapCon $ delta3 $ nabla_rp1), 
                     s `Set.member` occVar1  || any (\scomp -> scomp `Set.member` occVar1) (getAllSCVarsFromSubstitution (getSubstFromSC s))
                     ]  
                   , let dset' = Set.fromList [d | d <- Set.elems dset
                                 , d `Set.member` occVar1 
                                   || any (\dcomp -> dcomp `Set.member` occVar1) (getAllSCVarsFromSubstitution (getSubstFromSC d))]  
                   , not (Set.null dset')]
              shrink2 = 
                 splitToNCC $
                  Map.fromList 
                   [(s',dset') | (s',dset) <- [ (s,dset) | (s,dset) <- Map.assocs (splitCapCon $ delta3 $ nabla_rp2), 
                                s `Set.member` occVar2  || any (\scomp -> scomp `Set.member` occVar2) (getAllSCVarsFromSubstitution (getSubstFromSC s))
                                ]  
                   , let dset' = Set.fromList [d | d <- Set.elems dset
                                 , d `Set.member` occVar2 
                                   || any (\dcomp -> dcomp `Set.member` occVar2) (getAllSCVarsFromSubstitution (getSubstFromSC d))]  
                   , not (Set.null dset')]
                   
              cc1 = (splitFlatToNCC (concatMap getNCCsEnvDVC $  getEnvironments $ rpRedex rp1)) ++ shrink1 -- (delta3 $ nabla_rp1)
              cc2 = (splitFlatToNCC (concatMap getNCCsEnvDVC $  getEnvironments $ rpRedex rp2)) ++ shrink2 -- (delta3 $ nabla_rp2)
              givenNCC1 = applySingleFullRenamingNCC fullrenaming1 $ cc1
              givenNCC2 = applySingleFullRenamingNCC fullrenaming2 $ cc2
              neededNCC1 = applySingleFullRenamingNCC fullrenaming1 cc1 -- shrink1 -- (delta3 $ nabla_rp1)
              neededNCC2 = applySingleFullRenamingNCC fullrenaming2 cc2 -- shrink2 -- (delta3 $ nabla_rp2)
              renamedAlphaBase1 = applySingleFullRenamingAlphaBase fullrenaming1 (rpAlphaBase rp1)
              renamedAlphaBase2 = applySingleFullRenamingAlphaBase fullrenaming2 (rpAlphaBase rp2)
              jointAlphaBase  = renamedAlphaBase1 ++ renamedAlphaBase2
              givenDelta2_1 = [ (Env sub' e)
                              |
                                (Env sub e) <- (delta2 $ nabla_rp1)
                                ,sub' <- applySingleFullRenamingSubst fullrenaming1 sub
                              ]  
              givenDelta2_2 = [ (Env sub' e)
                              |
                                (Env sub e) <- (delta2 $ nabla_rp2)
                                ,sub' <- applySingleFullRenamingSubst fullrenaming2 sub
                              ]
              (test1,r1) =    nccImplies givenDelta2_1 givenNCC2  neededNCC1 jointAlphaBase -- renamedAlphaBase1
              (test2,r2) =    nccImplies givenDelta2_2 givenNCC1  neededNCC2 jointAlphaBase -- renamedAlphaBase2 
             in                                    -- ******************************************************************************************
                                                   -- DS: Okt 16: This check should be done, i.e.
                                                   -- nabla1 ==> nabla2 
                                                   -- and
                                                   -- nabla2 ==> nabla1
                                                   -- however, we first need the common alpha renaming and apply it also to
                                                   -- the constraints, at the moment the alpha-renaming is not really output of the 
                                                   -- alphaEquivalence check
                                                   --
                                                   -- problems:
                                                   --  * the renaming is not a function, but only a function is the position 
                                                   --    is included
                                                   --    however: if alpha-renaming and alpha repair is used before then it is a function
                                                   --  * how to apply alpha-renaming to NCCs (this has been avoided so far)
                                                  
                                                      if isJust r1 then 
{-                                                        trace ("TEST1 EX2=>EX1\n" ++ intercalate "\n\n" 
                                                        ["TESTOUTPUT",test1,
                                                         "REDEX1", show (rpRedex rp1),
                                                         "OCCVAR1", show occVar1,
                                                         "neededNCC1", (show $ splitCapCon $ neededNCC1),
                                                         "fullrenaming1", show fullrenaming1,
                                                         "REDEX2", (show (rpRedex rp2)),
                                                         "givenNCC2", (show $ splitCapCon $ givenNCC2),
                                                         "fullrenaming2", show fullrenaming2,
                                                         "alphaBase", show jointAlphaBase
                                                         ])-} 
                                                           ("",False)
                                                       else if isJust r2 then
{-                                                        trace ("TEST2 EX1=>EX2\n" ++ intercalate "\n\n" 
                                                         ["TESTOUTPUT",test2,
                                                         "REDEX2", show (rpRedex rp2),
                                                         "OCCVAR2", show occVar2,
                                                         "neededNCC2", (show $ splitCapCon $ neededNCC2),
                                                         "fullrenaming2", show fullrenaming2,
                                                         "REDEX1", (show (rpRedex rp1)),
                                                         "givenNCC1", (show $ splitCapCon $ givenNCC1),
                                                         "fullrenaming1", show fullrenaming1,
                                                         "alphaBase", show jointAlphaBase
                                                         ])-} 
                                                           ("",False)
                                                            else 
                                                              ("",b)
            else ("",b) 
                                          
checkSimLetEquivalence fresh rp1 rp2 =
  if  (rpRedex rp1) `simlet` (rpRedex rp2) 
     then 
             -- **************************************************************************************
             -- stuff for check equivalence of NCCs of both Reduction Problems
             let 
              nabla_rp1 = rpConstraintBase rp1
              nabla_rp2 = rpConstraintBase rp2
              expr1 = rpRedex rp1
              expr2 = rpRedex rp2
              occVar1 = getAllVarsFromExpr expr1
              occVar2 = getAllVarsFromExpr expr2
              shrink1 = 
                 splitToNCC $
                  Map.fromList 
                   [(s',dset') | (s',dset) <- [ (s,dset) | (s,dset) <- Map.assocs (splitCapCon $ delta3 $ nabla_rp1), 
                     s `Set.member` occVar1  || any (\scomp -> scomp `Set.member` occVar1) (getAllSCVarsFromSubstitution (getSubstFromSC s))
                     ]  
                   , let dset' = Set.fromList [d | d <- Set.elems dset
                                 , d `Set.member` occVar1 
                                   || any (\dcomp -> dcomp `Set.member` occVar1) (getAllSCVarsFromSubstitution (getSubstFromSC d))]  
                   , not (Set.null dset')]
              shrink2 = 
                 splitToNCC $
                  Map.fromList 
                   [(s',dset') | (s',dset) <- [ (s,dset) | (s,dset) <- Map.assocs (splitCapCon $ delta3 $ nabla_rp2), 
                                s `Set.member` occVar2  || any (\scomp -> scomp `Set.member` occVar2) (getAllSCVarsFromSubstitution (getSubstFromSC s))
                                ]  
                   , let dset' = Set.fromList [d | d <- Set.elems dset
                                 , d `Set.member` occVar2 
                                   || any (\dcomp -> dcomp `Set.member` occVar2) (getAllSCVarsFromSubstitution (getSubstFromSC d))]  
                   , not (Set.null dset')]
                   
              cc1 = (splitFlatToNCC (concatMap getNCCsEnvDVC $  getEnvironments $ rpRedex rp1)) ++ shrink1 -- (delta3 $ nabla_rp1)
              cc2 = (splitFlatToNCC (concatMap getNCCsEnvDVC $  getEnvironments $ rpRedex rp2)) ++ shrink2 -- (delta3 $ nabla_rp2)
              givenNCC1 = cc1
              givenNCC2 = cc2
              neededNCC1 = cc1 
              neededNCC2 = cc2
              renamedAlphaBase1 =  (rpAlphaBase rp1)
              renamedAlphaBase2 =  (rpAlphaBase rp2)
              jointAlphaBase  = renamedAlphaBase1 ++ renamedAlphaBase2
              givenDelta2_1 =  (delta2 $ nabla_rp1)
              givenDelta2_2 = (delta2 $ nabla_rp2)
              (test1,r1) =    nccImplies givenDelta2_1 givenNCC2  neededNCC1 jointAlphaBase -- renamedAlphaBase1
              (test2,r2) =    nccImplies givenDelta2_2 givenNCC1  neededNCC2 jointAlphaBase -- renamedAlphaBase2 
             in                                    -- ******************************************************************************************
                                                   -- DS: Okt 16: This check should be done, i.e.
                                                   -- nabla1 ==> nabla2 
                                                   -- and
                                                   -- nabla2 ==> nabla1
                                                   -- however, we first need the common alpha renaming and apply it also to
                                                   -- the constraints, at the moment the alpha-renaming is not really output of the 
                                                   -- alphaEquivalence check
                                                   --
                                                   -- problems:
                                                   --  * the renaming is not a function, but only a function is the position 
                                                   --    is included
                                                   --    however: if alpha-renaming and alpha repair is used before then it is a function
                                                   --  * how to apply alpha-renaming to NCCs (this has been avoided so far)
                                                  
                                                      if isJust r1 then 
--                                                         trace ("TEST1 EX2=>EX1\n" ++ intercalate "\n\n" 
--                                                         ["TESTOUTPUT",test1,
--                                                          "REDEX1", show (rpRedex rp1),
--                                                          "OCCVAR1", show occVar1,
--                                                          "neededNCC1", (show $ splitCapCon $ neededNCC1),
--                                                          "REDEX2", (show (rpRedex rp2)),
--                                                          "givenNCC2", (show $ splitCapCon $ givenNCC2),
--                                                          "alphaBase", show jointAlphaBase
--                                                          ]) 
                                                         ("",False)
                                                       else if isJust r2 then
--                                                         trace ("TEST2 EX1=>EX2\n" ++ intercalate "\n\n" 
--                                                          ["TESTOUTPUT",test2,
--                                                          "REDEX2", show (rpRedex rp2),
--                                                          "OCCVAR2", show occVar2,
--                                                          "neededNCC2", (show $ splitCapCon $ neededNCC2),
--                                                          "REDEX1", (show (rpRedex rp1)),
--                                                          "givenNCC1", (show $ splitCapCon $ givenNCC1),
--                                                          "alphaBase", show jointAlphaBase
--                                                          ]) 
                                                         ("",False)
                                                            else 
                                                              ("",True)
            else ("",False) 
                               
-- helper functions
zipGroups [] []  = []
zipGroups [] zs2 = [(k,[],g2) | (k,g2)  <- zs2]
zipGroups zs1 [] = [(k,g1,[]) | (k,g1)  <- zs1]
zipGroups zs1@((k1,g1):xs1) zs2@((k2,g2):xs2) 
    | k1 == k2 = (k1,g1,g2):(zipGroups xs1 xs2)
    | k1 <  k2 = (k1,g1,[]):(zipGroups xs1 zs2)
    | k1 >  k2 = (k2,[],g2):(zipGroups zs1 xs2)
zipGroups zs1 zs2 = error ("IN zipGroups" ++ "\n" ++ show zs1 ++ "\n" ++ show zs2)
    
groupNames base = [(a,map snd g) | g@((a,b):_)  <- groupBy (on (==) fst) $ sortBy (on compare fst) [(a,extractName b) | (a,b) <- base ]]
extractName (SubstName _ name)                      = name
extractName (SubstVar _ (VarLift sub name))         = name 
extractName (SubstVar _ (MetaVarLift sub name))     = name
extractName (SubstVar _ (ConstantVarLift sub name)) = name

-- compute possible name - name substitutions 
--
-- getTwoNameSubsts  :: [String] -> [(t, [String], [String])] -> [([String], Map.Map String String, Map.Map String String)]

getTwoNameSubsts fresh []                                 = [(fresh,Map.empty,Map.empty)]
-- easy case: both groups consist of exactly one name:
getTwoNameSubsts (fresh) ((k,[name1],[name2]):rest) = 
    let rek = getTwoNameSubsts fresh rest
    in [ (f, Map.insert name1 f1 m1, Map.insert name2 f1 m2) | (f1:f,m1,m2) <- rek ]

-- easy cases: one group is empty
getTwoNameSubsts fresh ((k,names1,[]):rest)               =
    let 
        rek = getTwoNameSubsts fresh rest
    in [ let (new,fresh') = splitAt (length names1) f
             m1' = Map.union m1  (Map.fromList (zip names1 new))
         in (fresh',m1',m2)  
       | (f,m1,m2) <- rek 
       ]

getTwoNameSubsts fresh ((k,[],names2):rest)               =
    let 
        rek = getTwoNameSubsts fresh rest
    in [ let (new,fresh') = splitAt (length names2) f
             m2' = Map.union m2  (Map.fromList (zip names2 new))
         in (fresh',m1,m2')  
       | (f,m1,m2) <- rek 
       ]

-- difficult case: both are non-empty and contain more than one name    
getTwoNameSubsts fresh ((k,names1,names2):rest)           =
    let 
       rek         = getTwoNameSubsts fresh rest 
    in
     concat 
      [ let 
            l1 = length names1
            l2 = length names2
            (new,fresh') = splitAt (max l1 l2) f
            m1' = if l1 >= l2 then
                    [Map.union m1 (Map.fromList (zip names1 new))]
                   else
                    [Map.union m1 (Map.fromList (zip names1 x)) | x <- nub $ map (take l1) $ permutations new]
            m2' = if l1 >= l2 then
                    [Map.union m2 (Map.fromList (zip names2 x)) | x <- nub $ map (take l2) $ permutations new]
                    else
                    [Map.union m2 (Map.fromList (zip names2 new))]
        in      
--           mytrace2 ("\nXXXX:" ++ show l1 ++":"++ show l2 ++  ":"  ++ show ( length (m1') ) ++ show ( length (m2') )  )
                -- "\n" ++ show (nub $ map (take l1) $ permutations new)
                -- ++
                -- "\n" ++ show (nub $ map (take l2) $ permutations new))
                
         [(fresh',m1new,m2new) | m1new <- m1', m2new <- m2']      
        | (f,m1,m2) <- rek
      ]    
       
changeNamesRP mp rp =
    rp {rpRedex     = changeNamesExpr mp (rpRedex rp)
       ,rpAlphaBase = map (\(a,b) -> (changeNamesSyntaxComponent mp a, changeNamesSubstItem mp b)) (rpAlphaBase rp)
       -- todo: change names somewhere else?
       }
    
    
-- ==========================================================================================
-- perform alpha renaming of a reduction problem 
-- TODO: Alpha-Renaming of Delta4?
alphaRenameProblem :: ReductionProblem -> ReductionProblem
alphaRenameProblem rp 
    | mytrace2 ((last $ show (rpRedex rp)):"ALPHARENAME STARTS")  False = undefined
alphaRenameProblem rp = 
    let frsh = rpFreshNames rp
        -- perform the renaming
        (frsh',expr',renaming) = renameExpr (delta3 $ rpConstraintBase rp) frsh (rpRedex rp)
        -- append the renaming to the alphaBase
        alpha = (rpAlphaBase rp) ++ renaming        
        -- adjust the contraintBase by computing the renamed environment and contexts and
        -- add them to Delta1, Delta2 if the non-renamed variables are in these sets:
        environmentVars        = nub $ environmentVarsInGamma (emptyGamma{expressionEq=[expr' :=?=: expr']})
        contextVars            = nub $ contextVarsInGamma (emptyGamma{expressionEq=[expr' :=?=: expr']})
        constraintBase = (rpConstraintBase rp){delta1=(updateDelta1 contextVars (delta1 $ rpConstraintBase rp))
                                              ,delta2=(updateDelta2 environmentVars (delta2 $ rpConstraintBase rp))}
                                              
    in -- update the problem, and simplify the substitutions by using the constraintBase
      let res = simplifyReductionProblem $ rp{rpFreshNames=frsh',rpRedex=expr',rpAlphaBase = alpha,rpConstraintBase=constraintBase, rpRule=(rpRule rp){ruleName=(False,"alpha",1,False)}}
      in mytrace2 ((last $ show $ rpRedex res) : "ALPHARENAMINGSTOPPS") res
 where
    updateDelta2 [] del2 = del2
    updateDelta2 (r:renamedVars) del2 =
        case lookupBy (\(Env sub name) (Env sub' name') -> name == name' && (substList sub') `isPrefixOf` (substList sub)) r del2 of
            Nothing -> updateDelta2 renamedVars del2
            Just _  -> (updateDelta2 renamedVars del2) ++ [r]
   
    updateDelta1 [] del1 = del1
    updateDelta1 (r:renamedVars) del1 =
        case lookupBy (\x y ->
                        case (x,y) of
                            (ConstantCtxt sub srt name,ConstantCtxt sub' srt' name') -> name == name' && srt == srt' && (substList sub') `isPrefixOf` (substList sub)
                            (MetaCtxt sub srt name,MetaCtxt sub' srt' name') -> name == name' && srt == srt' && (substList sub') `isPrefixOf` (substList sub)
                            _ -> False) r del1 
         of                    
            Nothing -> updateDelta1 renamedVars del1
            Just _  -> (updateDelta1 renamedVars del1) ++ [r]
-- ==========================================================================================    


-- ******************************************************************************************
-- REDUCTION 
-- ******************************************************************************************

-- ==========================================================================================
-- try to perform the reduction of a reduction problem for a set rules
-- already ready for parallelization

tryReduceSetOfRules reductionProblem rulesSet =
 let 
   res =
    concat (
         parMap 
           rseq 
           (\r -> tryReduce reductionProblem{rpRule = r})
           rulesSet
          )
  in  
--       trace ("\n Computed Successors: Rules: " ++ (intercalate "," $ map (showRuleName . ruleName . rpRule) res)
--             ++ " found " ++ show (length res) ++ " successors\n" 
      --      ++ (intercalate "\n * " $ map (show . rpRedex) res
--             )
      --     )
      res  

-- ==========================================================================================
-- try to perform the reduction of a reduction problem:
tryReduce :: ReductionProblem ->  [ReductionProblem]
-- tryReduce rp
--    | trace ("start tryReduce" ++ (showRuleName. ruleName . rpRule $ rp)) False = undefined
tryReduce rp@(ReductionProblem{rpRedex=expr,rpFreshNames=frsh,rpRule=crule,rpConstraintBase=cdel,rpAlphaBase=alp,rpFunInfo=finfo,rpLogging=rplog}) = 
  let  
   (freshRule,fresh') = makeFresh crule frsh
   (lhs :==>: rhs)    = rule freshRule
   igamma             = Gamma {varLiftEq     = []
                                         ,expressionEq  = [lhs :=?=: expr]
                                         ,environmentEq = []
                                         ,bindingEq     = []}
   del                = Delta {delta1=nonEmptyCtxts freshRule
                                         ,delta2=nonEmptyEnvs freshRule
                                         ,delta3=nonCapCon freshRule ++ computeConstraints lhs 
                                         ,delta4=emptyGamma
                                         }                              
   matchingProblem    = Problem
                        {solution      = []
                        ,gamma         = igamma
                        ,delta         = del
                        ,constantDelta = cdel
                        ,functions     = finfo
                        ,fresh         = fresh'
                        ,inputGamma    = igamma
                        ,inputDelta    = del
                        ,alphaBase     = alp
                        ,mpLogging     = []
                        ,expandAllEnvs = True
                        ,expandAllContexts = True -- irrelevant for matching at the moment
                        ,expandDelta4Constraints=True
                        }
   tableau            = matching matchingProblem
   solutions          = map fromSolved $ solvedAndOpenLeaves tableau
   results            = [rp{ rpRule=freshRule
                           , rpRedex=(constantify $ applySolution (solution sol) rhs)
                           , rpConstraintBase=cdel{delta3=fastLazyNub ((delta3 cdel) ++ (constantify $ applySolution (solution sol) (delta3 (delta sol))))}
                           , rpFreshNames=(fresh sol)
                           , rpLogging = rplog ++ (mpLogging sol) ++
                                            (unlines ["REDUCTIONLOG: START"
                                                    , "REDUCTIONLOG: successfully applied reduction " ++ showRuleName (ruleName crule) ++ "  to expression "
                                                    , "REDUCTIONLOG: " ++ show expr
                                                    , "REDUCTIONLOG: resulting in"
                                                    , "REDUCTIONLOG: " ++ show (constantify $ applySolution (solution sol) rhs)
                                            ])
                            } | sol1 <- solutions, let sol = sol1{constantDelta=(constantDelta sol1){
                                                                   delta1 = (delta1 (constantDelta sol1)) ++ (constantify $ delta1 $ (delta sol1))
                                                                  ,delta2 = (delta2 (constantDelta sol1)) ++ (delta2 $ (delta sol1))
                                                                  -- ,delta3 = (delta3 (constantDelta sol1)) ++ (constantify $ delta3 $ (delta sol1))
                                                     }}
                        ]
   noDuplicates = fastLazyNubOn rpRedex results
   -- noDuplicates = nubBy (\x y -> snd $ checkAlphaEquivalence ["_AEQCHECK" ++ show i | i <- [1..]] x y) results
  in -- trace ((showRuleName . ruleName . rpRule $ rp) ++ ":" ++  (show $ length noDuplicates)) $
 -- (trace ((show $ length results) ++ " > " ++ (show $ length noDuplicates))) $
--        if ("lbeta" `isInfixOf` (showRuleName. ruleName . rpRule $ rp)) && null (noDuplicates) then 
--          trace ("\nREDEX:\t" ++ show (rpRedex rp) ++ "\n" ++ (unlines $ map  show $ allLeaves tableau)) noDuplicates
--         else 
        -- if (not . null $ noDuplicates) then trace ("success" ++ (showRuleName. ruleName . rpRule $ rp)) noDuplicates else
    map shrinkNCC      results -- noDuplicates

          
-- ========================================          
-- apply simplification of alpha-renamings
          
simplifyReductionProblem rp = 
 rp {rpRedex = simplifyExpr (rpAlphaBase rp) (splitCapCon $ delta3 $ rpConstraintBase rp) (rpRedex rp)}

-- =========================================
-- some helper functions                                     
fromSolved (Solved p) = p
fromSolved other = error ("(fromSolved): " ++ show other)

                                     
-- applyConstraintEquation tries to apply equations from the  environment constraints in Delta4             
applyConstraintEquation reductionProblem =
  let     
   solConstr1 = [(constantify l,constantify r) | (r :=?=: l) <- environmentEq $ delta4 $ rpConstraintBase reductionProblem ]
   solConstr2 = [(constantify l,constantify r) | (l :=?=: r) <- environmentEq $ delta4 $ rpConstraintBase reductionProblem ]
 
   expr1      = constantify $ foldr (\(l,r) e -> substEnvConstraint r l e) (rpRedex reductionProblem)  solConstr1
   expr2      = constantify $ foldr (\(l,r) e -> substEnvConstraint r l e) (rpRedex reductionProblem)  solConstr2
   res1       =   (if constantify (rpRedex reductionProblem) == expr1 then [] else  [reductionProblem{rpRedex=expr1,rpRule=identity{ruleName=(False,"D4ConstraintLToR",1,False)}}])
   res2       = (if constantify (rpRedex reductionProblem) == expr2 then [] else  [reductionProblem{rpRedex=expr2,rpRule=identity{ruleName=(False,"D4ConstraintRToL",1,False)}}])
  in res1 ++ res2
  

-- interface for starting the repair procedure for the alpha renaming
--
  
alphaRepairProblem rp = 
    let (f,newAlphaBase,e) = startAlphaRepair (rpAlphaBase rp)  (rpFreshNames rp) (rpRedex rp) (rpRedex rp)
        oldAlphaBase = rpAlphaBase rp
        getFirsts [] _ = []
        getFirsts ((comp,ren):rest) have
          | Just _ <- lookup comp have = getFirsts rest have
          | otherwise = (comp,ren):(getFirsts rest ((comp,ren):have))
        oldAlphaBase' = getFirsts oldAlphaBase []
        subs = [(comp,ren,lookupAll comp newAlphaBase)  | (comp,ren) <- oldAlphaBase']
        splitMulti [] = [[]] 
        splitMulti (el@(comp,ren,[]):rest)    = splitMulti rest
        splitMulti (el@(comp,ren,[new]):rest) = [(toSubs comp ren new):recurse | recurse <- splitMulti rest] 
        splitMulti (el@(comp,ren,new):rest)   = [(toSubs comp ren n):recurse |  recurse <- splitMulti rest,n <- new] 
        allSubs = splitMulti subs  
        lookupAll key list = [v | (k,v) <- list, k == key]
        cdel = (rpConstraintBase rp)
        cdel3 = delta3 cdel
        newcdel3 =
           splitToNCC $ Map.fromList
             [
              (fst (head gs), Set.fromList (map snd gs))
              |
               gs <-
               groupBy ((==) `on` fst)        
               ( sortBy (compare `on` fst)
                [(applySubsSCList s t,applySubsSCList s d) 
                    | (t,dset) <- Map.assocs (splitCapCon cdel3)
                    , d <- Set.elems dset, s <- allSubs
                ]
                )
              ]
        -- newcdel3s = fastLazyNub $ [(Left (applySubLRList s t),applySubLRList s d) | (Left t,d) <- cdel3, s <- allSubs]
                                 -- ++ [(Right (applySubLRListEnv s t),applySubLRList s d) | (Right t,d) <- cdel3, s <- allSubs]
        newcdel = cdel{delta3=cdel3 ++ newcdel3} -- okt 16 cdel3 ++ 
        -- replaceInExprBy :: SyntaxComponent -> String -> String -> Expression -> Expression

        
    in  (if  not (null ([() | (a,b) <- unfoldSplittedNCCs (splitCapCon (delta3 newcdel)), a==b]))
         then     
          trace ("\nAP: BASE BEFORE: " ++ show (oldAlphaBase)) $
            trace ("\nAP: BASE AFTER: " ++ show newAlphaBase) $
              trace ("\nAP: ConstraintBase before:" ++ show (splitCapCon $ cdel3)) $
               trace ("\nAP: Subs:" ++ show subs) $
                  trace ("\nAP: ConstraintBase after:" ++ show (splitCapCon $ newcdel3)) $ id
           else id)     
                rp{rpFreshNames=f,rpAlphaBase=newAlphaBase++oldAlphaBase,rpRedex=e,rpConstraintBase=newcdel,rpRule=identity{ruleName=(False,"alphaRepair",1,False)}}
      

applySubSC (Left (v0,v1,v2)) (SCVar v) = SCVar (replaceTriplesVar v0 v1 v2 v)
applySubSC (Right (sc,n1,n2)) (SCVar v) = SCVar (replaceInVarBy sc n2 n1 v)

applySubSC (Left (v0,v1,v2)) (SCExpr e) = SCExpr (replaceTriples v0 v1 v2 e)
applySubSC (Right (sc,n1,n2)) (SCExpr e) = SCExpr (replaceInExprBy sc n2 n1 e)

applySubSC (Left (v0,v1,v2)) (SCCtxt e) = SCCtxt (replaceTriples v0 v1 v2 e)
applySubSC (Right (sc,n1,n2)) (SCCtxt e) = SCCtxt (replaceInExprBy sc n2 n1 e)

applySubSC (Left (v0,v1,v2)) (SCMetaEnv e) = 
 let env=empty{metaEnvVars=[e]}
     env'=replaceTriplesEnv v0 v1 v2 env
     [e'] = metaEnvVars env'
 in (SCMetaEnv e')
applySubSC (Right (sc,n1,n2)) (SCMetaEnv e) = 
 let env=empty{metaEnvVars=[e]}
     env'=replaceInEnvBy sc n2 n1 env
     [e'] = metaEnvVars env'
 in (SCMetaEnv e')

applySubSC (Left (v0,v1,v2)) (SCConstantEnv e) = 
 let env=empty{constantEnvVars=[e]}
     env'=replaceTriplesEnv v0 v1 v2 env
     [e'] = constantEnvVars env'
 in (SCConstantEnv e')
 
applySubSC (Right (sc,n1,n2)) (SCConstantEnv e) = 
 let env=empty{constantEnvVars=[e]}
     env'=replaceInEnvBy sc n2 n1 env
     [e'] = constantEnvVars env'
 in (SCConstantEnv e')

applySubSC (Left (v0,v1,v2)) (SCMetaChain sub typ name) = 
   let env= empty{metaChains=[MetaChain 
                            sub 
                            typ 
                            name 
                            (ConstantVarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
       env' = replaceTriplesEnv v0 v1 v2 env
       [MetaChain sub1 typ1 name1 _ _] = metaChains env' 
    in SCMetaChain sub1 typ1 name1
   
applySubSC (Right (sc,n1,n2))  (SCMetaChain sub typ name) = 
   let env= empty{metaChains=[MetaChain 
                            sub 
                            typ 
                            name 
                            (ConstantVarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
       env' = replaceInEnvBy sc n2 n1 env
       [MetaChain sub1 typ1 name1 _ _] = metaChains env' 
    in SCMetaChain sub1 typ1 name1
      
      
applySubSC (Left (v0,v1,v2)) (SCConstantChain sub typ name) = 
   let env= empty{constantChains=[ConstantChain 
                            sub 
                            typ 
                            name 
                            (ConstantVarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
       env' = replaceTriplesEnv v0 v1 v2 env
       [ConstantChain sub1 typ1 name1 _ _] = constantChains env' 
    in SCConstantChain sub1 typ1 name1
   
applySubSC (Right (sc,n1,n2))  (SCConstantChain sub typ name) = 
   let env= empty{constantChains=[ConstantChain 
                            sub 
                            typ 
                            name 
                            (ConstantVarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
       env' = replaceInEnvBy sc n2 n1 env
       [ConstantChain sub1 typ1 name1 _ _] = constantChains env' 
    in SCConstantChain sub1 typ1 name1
      
applySubSC x y = error $ "in applySubSC:" ++ show (x,y)      
applySubsSCList [] sc = sc
applySubsSCList (x:xs) sc = applySubsSCList xs (applySubSC x sc)      
      
{-
applySubLR (Left (v0,v1,v2)) e = replaceTriples v0 v1 v2 e
applySubLR (Right (sc,n1,n2)) e = replaceInExprBy sc n2 n1 e
applySubLRList (x:xs) e 
 | trace (show ("RP:",x,e)) False = undefined
applySubLRList (x:xs) e = applySubLRList xs (applySubLR x e)      
applySubLRList [] e = e 
applySubLREnv (Left (v0,v1,v2)) e = replaceTriplesEnv v0 v1 v2 e
applySubLREnv (Right (sc,n1,n2)) e = replaceInEnvBy sc n2 n1  e
applySubLRListEnv (x:xs) e 
 | trace (show ("RP:",x,e)) False = undefined
applySubLRListEnv (x:xs) e = applySubLRListEnv xs (applySubLREnv x e)      
applySubLRListEnv [] e = e 
-}

toSubs sc@(SCVar v0)  (SubstVar v1 v1') (SubstVar v2 v2')
   | v0 == v1 && v0 == v2 = Left (v0,v1',v2')
   
-- toSubs sc@(SCVar (VarLift sub v))  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
--   | sc == comp1 && sc == comp2 = (sc,name1,name2)

toSubs sc@(SCExpr e0)  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
   | sc == comp1 && sc == comp2 = Right (sc,name1,name2)
toSubs sc@(SCCtxt e0)  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
   | sc == comp1 && sc == comp2 = Right (sc,name1,name2)
toSubs sc@(SCMetaEnv e0)  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
   | sc == comp1 && sc == comp2 = Right (sc,name1,name2)
toSubs sc@(SCMetaChain sub t s)  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
   | sc == comp1 && sc == comp2 = Right (sc,name1,name2)
toSubs sc@(SCConstantEnv e0)  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
   | sc == comp1 && sc == comp2 = Right (sc,name1,name2)
toSubs sc@(SCConstantChain sub t s)  (SubstName (None comp1) name1) (SubstName (None comp2) name2)
   | sc == comp1 && sc == comp2 = Right (sc,name1,name2)

toSubs a b c = error ("IN toSubs:" ++ show (a,b,c))   
   
-- toSubs (SCVar (VarLift
  -- |  SCExpr Expression
  -- |  SCCtxt Expression
  -- |  SCMetaEnv EnvVar
  -- |  SCMetaChain Substitution ChainType String  
  -- |  SCConstantEnv EnvVar
  -- |  SCConstantChain Substitution ChainType String  
      

