-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Unification.Unification 
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- The unification algorithm for higher-order 
-- expressions with recursive bindings and 
-- special environments for small step reductions
--
-----------------------------------------------------------------------------
module LRSX.Unification.Unification where
import Data.Function
import Data.List
import Data.Maybe
--
import Debug.Trace
--
import LRSX.Language.Syntax
import LRSX.Util.Util
import LRSX.Unification.UnificationProblem
import qualified Data.Set as Set
import qualified Data.Map as Map
import LRSX.Matching.ConstraintInclusion
-- ====================================================================================================================
-- UNIFICATION
-- ====================================================================================================================



-- ********************************************************************************************************************
-- The unification function performs the whole unification and builds a tableau
-- ********************************************************************************************************************

unification :: Problem -> Tableau
unification problem 
  | Just err <- checkGammaUnifInp (gamma problem,
                                   delta1 (delta problem),
                                   delta2 (delta problem)) = error $ "unification: input check fails:" ++ show err ++ "\nInputProblem\n" ++ show problem

unification problem = expand problem 
 where
  expand pb 
   | solved (gamma pb)             
   , Empty <- unifyStep  (heur (expandDelta4Constraints pb) (expandAllContexts pb) (expandAllEnvs pb) ) (pb) 
   , Empty <- unifyStep [applyFailRulesDelta4] pb = 
       let sol             = composeAll (solution pb)
           instExpr        = applySolution sol (gamma problem)
           instDelta       = applySolution sol (delta3 $ delta pb)    
           capcon          = checkNCCs (delta2 $ delta pb) instDelta
                             -- checkNonCaptureConstraints instDelta
       in case checkWeakDVCGamma instExpr of 
           Nothing -> 
                 if isNothing capcon -- check capture constraints
                 then -- consistency check for the solution: 
                     if (any (\(x :=?=: y) -> x /= y) (expressionEq instExpr)) && (nullGamma $ delta4 $ delta pb) 
                       then error $ "\nPANIC: Solution is not a solution\n" 
                             ++  "\nInstExpr:\n" ++  (unlines $ map (\(x :=?=: y) -> "\n" ++ show x ++ "\n=?=" ++ "\n" ++ show y) (expressionEq instExpr))
                             ++  "\nInstExpr:\n" ++ show (instExpr)
                             ++  "\nComposedSol\n" ++ show sol
                             ++  "\nCurrentState\n"
                             ++  "\nSol\n" ++ show (solution pb)
                             ++  "\nDelta\n" ++ show (delta pb)                             
                             ++  "\nInput\n" ++ show (gamma problem)
                             
                       else  
                       Solved pb{solution=sol} -- (instExpr, sol, delta pb,pb)
              else Node pb [Fail $ "capture constraint fail" ++ (show (fromJust capcon))]
           Just err -> Node pb [Fail $ "Weak DVC failure:" ++ err ++ " " ++ (show (expressionEq instExpr))]
  expand pb 
      | solved (gamma pb) =
            case unifyStep [applyFailRulesDelta4] pb of
                Some problems -> Node pb (map expand problems)
                UnifFail err  -> Node pb [Fail err]
                Empty -> case unifyStep (heur (expandDelta4Constraints pb) (expandAllContexts pb) (expandAllEnvs pb)) pb of 
                           Some problems -> Node pb (map expand problems)
                           UnifFail err  -> Node pb [Fail err]
                           Empty          -> error $ "in expand " ++ show (pb) -- show $ Node (solution pb, gamma pb, delta pb) []
      | otherwise =
                case unifyStep (heur (expandDelta4Constraints pb) (expandAllContexts pb) (expandAllEnvs pb)) pb of 
                    Some problems -> Node pb (map expand problems)
                    UnifFail err  -> Node pb [Fail err]
                    Empty         -> error $ "in expand " ++ show (pb) -- show $ Node (solution pb, gamma pb, delta pb) []
     
  heur flagExpandDelta4 flagAllContexts flagAllEnvs = 
         [applyFailRules
         ,applyFailRules . commuteProblem
         ,applyFirstOrderRules
         ,applyFirstOrderRules . commuteProblem
         ,applyContextRules 
         ,applyContextRules . commuteProblem 
         ,applyEnvRules
         ,applyEnvRules . commuteProblem
         ,applyGuessRules
         ,applyGuessRules . commuteProblem
         ]
         ++ (if flagExpandDelta4 
                then [expandDelta4ConstraintsIfPossible
                     ,expandDelta4ConstraintsIfPossible . commuteDelta4
                     ]
                else [])
         ++ (if flagAllContexts 
                then [applyFullGuessRules]   -- guess all context vars
                else [])
         ++ (if flagAllEnvs 
                then [applyFullGuessRulesEnv]   -- guess all context vars
                else [])
         
  unifyStep [] pb = Empty
  unifyStep (uniffn:xs) pb =
    case uniffn pb of 
      Empty -> unifyStep xs pb
      Some [] -> unifyStep xs pb
      Some ps -> Some ps
      UnifFail err -> UnifFail err

-- ********************************************************************************************************************
-- The tableau of the unification algorithm

data Tableau = Node Problem [Tableau] 
             | Fail String
             | Solved Problem -- (Gamma,Solution,Delta,Problem) -- Solved nodes: Unified equations, solution, constraints on environments, final problem

             
instance Show Tableau where
 show t = showTableau [1] t
  where
   showTableau ind (Node p xs) = 
     indent ind ++ "Open: " ++ show (solution p,gamma p, delta p) ++ "\n" ++ (concat [showTableau (if length xs > 1 then ind ++ [i] else ind) x | (i,x) <- zip [1..] xs])
   showTableau ind (Fail a)   = indent ind ++ "RULE: " ++ show a ++ "\n"
   showTableau ind (Solved pb) = 
       let sol = solution pb
           gam = applySolution sol (inputGamma pb)
           del = delta pb
       in indent ind ++ "Solved: Unified Equations:" ++ show (gam,sol,del) ++ "\n"
     
   indent ind = let px = (concat $ intersperse "." (map show ind)) 
                in  px ++ (replicate (mx - (length px)) ' '     ) ++ ": "
   mx = 10
   branchDepth i (Node _ [x]) = branchDepth i x
   branchDepth i (Node _ xs) =  let a = [branchDepth (i ++ [v]) x  | (v,x) <- zip [1..] xs]
                                in if null a then i else maximumBy (compare `on` (length . show)) a   
   branchDepth i _ = i 
    
-- helper function for the tableau:

-- compute solved and open leaves 
solvedAndOpenLeaves (Node a []) = [Node a []]
solvedAndOpenLeaves (Node _ cs) = concatMap solvedAndOpenLeaves cs
solvedAndOpenLeaves (Fail x)    = []
solvedAndOpenLeaves (Solved x)  = [Solved x] 

-- compute the solved leaves only    
solvedLeaves (Node _ cs) = concatMap solvedLeaves cs
solvedLeaves (Fail _)    = []
solvedLeaves (Solved x)  = [x] 

-- compute all leaves 
allLeaves (Node a []) = [Node a []]
allLeaves (Node _ cs) = concatMap allLeaves cs
allLeaves (Fail x)    = [Fail x]
allLeaves (Solved x)  = [Solved x] 

-- extract Delta
getDeltaFromSolved (Solved pb) = delta pb 
-- ********************************************************************************************************************


-- ********************************************************************************************************************
-- Result of one step of the unification algorithm 
data UnifResult = UnifFail String  -- unification fails
                | Empty            -- no unification rule applicable
                | Some [Problem]   -- new problems after one step
 deriving(Show)
-- ********************************************************************************************************************

  
-- ********************************************************************************************************************
-- Fail rules
-- ********************************************************************************************************************
-- the function applyFailRules checks whether a fail rule is applicable to the problem,
-- if yes, then UnifFail is delivered as result, otherwithe Empty is delivered.
-- the extra fail rule, which is only applied to the final state, and checks whether
-- the constraints in Delta4 are unsolvable is not included in applyFailRules, it is
-- available via applyFailRulesDelta4

applyFailRules :: Problem -> UnifResult
applyFailRules problem@(Problem{solution=sol,gamma=gam,delta=del})
--
--  Fail-Rules for variable-equations:
--  RULE: (Sol,Gamma cup {x = y}, Delta) --> Fail if x/= y
--
  |  (eq:_) <- [eq | eq@(VarLift s1 x :=?=: VarLift s2 y) <- (varLiftEq gam), x /= y]
    =  UnifFail $ show eq 
--    
--  Fail-Rules for expression-equations:
--
 | (eq:_) <- (  
--  RULE: (Sol,Gamma cup {f r1 ... rn = var y}, Delta) --> Fail
                [eq | eq@((Fn name args) :=?=: VarL _) <- (expressionEq gam)]
--  RULE: (Sol,Gamma cup {f r1 ... rn = letrec ....}, Delta) --> Fail
             ++ [eq | eq@((Fn _ _) :=?=: (Letrec _ _)) <- (expressionEq gam)]
--  RULE: (Sol,Gamma cup {f r1 ... rn = g r1' ... rn'}, Delta) --> Fail
             ++ [eq | eq@((Fn name _) :=?=: (Fn name' _)) <- (expressionEq gam), name /= name']
--  RULE: (Sol,Gamma cup {var x = letrec ...}, Delta) --> Fail
             ++ [eq | eq@(VarL  _ :=?=: (Letrec _ _)) <- (expressionEq gam)]
--  RULE: (Sol,Gamma cup {S = t ...}, Delta) --> Fail, if S is proper subterm of t and t /= D[t'] for some context variable D
             ++ [eq | eq@(vars@(MetaExpression s0 s) :=?=: t) <- (expressionEq gam)
                    , (properSubTermOf vars t)
                    , not (isMetaCtxtSubst t)]
--  RULE: (Sol,Gamma cup {S = D[t] ...}, Delta) --> Fail, if S is subterm of t and D in Delta1 (i.e. D is non-empty)
             ++ [eq | eq@(vars@(MetaExpression s0 s) :=?=: MetaCtxtSubst s1 srt1 d1 e1)   <- (expressionEq gam)
                    , (subTermOf vars e1)
                    , (MetaCtxt emptySubstitution srt1 d1) `elem` (delta1 del)]
--  RULE: (Sol,Gamma cup {A[s] = letrec ...}) --> Fail if A in Delta1 (i.e. A is non-empty)
             ++ [eq | eq@(vars@(MetaCtxtSubst s0 ACtxt d _) :=?=: Letrec _ _) <- (expressionEq gam)
                    , (MetaCtxt emptySubstitution ACtxt d) `elem` (delta1 del)]
--  RULE: (Sol,Gamma cup {D[s] = var x ...}) --> Fail if D in Delta1 (i.e. D is non-empty)
             ++ [eq | eq@(vars@(MetaCtxtSubst s0 srt d _) :=?=: VarL _) <- (expressionEq gam)
                ,(MetaCtxt s0 srt d) `elem` (delta1 del)]
--  RULE: (Sol,Gamma cup {D[s] = f}) --> Fail if ar(f) = 0 and D in Delta1 (i.e. D is non-empty)
             ++ [eq | eq@(vars@(MetaCtxtSubst s0 srt d _) :=?=: VarL _) <- (expressionEq gam)
                ,(MetaCtxt s0 srt d) `elem` (delta1 del)]
--  RULE: (Sol,Gamma cup {D[s] = f}) --> Fail if ar(f) = 0 and D in Delta1 (i.e. D is non-empty)
             ++ [eq | eq@(vars@(MetaCtxtSubst s0 srt d _) :=?=: Fn f []) <- (expressionEq gam)
                ,(MetaCtxt s0 srt d) `elem` (delta1 del)]
--  RULE: (Sol,Gamma cup {D[s] = f r1 ... rn}) --> Fail 
--        if D is an A or T-context, 
--        and the operator arity of of f is (n1,...nm) where all ni > 0,
--        and D in Delta1 (i.e. D is non-empty)
             ++ [eq | eq@(vars@(MetaCtxtSubst s0 srt d _) :=?=: Fn f args) <- (expressionEq gam)
                , (MetaCtxt s0 srt d) `elem` (delta1 del)
                , srt < Ctxt
                , let bari:_ = [bindarity | (name,arity,strictpos,bindarity) <- functions problem, f == name]
                , all (> 0) bari
                , not (null bari)
                ]
--  RULE: (Sol,Gamma cup {D[s] = f r1 ... rn}) --> Fail 
--        if D is an A or T or C-context, 
--        and the operator arity of of f is (n1,...nm) where all ni = Var  (in the program this means they are -1)
--        and D in Delta1 (i.e. D is non-empty)
             ++ [eq | eq@(vars@(MetaCtxtSubst s0 srt d _) :=?=: Fn f args) <- (expressionEq gam)
                ,(MetaCtxt s0 srt d) `elem` (delta1 del)
                , srt <= Ctxt
                , let bari:_ = [bindarity | (name,arity,strictpos,bindarity) <- functions problem, f == name]
                , all (< 0) bari
                , not (null bari)
                ]
          )
    = UnifFail $ show eq
-- RULE: (Sol,Gamma cup {A[s] = f s1 ... sn}) --> Fail if
--  f has no strict positions, and A in Delta1 (A is not empty)
 | Just eq@((ctxt@(MetaCtxtSubst s0 ACtxt cname s) :=?=: (Fn fname arguments)),rest)
     <- splitBy (\(l:=?=:r) -> case (l,r) of 
                                (MetaCtxtSubst s0 ACtxt cname s, Fn fname _) ->  (MetaCtxt s0 ACtxt cname) `elem` (delta1 del)
                                _ -> False) (expressionEq gam),
   Just [] <- (getStrictPos problem fname)  
   = UnifFail $ show eq
--    
--  Fail-Rules for environment-equations:
--
--  RULE:(Sol,Gamma cup env :=: EE[z,s]; EE'[z,s'];env') -> Fail since same z occurs twice as binder in the same environment
  | (eq:_) <- (  [eq | eq@(l :=?=: r) <- environmentEq gam
                  , not (null [ ee | (MetaChain sub _ ee z s) <- metaChains r
                  ,(MetaChain sub' _ ee' z' s') <- metaChains r
                  , ee /= ee'
                  , z == z'])
                ]   
--  RULE: (Sol,Gamma cup env :=: EE[z,s]; z=s';env') -> Fail since same z occurs twice as binder in the same environment
             ++ [eq | eq@(l :=?=: r) <- environmentEq gam
                  , (MetaChain sub _ ee z s) <- metaChains r
                  , (y:=_) <- (bindings r)
                  , y == z
                ]
--  RULE: (Sol,Gamma cup {env = {}} -> Fail, if env is non-empty, i.e. 
--   env=b;env'  or env=EE[z,s];env'  or env=E;env' with E in Delta2
             ++ [eq | eq@(left :=?=: right) <- (environmentEq gam)
                       ,isEmpty right
                       ,  (not (null (bindings left )) 
                        || not (null (metaChains left)) 
                        || not (null [ v | v <- metaEnvVars left
                                     , v `elem` (delta2 del)] ))
                ]
             )
       = UnifFail $ show eq
--
-- Otherwise, no Fail-rule is applicable:
--       
 | otherwise = Empty 

-- 
-- Fail-Rule for Delta4, checked only once at the end of unification
--    First, check whether there are constraint equations 
--    in Delta4, then check if the constraints are solvable
--    (done by the function solvableChains) 
--

applyFailRulesDelta4 :: Problem -> UnifResult
applyFailRulesDelta4 problem@(Problem{solution=sol,gamma=gam,delta=del})
 | eqs <- (environmentEq $ delta4 del)
 , not (null eqs)
  , leftEI <- [v | (left :=?=: _) <- eqs, v <- metaEnvVars left]
  , deltaEI <- [ei | ei <- leftEI, ei `elem` (delta2 del)]
  , Just err <- solvableChains problem
    = UnifFail $ "unsolvable E1,...,En = EE[z,s]" 
-- Otherwise, the constraints are solvable and thus the fail-rule does not deliver a RULE:
 | otherwise = Empty 
-- ********************************************************************************************************************


 
-- ********************************************************************************************************************
 
-- ********************************************************************************************************************
-- First-Order Cases
-- ********************************************************************************************************************
-- applyFirstOrderRules tries to apply a first-order unification rule to the problem
-- if no rule is applicable, then Empty is delivered, otherwise the result is Some [problem] which is
-- the subsequent unification problem (no branching is required, and thus the result is a singleton) 
--
applyFirstOrderRules :: Problem -> UnifResult
applyFirstOrderRules problem@(Problem{solution=sol,gamma=gam,delta=del})
--
--  RULE: (Sol,Gamma cup {X = Y}, Delta) --> (Sol circ {X |-> Y}, Gamma[Y/X], Delta [Y/X])
--         where X is a Meta-Variable, Y can be meta or concrete
--
 | Just ((varx@(MetaVarLift _ x) :=?=: vary),rest) 
     <-  splitBy (\(l:=?=:r) -> case l of 
                                 MetaVarLift _ x -> True
                                 _                 -> False) (varLiftEq gam)
    = let sol' = (MapVarLift $ varx :|-> vary):sol
          gam' = subst vary varx (gam{varLiftEq = rest}) 
          del' = subst vary varx del
      in Some [problem{solution=sol',gamma=gam',delta=del'}]
--
--  RULE: (Sol,Gamma cup {x = x}, Delta) --> (Sol, Gamma, Delta)
--         where x is a concrete variable
--
  |  Just (eq@((VarLift _ x) :=?=: (VarLift _ y)),rest) 
       <-  splitBy (\(l :=?=: r) 
                        -> case (l,r) of 
                            (VarLift _ x,VarLift _ y) -> x == y
                            _                         -> False) 
                   (varLiftEq gam)
    = Some [problem{gamma=gam{varLiftEq = rest}}]
--    
--
--  RULE: (Sol,Gamma cup {S = s}, Delta) --> (Sol circ {S |-> s}, Gamma[s/S], Delta [s/S])
--         if S is not a proper subterm of s
--         not that proper is meant syntactically, 
--         i.e. S is a proper subterm of D[S] 
--         even if D may be the empty context
--         (the first-order rule is then not applicable, first the context variable has to be guesses empty/non-empty)                                        
--  
 | Just ((vars@(MetaExpression _ s) :=?=: t),rest) 
     <- splitBy (\(l:=?=:r) 
                    -> case l of 
                        MetaExpression _  s -> not (properSubTermOf l r )   
                        _                   -> False                    ) 
                (expressionEq gam)
    = let sol' = (MapExpr $ vars :|-> t):sol
          gam' = subst t vars (gam{expressionEq = rest}) 
          del' = subst t vars  del
      in Some [problem{solution=sol',gamma=gam',delta=del'} ]

--
--  RULE: (Sol,Gamma cup {f vect(X1).s1...vect(Xn).sn = f vect(Y1).t1...vect(Yn).tn}, Delta) --> (Sol, Gamma cup {si=ti} cup {X11=Y11,...,Xnmn=Ynmn}, Delta)
--        where vect(Zi) = Zi1...Zimi, note that all binders are processed in one blow (to avoid higher-order-expression-equations)
--
 | Just (((Fn name args) :=?=: Fn name' args'),rest) 
     <-   splitBy (\(l:=?=:r) 
                      -> case (l,r) of 
                          (Fn name _, Fn name' _) -> name == name'
                          _                       -> False         ) 
                  (expressionEq gam)
    = let (vs,es) = -- computes the new variable equations and the new expression equations
            let (vareqs,expreqs) = unzip (zipWith compute args args') -- 
            in (concat vareqs,concat expreqs)
          -- compute variable equations and expression equations  
          -- case for variable positions
          compute (VarPos v1) (VarPos v2)
            = let vareq = v1 :=?=: v2
              in ([vareq],[])
          -- case for higher-order binder:
          compute b1@(Binder v1 e1) b2@(Binder v2 e2)
            | length v1 == length v2 = 
                let vareq = zipWith (:=?=:) v1 v2
                    expreq = [e1 :=?=: e2]
                in (vareq,expreq)
            | otherwise = error ("found binders with different arity: "  ++ show b1 ++ " and " ++ show b2)
          -- error handling:
          compute b1 b2 =
            error ("Binder found at variable position with: " ++ show b1 ++ " and " ++ show b2)
          -- update Gamma:
          gam' = (gam{expressionEq = es ++ rest, varLiftEq = vs ++ (varLiftEq gam)}) 
      in Some [problem{gamma=gam'} ]
--  
--  RULE: (Sol,Gamma cup {var X = var Y}, Delta) --> (Sol, Gamma cup {X=Y}, Delta)
--  (for all cases of X,Y (Meta or concrete)
--  
  | Just ((VarL varx :=?=: VarL vary),rest)
       <-  splitBy (\(l:=?=:r) -> 
                        case (l,r) of 
                             (VarL _, VarL _) -> True
                             _                -> False) 
                   (expressionEq gam)
     = let gam' = gam{ varLiftEq = (varx:=?=:vary):(varLiftEq gam)
                      , expressionEq = rest} 
       in Some [problem{gamma=gam'}]
--  
--  RULE: (Sol,Gamma cup {x.s = y.t}, Delta) --> (Sol, Gamma cup {x=y,s=t}), Delta 
--  (for all cases of X,Y (Meta or concrete)
-- 
  | ((var1 := e1) :=?=: (var2 := e2):rest) <-  (bindingEq gam)
     = let gam' = gam{ varLiftEq = (var1:=?=:var2):(varLiftEq gam)
                      , expressionEq = (e1 :=?=: e2):(expressionEq gam)
                      , bindingEq = rest} 
      in Some [problem{gamma=gam'}]
--  
--  RULE: (Sol,Gamma cup {letrec env1 in s1 = letrec env2 in s2}, Delta) --> (Sol, Gamma cup {env1=env2,s1=s2}, Delta)
--  
  | Just ((Letrec env1 s1 :=?=: Letrec env2 s2),rest) 
      <- splitBy (\(l:=?=:r) -> 
                        case (l,r) of 
                         (Letrec _ _, Letrec _ _) -> True
                         _                        -> False) 
                 (expressionEq gam)
     = let gam' = gam{environmentEq = (env1:=?=:env2):(environmentEq gam)
                     ,expressionEq = (s1 :=?=: s2):rest} 
       in Some [problem{gamma=gam'}] 
--      
-- RULE: (Sol,Gamma cup {{} = {}},Delta) -> (Sol,Gamma,Delta)
--     
  | Just (env1:=?=: env2,rest) 
      <- splitBy  
            (\(l :=?=: r) -> isEmpty l && isEmpty r) 
            (environmentEq gam)
     = Some [problem{gamma=gam{environmentEq = rest}} ]
--      
-- no first-order rule applicable:
--
  | otherwise = Empty
-- ********************************************************************************************************************

-- ********************************************************************************************************************
-- Rules for Contexts 
-- ********************************************************************************************************************
-- The rules for contexts work on expression equations, where at least on side of the
-- equation is a context. However, for guessing a context variable to be empty or non-empty
-- it suffices that the context variable occurs somewhere in Delta 
-- the rules are split into two functions:
-- applyGuessRules to guess contexts to be empty/non-empty
-- applyContextRules for all other rules
-- (reason: one can play around with different heuristics for applying the guess rules)
--

applyFullGuessRules problem@(Problem{solution=sol,gamma=gam,delta=del})
 | q@((MetaCtxt _ srt name):r) <- [x | x@(MetaCtxt sub srt name) <- (contextVarsInSol $ composeAll sol)
                                   , not (x `elem` (delta1 del) || (MetaCtxt emptySubstitution srt name) `elem` (delta1 del)) ]
                                   
    = let problem1 = 
                let oldDel_1 = delta1 del
                    newDel_1 = (MetaCtxt emptySubstitution  srt name):oldDel_1
                in  problem{delta = del{delta1 = newDel_1}}
          problem2 = 
                let sol' = (MapExpr $ MetaCtxt emptySubstitution  srt name :|-> hole):sol
                    gam' = subst hole (MetaCtxt emptySubstitution srt name) gam
                    del' = subst hole (MetaCtxt emptySubstitution srt name) del
               in problem{solution=sol',gamma=gam', delta=del'}
      in Some [problem1,problem2]
-- no guess rule applicable
 | otherwise = Empty
--      
-- Guessing context-variables as empty/non-empty
-- 
-- if there is a context-variable D in Gamma on Top, but not in Delta1,
-- then branch into two problems: either D is guessed as empty and replaced by the empty context,
-- or D is guessed as non-empty and thus added to Delta1
-- (Sol,Gamma cup D[s]=t,Delta) -> (Sol[[.]/D],(Gamma cup {s=t})[[.]/D],Delta[[.]/D]])  | (Sol,Gamma,Delta1 cup {D}) 
--  if D occurs in Gamma and D not in Delta1
applyGuessRules problem@(Problem{solution=sol,gamma=gam,delta=del})
 | Just ((ctxt@(MetaCtxtSubst sub srt name s) :=?=: _),rest)
     <- splitBy (\(l:=?=:r) 
                    -> case (l,r) of 
                        (MetaCtxtSubst sub srt cname s, _) -> not ((MetaCtxt sub srt cname) `elem` (delta1 del))
                        _ -> False) 
                (expressionEq gam)
    = let problem1 = 
                let oldDel_1 = delta1 del
                    newDel_1 = (MetaCtxt emptySubstitution  srt name):oldDel_1
                in  problem{delta = del{delta1 = newDel_1}}
          problem2 = 
                let sol' = (MapExpr $ MetaCtxt emptySubstitution  srt name :|-> hole):sol
                    gam' = subst hole (MetaCtxt emptySubstitution srt name) gam
                    del' = subst hole (MetaCtxt emptySubstitution srt name) del
               in problem{solution=sol',gamma=gam', delta=del'}
      in Some [problem1,problem2]
-- no guess rule applicable
 | otherwise = Empty
--
--
--     
applyContextRules problem@(Problem{solution=sol,gamma=gam,delta=del})
--
-- RULE: (Sol,Gamma cup {A[s] = f s1 ... sn},Delta) -> (Sol cup {A |-> f s1 ... A' ... sn}, Gamma cup {si =?= A'[s]}, Delta[f s1 .... A' .... sn/A]) 
--  if A in Delta1, and for each strict position of f_i 
--
--
 | Just ((ctxt@(MetaCtxtSubst sub ACtxt cname s) :=?=: (Fn fname arguments)),rest)
     <- splitBy (\(l:=?=:r) 
                    -> case (l,r) of 
                        (MetaCtxtSubst sub ACtxt cname s, Fn fname _) ->  (MetaCtxt sub ACtxt cname) `elem` (delta1 del)
                        _ -> False) 
                (expressionEq gam)
   -- are there positions with operator arity = 0 ?
 , not (null [v | Binder v si <- arguments, null v])
   -- are there strict positions?
 , Just (positions@(_:_)) <- (getStrictPos problem fname)  
    = Some
       [let (name_d':fresh') = fresh problem
            a'_s = MetaCtxtSubst emptySubstitution  ACtxt ("A"++name_d') s 
            a' = MetaCtxt emptySubstitution ACtxt ("A"++name_d') 
            a = MetaCtxt emptySubstitution ACtxt cname
            gam' = gam{expressionEq = (expr :=?=: a'_s):rest}  
            del' = subst rhs a del 
            rhs = Fn fname (before ++ [Binder [] a'] ++ after)
            sol' = (MapExpr $ a :|-> rhs):sol
        in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
               | (pos,(before,(Binder _ expr),after)) <- zip [1..] (splits2 arguments)
               , pos `elem` positions
        ]
      
--
-- RULE: (Sol,Gamma cup {T[s] = f s1 .. sn},Delta) -> (Sol cup {T |-> f s1 ... T' ... sn},Gamma cup {s=T'[s_i]},Delta[f s1 ... T' ...sn/T]) , 
--       if T in Delta1, and for all positions of f with operator arity = 0
--
 | Just ((ctxt@(MetaCtxtSubst sub TCtxt cname s) :=?=: (Fn fname (args))),rest)
     <- splitBy (\(l:=?=:r) 
                    -> case (l,r) of
                        (MetaCtxtSubst sub TCtxt cname s, Fn fname args) -> 
                                   (MetaCtxt sub TCtxt cname) `elem` (delta1 del)
                        _ -> False) 
                (expressionEq gam)
 , not (null [v | Binder v si <- args, null v])
    = Some
      [let
         (name_t':fresh') = fresh problem
         t'_s = MetaCtxtSubst emptySubstitution TCtxt  ("T"++name_t') s 
         t' = MetaCtxt emptySubstitution TCtxt ("T"++name_t')  
         t = MetaCtxt emptySubstitution TCtxt cname
         gam' = gam{expressionEq = (si :=?=: t'_s):rest}
         del' = subst rhs t del 
         rhs = Fn fname (before ++ [Binder [] t'] ++ after)
         sol' = (MapExpr $ t :|-> rhs):sol
       in 
         problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
       | (before,(Binder v si),after) <- splits2 args, null v  -- only positions without binders!
       ]
--
-- RULE: (Sol,Gamma cup {C[s] = f s1 ..x1...xm.si .. sn},Delta) -> (Sol cup {C |-> f s1 ... x1...xm.C' ... sn},Gamma cup {s=C'[s_i]},Delta[f s1 ... x1...xm.C' ...sn/T]) , 
--       if C in Delta1, and for all positions of f with operator arity >= 0
--
 | Just ((ctxt@(MetaCtxtSubst sub Ctxt cname s) :=?=: (Fn fname (args))),rest)
     <- splitBy (\(l:=?=:r) 
                    -> case (l,r) of
                        (MetaCtxtSubst sub Ctxt cname s, Fn fname args) -> (MetaCtxt sub Ctxt cname) `elem` (delta1 del)
                        _ -> False)
                (expressionEq gam)
 , not (null [v | Binder v si <- args])                                 
    = Some
      [let
         (name_c':fresh') = fresh problem
         c'_s = MetaCtxtSubst emptySubstitution Ctxt ("C"++name_c') s 
         c' = MetaCtxt emptySubstitution Ctxt ("C"++name_c')  
         c = MetaCtxt emptySubstitution Ctxt cname
         gam' = gam{expressionEq = (si :=?=: c'_s):rest}
         del' = subst rhs c del 
         rhs = Fn fname (before ++ [Binder v c'] ++ after)
         sol' = (MapExpr $ c :|-> rhs):sol
       in 
         problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
       | (before,(Binder v si),after) <- splits2 args       -- hole may be below binders!
       ]

--
-- RULE: (Sol,Gamma cup {D[s] = letrec env in s'},Delta) 
--        ->    (Sol cup {D |-> letrec env in D'},Gamma cup {D'[s]=s'},Delta[letrec env in D'/D])
--           |  (Sol cup {D |-> letrec x'=D'; E in s'},Gamma cup {x'=D'[s];E  = env},Delta[letrec x'=D'; E in s'/D])
--       if D in Delta1, and D is of sort T or C 
--

 | Just ((ctxt@(MetaCtxtSubst sub srt cname s) :=?=: (Letrec env s')),rest) 
     <- splitBy (\(l:=?=:r) -> 
                    case (l,r) of
                      (MetaCtxtSubst sub srt cname s, Letrec env s') -> (MetaCtxt sub srt cname) `elem` (delta1 del)
                                                                        && (srt `elem` [TCtxt,Ctxt])
                      _                                              -> False) 
                (expressionEq gam)
    = Some [
      let
       (name_d':fresh') = fresh problem
       d'_s = MetaCtxtSubst emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') s 
       d' = MetaCtxt emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') 
       d = MetaCtxt emptySubstitution srt cname
       gam' = gam{expressionEq = (d'_s :=?=: s'):rest}
       del' = subst rhs d $ del
       rhs = Letrec env d'
       sol' = (MapExpr $ d :|-> rhs ):sol
      in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
      ,
      let
       (name_d':name_x':name_e:fresh') = fresh problem
       x' = MetaVarLift emptySubstitution  $ "X" ++ name_x'
       d'_s = MetaCtxtSubst emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') s 
       d' = MetaCtxt emptySubstitution srt ((if srt == Ctxt then "C" else "T")++name_d') 
       d = MetaCtxt emptySubstitution srt cname
       e = "E" ++ name_e
       gam' = gam{expressionEq = rest, environmentEq =  ((binding (x' := d'_s)){metaEnvVars=[Env emptySubstitution e]} :=?=: env):(environmentEq gam)}
       del' = subst rhs d del
       rhs = Letrec ((binding (x' := d')){metaEnvVars=[Env emptySubstitution e]}) s'
       sol' = (MapExpr $ d :|-> rhs):sol
      in problem{delta=del',gamma=gam',solution=sol',fresh=fresh'}
      ]
--
-- prefix rule and forking rule, all in one case:
--
-- the prefix rule is:
--
-- RULE: (Sol,Gamma cup {D1[s]  = D2[s']}, Delta)  
--       -->  (Sol cup {D1 |-> D3, D2 -> D3D4}, Gamma cup {s=D4[s']}, Delta [D3/D1, D3D4/D2]) and sort D3 = sort D1; sort D4 = sort D2
--          | (Sol cup {D1 |-> D3D4, D2 -> D3}, Gamma cup {D4[s]=s'}, Delta [D3D4/D1, D3/D2]) and sort D3 = sort D1; sort D4 = sort D1
-- if D1,D2 in Delta1, D1 <= D2
 
 | Just ((ctxt@(MetaCtxtSubst sub1 srtd1@srtd d1@d s) :=?=: (MetaCtxtSubst sub2 srtd2@srtd' d2@d' s')),rest)
     <- splitBy (\(l:=?=:r) 
                   -> case (l,r) of 
                       (MetaCtxtSubst sub1 srtd1 d1 s, MetaCtxtSubst sub2 srtd2 d2 s') ->  
                                (MetaCtxt sub1 srtd1 d1) `elem` (delta1 del)
                             && (MetaCtxt sub2 srtd2 d2) `elem` (delta1 del)
                             && srtd1 <= srtd2
                       _    -> False) 
                (expressionEq gam)
    = Some $ [
      let
       (name_d3:fresh') = fresh problem
       (name_d4:fresh'') = fresh'
       srtd3  = srtd1
       srtd4  = srtd2
       d3 =  (case srtd3 of 
                   Ctxt -> "C"
                   TCtxt -> "T"
                   ACtxt -> "A") ++ name_d3
       d4 = (case srtd4 of 
                   Ctxt -> "C"
                   TCtxt -> "T"
                   ACtxt -> "A") ++ name_d4
       d2_rhs = MetaCtxtSubst 
                    emptySubstitution 
                    srtd3 
                    d3 
                    (MetaCtxt emptySubstitution srtd4 d4)  -- D3[D4]
       d1_rhs = (MetaCtxt emptySubstitution srtd3 d3)                                           -- D3
       sol' = (MapExpr $ MetaCtxt emptySubstitution  srtd2 d2 :|-> d2_rhs)
                :(MapExpr $ MetaCtxt emptySubstitution  srtd1 d1 :|-> d1_rhs)
                :sol  -- D2 -> D3[D4]; D1 -> D3
       gam' = gam{expressionEq = (s :=?=: d4sp):rest}
       d4sp = MetaCtxtSubst emptySubstitution srtd4 d4 s'
       del' = subst 
                d1_rhs 
                (MetaCtxt emptySubstitution srtd1 d1) 
                (subst 
                    d2_rhs 
                    (MetaCtxt emptySubstitution srtd2 d2) 
                    del
                )
       del'' = del'{delta1=delta1 del' ++ [(MetaCtxt emptySubstitution srtd3 d3)]}        
      in problem{delta=del'',gamma=gam',solution=sol',fresh=fresh''} 
     ,
      let
       (name_d3:fresh') = fresh problem
       (name_d4:fresh'') = fresh'
       srtd3  = srtd1
       srtd4  = srtd1
       d3 =  (case srtd3 of 
                   Ctxt -> "C"
                   TCtxt -> "T"
                   ACtxt -> "A") ++ name_d3
       d4 = (case srtd4 of 
                   Ctxt -> "C"
                   TCtxt -> "T"
                   ACtxt -> "A") ++ name_d4
       d2_rhs = (MetaCtxt emptySubstitution srtd3 d3)     -- D3
       d1_rhs =  MetaCtxtSubst 
                    emptySubstitution 
                    srtd3 
                    d3 
                    (MetaCtxt emptySubstitution srtd4 d4) -- D3[D4]
       sol' = (MapExpr $ MetaCtxt emptySubstitution  srtd2 d2 :|-> d2_rhs)
                :(MapExpr $ MetaCtxt emptySubstitution  srtd1 d1 :|-> d1_rhs)
                :sol -- D2 -> D3; D1 ->  D3[D4]
       gam' = gam{expressionEq = (s' :=?=: d4s):rest}
       d4s = MetaCtxtSubst emptySubstitution srtd4 d4 s
       del' = subst 
                d1_rhs 
                (MetaCtxt emptySubstitution srtd1 d1) 
                (subst 
                    d2_rhs 
                    (MetaCtxt emptySubstitution srtd2 d2) 
                    del
                ) -- D3D4/D1 
       del'' = del'{delta1=delta1 del' ++ [(MetaCtxt emptySubstitution srtd3 d3)]}        
      in problem{delta=del'',gamma=gam',solution=sol',fresh=fresh''}    
     ]
--
-- the forking rule is for functions symbols:
--
-- (Sol,Gamma cup {D1[s] = D2[t]),Delta) --> |_Possibilites (Sol circ sigma, Gamma, Delta sigma)
--  where D1 <= D2, 
--        sigma = D1 |-> D1'[f S1 ... vectX.D1''    ... vectY.D2''[t]...]
--                D2 |-> D1'[f S1 ....vectX.D1''[s] ... vectY.D2''   ....]
--                                    -------------     -------------
--                                         i                j
--  and sort of D1' = sort of D1
--      sort of D1'' = sort of D1
--      sort of D2'' = sort of D2
--  if D1 is an ACtxt then hole i must be in strict postion      
--  and if D2 is also an ACtxt then hole j must also be in another strict postion -- 
--  and if D2 is another context, then j = 2...
--  if D1 is a TCtxt or a Ctxt, then (i,j) can be all positions, but only for C it can be below a binder
     ++
     [ 
       let
        d1 = d
        d2 = d'
        srtd1 = srtd
        srtd2 = srtd'
        (name_d1':name_d1'':name_d2'':fresh') = fresh problem
        fresh'' = drop (arity + (sum bindarity) + (length (filter (<0) bindarity))) fresh'
        xi_s = go (take (arity + (sum bindarity) + (length (filter (<0) bindarity))) fresh') bindarity
        go names [] = []
        go names (bar:bs) 
         | bar >= 0 =        
          let bv = map (MetaVarLift emptySubstitution) (map ("X" ++) $ take bar names)
              (metaS:rest) = (drop bar names)
          in (Binder bv (MetaExpression emptySubstitution ("S" ++ metaS))):(go rest bs)
        go names (bar:bs) 
         | bar == (-1) =        
          let (bv:_) = map (MetaVarLift emptySubstitution) (map ("X" ++) $ take 1 names)
              (metaS:rest) = (drop 1 names)
          in (VarPos bv):(go rest bs)
        d1' =  (repr srtd1') ++ name_d1'
        d1'' = (repr srtd1'') ++ name_d1''
        d2'' = (repr srtd2'') ++ name_d2''
        srtd1' = srtd1
        srtd1'' = srtd1
        srtd2'' = srtd2
        rhsd1 = MetaCtxtSubst emptySubstitution
                  srtd1' 
                  d1' 
                  (let (Binder v _)  = xi_s !!! (i-1)
                       (Binder v' _) = xi_s !!! (j-1)
                   in  (Fn 
                            name 
                            ( (replacePos i (Binder v (MetaCtxt emptySubstitution srtd1'' d1''))) 
                                $ (replacePos j (Binder v' (MetaCtxtSubst emptySubstitution srtd2'' d2'' s'))) xi_s)
                        )
                  )
        rhsd2 = MetaCtxtSubst  emptySubstitution
                  srtd1' 
                  d1' 
                  (let (Binder v _)  = xi_s !!! (i-1)
                       (Binder v' _) = xi_s !!! (j-1)
                   in  (Fn 
                            name 
                            ( (replacePos i (Binder v (MetaCtxtSubst emptySubstitution srtd1'' d1'' s)) 
                                $ (replacePos j (Binder v' (MetaCtxt emptySubstitution srtd2'' d2'')) xi_s ))
                            )
                        )
                  )
        sol' = (MapExpr $ (MetaCtxt emptySubstitution srtd1 d1 :|-> rhsd1))
                :(MapExpr $ (MetaCtxt emptySubstitution srtd2 d2 :|-> rhsd2))
                :sol
        gam' = gam{expressionEq = rest}
        del' = subst rhsd2 (MetaCtxt emptySubstitution srtd2 d2 ) 
                $ subst rhsd1 (MetaCtxt emptySubstitution srtd1 d1) 
                $ del
       in  problem{gamma=gam',solution=sol',delta=del',fresh=fresh''}
        | (name,arity,strictpos,bindarity) <- functions problem
          , arity > 1
          , (i,j) <- if srtd == ACtxt then  
                      if srtd' == ACtxt then [(i,j) | i <- strictpos, j <- strictpos, i/=j]
                        else [(i,j) | i <- strictpos
                                    , (j,jb) <- zip [1..arity] bindarity
                                    , i /= j 
                                    , if srtd' == TCtxt then jb == 0 else jb >= 0
                             ]  
                      else [ (i,j) | (i,ib) <- zip [1..arity] bindarity
                                        , (j,jb) <- zip [1..arity] bindarity 
                                        , i /= j
                                        , if srtd == TCtxt then ib == 0 else ib >= 0
                                        , if srtd' == TCtxt then jb == 0 else jb >= 0
                                ]
          
       ] 
     ++
-- the forking rule is for letrec
--
-- (Sol,Gamma cup {D1[s] = D2[t]),Delta) --> |_Possibilites (Sol circ sigma, Gamma, Delta sigma)
--  where D1 <= D2, and D1 not an A-Ctxt
--   case 1:
--        sigma = D1 |-> D1'[letrec Y=D1'';E in D2''[t]
--                D2 |-> D1'[letrec Y=D1''[s] in D2 
--   case 2:
--        sigma = D1 |-> D1'[letrec Y=D2''[t];E in D1'']
--                D2 |-> D1'[letrec Y=D2'';E in D1''[s]]
--   case 3:
--        sigma = D1 |-> D1'[letrec Y1=D1''; Y2=D2''[t];E in S]
--                D2 |-> D1'[letrec Y1=D1''[s]; Y2=D2'';E in S]
--
--  and sort of D1' = sort of D1
--      sort of D1'' = sort of D1
--      sort of D2'' = sort of D2
     if srtd == ACtxt then [] else
     [ 
--     case 1:
       let
        d1 = d
        d2 = d'
        srtd1 = srtd
        srtd2 = srtd'
        (name_d1':name_d1'':name_d2'':name_e:name_xn:fresh') = fresh problem
        d1'  = (repr srtd1') ++ name_d1'
        d1'' = (repr srtd1'') ++ name_d1''
        d2'' = (repr srtd2'') ++ name_d2''
        srtd1' = srtd1
        srtd1'' = srtd1
        srtd2'' = srtd2
        e       = "E" ++ name_e
        x      = MetaVarLift emptySubstitution $ "X" ++ name_xn
        rhsd1 = MetaCtxtSubst 
                    emptySubstitution 
                    srtd1' 
                    d1' 
                    (Letrec 
                        (binding (x := MetaCtxt emptySubstitution srtd1'' d1'')){metaEnvVars = [Env emptySubstitution e]} 
                        (MetaCtxtSubst emptySubstitution srtd2'' d2'' s')
                    )
        rhsd2 = MetaCtxtSubst 
                    emptySubstitution 
                    srtd1' 
                    d1' 
                    (Letrec 
                        (binding (x := MetaCtxtSubst emptySubstitution srtd1'' d1'' s)){metaEnvVars = [Env emptySubstitution e]} 
                        (MetaCtxt emptySubstitution srtd2'' d2'')
                    )
        sol' = (MapExpr $ MetaCtxt emptySubstitution srtd1 d1 :|-> rhsd1):(MapExpr $ MetaCtxt emptySubstitution srtd2 d2 :|-> rhsd2):sol
        gam' = gam{expressionEq = rest}
        del' = subst rhsd2 (MetaCtxt emptySubstitution srtd2 d2) $ subst rhsd1 (MetaCtxt emptySubstitution srtd1 d1) $ del
        in  problem{gamma=gam',solution=sol',delta=del',fresh=fresh'}
       ,
--     case 2:
       let
        d1 = d
        d2 = d'
        (name_d1':name_d1'':name_d2'':name_e:name_xn:fresh') = fresh problem
        d1'  = (repr srtd1') ++ name_d1'
        d1'' = (repr srtd1'') ++ name_d1''
        d2'' = (repr srtd2'') ++ name_d2''
        srtd1 = srtd
        srtd2 = srtd'
        srtd1' = srtd1
        srtd1'' = srtd1
        srtd2'' = srtd2
        e       = "E" ++ name_e
        x      = MetaVarLift emptySubstitution $ "X" ++ name_xn
        rhsd1 = MetaCtxtSubst 
                    emptySubstitution 
                    srtd1' 
                    d1' 
                    (Letrec 
                        (binding (x := MetaCtxtSubst emptySubstitution srtd2'' d2'' s')){metaEnvVars = [Env emptySubstitution e]} 
                        (MetaCtxt emptySubstitution srtd1'' d1'' )
                    )
        rhsd2 = MetaCtxtSubst 
                    emptySubstitution 
                    srtd1' 
                    d1' 
                    (Letrec 
                        (binding (x := MetaCtxt emptySubstitution srtd2'' d2'' )){metaEnvVars = [Env emptySubstitution e]} 
                        (MetaCtxtSubst emptySubstitution srtd1'' d1'' s)
                    )
        sol' = (MapExpr $ (MetaCtxt emptySubstitution srtd1 d1 :|-> rhsd1))
                :(MapExpr $ MetaCtxt emptySubstitution srtd2 d2 :|-> rhsd2)
                :sol
        gam' = gam{expressionEq = rest}
        del' = subst rhsd2 (MetaCtxt emptySubstitution srtd2 d2)
                $ subst rhsd1 (MetaCtxt emptySubstitution srtd1 d1) 
                $ del 
       in  problem{gamma=gam',solution=sol',delta=del',fresh=fresh'}
     ,       
--     case 3:
       let
        d1 = d
        d2 = d'
        (name_d1':name_d1'':name_d2'':name_e:name_x1:name_x2:name_s:fresh') = fresh problem
        d1'  = (repr srtd1') ++ name_d1'
        d1'' = (repr srtd1'') ++ name_d1''
        d2'' = (repr srtd2'') ++ name_d2''
        srtd1 = srtd
        srtd2 = srtd'
        srtd1' = srtd1
        srtd1'' = srtd1
        srtd2'' = srtd2
        e       = "E" ++ name_e
        x1      = MetaVarLift emptySubstitution $ "X" ++ name_x1
        x2      = MetaVarLift emptySubstitution $ "X" ++ name_x2
        mS      = MetaExpression  emptySubstitution $ "S" ++ name_s
        rhsd1 = MetaCtxtSubst 
                    emptySubstitution 
                    srtd1' 
                    d1' 
                    (Letrec 
                        (empty{bindings = [(x1 := MetaCtxt emptySubstitution  srtd1'' d1'')
                                          ,x2 := MetaCtxtSubst emptySubstitution srtd2'' d2'' s']
                              ,metaChains=[]
                              ,metaEnvVars=[Env emptySubstitution e]}) 
                        mS
                    )
        rhsd2 = MetaCtxtSubst 
                    emptySubstitution 
                    srtd1' 
                    d1' 
                    (Letrec (empty{bindings = [(x1 := MetaCtxtSubst emptySubstitution srtd1''  d1'' s)
                                              ,x2 := MetaCtxt emptySubstitution srtd2'' d2'']
                                  ,metaChains=[]
                                  ,metaEnvVars=[Env emptySubstitution e]}) 
                                  mS)
        sol' =  (MapExpr $ MetaCtxt emptySubstitution srtd1 d1 :|-> rhsd1)
               :(MapExpr $ MetaCtxt emptySubstitution srtd2 d2 :|-> rhsd2)
               :sol
        gam' = gam{expressionEq = rest}
        del' = subst rhsd2 (MetaCtxt emptySubstitution srtd2 d2) 
               $ subst rhsd1 (MetaCtxt emptySubstitution srtd1 d1) 
               $ del
       in  problem{gamma=gam',solution=sol',delta=del',fresh=fresh'}
     ]      
--  
-- no context rule applicable:
--
 | otherwise = Empty
-- ********************************************************************************************************************


-- ********************************************************************************************************************
-- Rules for Environments
-- ********************************************************************************************************************
-- The rules for enviroments  work on environment equations
--
applyEnvRules :: Problem -> UnifResult
applyEnvRules problem@(Problem{solution=sol,gamma=gam,delta=del})
--  
--  RULE: (Sol,Gamma cup {E = env}, Delta) --> (Sol cup {E |-> env}, Gamma,Delta[env/E])
--        if env is nonempty 
--  (env-E1)
--  
  | ((left :=?=: r),rest):_ 
      <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
                               ,isMetaEnv left
                               ,isSingleton  (metaEnvVars left)
                               ,let [e] = metaEnvVars left
                               ,e `notElem` (delta2 del) 
                                || (not (null (bindings r))) 
                                || (not (null (metaChains r))) 
                                || (not (null [v | v <- metaEnvVars r, v `elem` (delta2 del)])) 
                               ]
    ,[Env sub e] <- (metaEnvVars left)
     = 
       let sol' = (MapEnv $ Env sub e :|-> r): sol
           gam' = gam{environmentEq = rest}
           del' = substEnv r e del 
       in Some [problem{solution=sol',gamma=gam',delta=del'}] 
--  
--  RULE: (Sol,Gamma cup {E1...En = {}}, Delta) --> (Sol cup {E1 |-> {},...,En |-> {}}, Gamma,Delta[{}/E1,...,{}/En])
--  if for all i: Ei not in Delta2 (i.e. all Ei must be empty) 
-- (env-E2)
--  
  | ((left :=?=: r),rest):_ 
      <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
                               ,null (bindings left)
                               ,null (metaChains left)
                               ,isEmpty r
                               ,not (null $ metaEnvVars left)
                               ,and [e `notElem` (delta2 del) | e <- metaEnvVars left] 
                               ]
    ,es <- (metaEnvVars left)
     = 
       let sol' = [(MapEnv $ ei :|-> empty) | ei <- es ] ++ sol
           gam' = gam{environmentEq = rest}
           del' = foldr (\e d -> substEmpty e d) del [q | Env sub q <- es]
       in Some [problem{solution=sol',gamma=gam',delta=del'}]        
--  
--  RULE: (Sol,Gamma cup {E = E1,...,En}}, Delta) -> |_forall i: (Sol, Gamma cup {E=E1,...,En}, Delta2 cup {Ei})
--   if E in Delta2, if for all i: Ei not in Delta2 (i.e. guess some Ei to be non-empty)
-- (env-E3) 
--  
  | ((left :=?=: r),rest):_ 
      <-  [splitting | splitting@(left :=?=: r,rest) <- splits (environmentEq gam)
                               ,isMetaEnv left
                               ,isSingleton (metaEnvVars left)
                               ,let [e] = metaEnvVars left
                               ,null (bindings r)
                               ,null (metaChains r)
                               ,not (null (metaEnvVars r))
                               ,let es = metaEnvVars r
                               ,e `elem` (delta2 del)
                               ,and [ei `notElem` (delta2 del) | ei <- es] 
                               ]
    ,[e] <- (metaEnvVars left)
    ,es <- metaEnvVars r
     =  Some [problem{delta = del{delta2 = ei:(delta2 del)}} | ei <- es] 

--  
--  RULE: (Sol,Gamma cup {b1;env1 = env2}, Delta --> |_Possibilites (Sol, Gamma cup ..., Delta)
--   several cases to match the binding b1 against something in env2 (details see below)
--  (env-b)
--  
  | (((left :=?=: right),rest):_) <- [(left :=?=: right,rest) | splitting@(left :=?=: right,rest) <- splits (environmentEq gam)
                                                              , hasBinding left
                                                              , not (null (bindings right) && null (metaEnvVars right) && null (metaChains right))]
  ,((b1,bindings_left'):_) <- splits (bindings left)
     =
      Some $
--  RULE: (Sol,Gamma cup {b1;env1 = env2}, Delta --> |forall b2;env2' = env2 (Sol, Gamma cup {env1=env2; b1=b2} ..., Delta)
              [ -- for each binding b2 from right
                let gam'   = gam {bindingEq = (b1 :=?=: b2):bindingEq gam
                                 ,environmentEq = (left' :=?=: right'):rest}
                    left'  = left{bindings = bindings_left'}
                    right' = right{bindings = bindings_right'}
                in problem{gamma=gam'}                           
                | (b2,bindings_right') <- splits (bindings right)
              ]
--  RULE: (Sol,Gamma cup {b1;env1 = env2}, Delta --> |forall E;env2' = env2 (Sol cup {E -> b1;E'}, Gamma cup {env1 = E';env2'}, Delta[b1;E'/E])
              ++
              [ -- for each metaEnv E from right
                let (name_e':fresh') = fresh problem
                    e' = Env emptySubstitution ("E" ++ name_e')
                    b1e' = empty{bindings=[b1],metaChains=[],metaEnvVars=[e']}
                    sol' = (MapEnv $  Env emptySubstitution e :|-> b1e'):sol
                    gam' = gam {environmentEq = (left' :=?=: right'):rest}
                    left' = left{bindings = bindings_left'}
                    right' = right{metaEnvVars = e':metaEnvVars_right'}
                    del' = substEnv b1e' e del
                in problem{solution=sol',fresh=fresh',gamma=gam',delta=del'}                           
                | (Env sub e,metaEnvVars_right') <-  splits (metaEnvVars right)
              ]
              ++
--  RULE: (Sol,Gamma cup {b1;env1 = env2}, Delta --> |forall EE[z,s];env2' = env2 
--     case 1:
--                (Sol cup {EE[z,s] |-> z=A[s]}, 
--                 Gamma cup {env1 = env2'; b1 = z=A[s]}, 
--                 Delta[z=A[s]/EE[z,s])
--     case 2:
--             |  (Sol cup {EE[z,s] |-> z=A[var X]; EE'[X,s]}, 
--                 Gamma cup {env1 = EE'[X,s];env2'; b1 = z=A[var X]}, 
--                 Delta[z=A[var X],EE'[X,s]/EE[z,s]])
--     case 3:
--             |  (Sol cup {EE[z,s] |-> EE1[z,var Y];Y=A[var X];EE2[X,s]}, 
--                 Gamma cup {env1 = EE1[z,var Y];EE2[X,s];env2'; b1 = z=A[var X]}, 
--                 Delta[EE1[z,var Y];Y=A[var X];EE2[X,s]/EE[z,s]])
--     case 4:
--             |  (Sol cup {EE[z,s] |-> EE1[z,var Y];Y=A[s]}, 
--                 Gamma cup {env1 = EE1[z,var Y];env2'; b1 = Y=A[s]}, 
--                 Delta[EE1[z,var Y];Y=A[s]/EE[z,s]])

-- analogous for VV-Chains where the A-Ctxt is not generated, and VV-Chains are splitted into VV-Chains

              (concat [ -- for each metaChain EE from right
--     case 1:
                 [let 
                      sol' = (MapCV (MetaChain emptySubstitution typ ee z s :|-> sigma_right)):sol
                      (name_ctxtname:fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      z_A_s = \z s ->
                                case typ of 
                                 AChain -> (z := MetaCtxtSubst emptySubstitution ACtxt ctxtname s)
                                 VarChain -> (z := s)
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (b1 :=?=: z_A_s z s):(bindingEq gam)
                                }
                      sigma_right = (\z s -> binding (z_A_s z s))
                      left' = left{bindings=bindings_left'}
                      right' = right{metaChains = metaChains_right'}
                      del' = substChainVar sigma_right (MetaChain emptySubstitution  typ ee z s) del
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                  ,
--     case 2:
                  let 
                      sol' = (MapCV $ MetaChain emptySubstitution  typ ee z s :|-> sigma_right):sol
                      sigma_right = \z s -> (binding (z_A_vletX' z)){metaChains = [ee' s]}
                      (name_ctxtname:name_x':name_ee':fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      x' =  MetaVarLift emptySubstitution  ("X"++name_x')
                      ee' = \s -> case typ of 
                              AChain -> MetaChain emptySubstitution  AChain ("EE" ++ name_ee') x' s 
                              VarChain -> MetaChain emptySubstitution  VarChain ("VV" ++ name_ee') x' s 
                      z_A_vletX' = \z -> case typ of 
                                       AChain -> (z := MetaCtxtSubst emptySubstitution  ACtxt ctxtname (VarL x'))
                                       VarChain -> (z:= (VarL x'))
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (b1 :=?=: z_A_vletX' z):(bindingEq gam)
                                }
                      left' = left{bindings=bindings_left'}
                      right' = right{metaChains = (ee' s):metaChains_right'}
                      del' = substChainVar sigma_right (MetaChain emptySubstitution  typ ee z s) del 
                      
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                  ,
--     case 3:
                  let 
                      sol' = (MapCV $ MetaChain emptySubstitution  typ ee z s :|-> sigma_right):sol
                      sigma_right = \z s -> (binding x1_A_vletX2){metaChains = [ee1 z,ee2 s]}
                      (name_ctxtname:name_x1:name_x2:name_ee1:name_ee2:fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      x1 =  MetaVarLift emptySubstitution ("X"++name_x1)
                      x2 =  MetaVarLift emptySubstitution ("X"++name_x2)
                      ee1 = \z -> 
                             case typ of
                               AChain -> MetaChain emptySubstitution  AChain ("EE" ++ name_ee1) z (VarL x1)
                               VarChain -> MetaChain emptySubstitution  VarChain ("VV" ++ name_ee1) z (VarL x1)
                      ee2 = \s ->
                             case typ of 
                               AChain -> MetaChain emptySubstitution  AChain ("EE" ++ name_ee2) x2 s
                               VarChain -> MetaChain emptySubstitution  VarChain ("VV" ++ name_ee2) x2 s
                      x1_A_vletX2 = case typ of 
                                       AChain -> (x1 := MetaCtxtSubst emptySubstitution  ACtxt ctxtname (VarL x2))
                                       VarChain -> (x1 := VarL x2) 
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (b1 :=?=: x1_A_vletX2):(bindingEq gam)
                                }
                      left' = left{bindings=bindings_left'}
                      right' = right{metaChains = [ee1 z,ee2 s]++metaChains_right'}
                      del' = substChainVar sigma_right (MetaChain emptySubstitution typ ee z s) del
                      
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                 ,
--     case 4:
                  let 
                      sol' = (MapCV $ MetaChain emptySubstitution typ ee z s :|-> sigma_right):sol
                      sigma_right= \z s -> (binding (x1_A_s s)){metaChains = [ee' z]}
                      (name_ctxtname:name_x1:name_ee':fresh') = fresh problem
                      ctxtname = "A" ++ name_ctxtname
                      x1 =  MetaVarLift emptySubstitution ("X"++name_x1)
                      ee' = \z -> 
                             case typ of 
                              AChain -> MetaChain emptySubstitution AChain ("EE" ++ name_ee') z (VarL x1)
                              VarChain -> MetaChain emptySubstitution VarChain ("VV" ++ name_ee') z (VarL x1)
                      x1_A_s = \s -> 
                                case typ of 
                                 AChain -> (x1 := MetaCtxtSubst emptySubstitution ACtxt ctxtname s)
                                 VarChain -> (x1 := s)
                      gam' = gam{environmentEq =(left' :=?=: right'):rest
                                ,bindingEq = (b1 :=?=: (x1_A_s s)):(bindingEq gam)
                                }
                      left' = left{bindings=bindings_left'}
                      right' = right{metaChains = (ee' z):metaChains_right'}
                      del' = substChainVar sigma_right (MetaChain emptySubstitution typ ee z s) del
                  in problem{fresh = fresh',gamma = gam',solution=sol',delta=del'}
                 ]
                 

                | ((MetaChain sub typ ee z s),metaChains_right') <- splits (metaChains right)
              ])
           -- for all b1 in the left

          
-- RULE: (Sol,Gamma cup {E1;...;En = EE1[x1,s1]; ...; EEk[xk;sk]; E1';...;Em'},Delta)
--   -->  (Sol circ sigma,Gamma, (Delta1,Delta2 cup Delta2',Delta3, Delta4 cup j=1,...,n{E1,j; ...; En,j = EEj[xj,sj]} )
--  where sigma = i=1 to n {Ei |-> Ei,1 ; ... ; Ei,k+m} cup j=1...m{Ej' |-> E1,k+j;...;En k+j}
--    and Delta2' is a minimal subset of {Eij} s.t.
--    * Ei in Delta2 => some Ei,h in Delta2', 
--    * Ej' in Delta2 implies some Eh,k+j in Delta2',and
--    * some Eh,j in Delta2' for all 1 <= j <= k
-- (envCh)

 | ((left :=?=: right,rest):_)
     <- [splittings | splittings@(left :=?=: right,rest) <- splits (environmentEq gam)
         , not (hasBinding left)
         , not (hasBinding right)          
         , not (hasMetaChain left)
         , (length (metaEnvVars left) > 1)
         , (hasMetaChain right || ((length (metaEnvVars right)) > 1))
         ] 
  =
   let frsh1       = (fresh problem)
       bigEsLeft   = metaEnvVars left
       bigEsLeftNumbered = zip [1..] bigEsLeft
       bigEsRight  = metaEnvVars right      
       chainsRight = metaChains right
       k           = length chainsRight
       n           = length bigEsLeft
       m           = length bigEsRight
       newEs       = let names = map (\x -> Env emptySubstitution $ "E" ++ x) (take (n*(k+m)) frsh1)
                         indices = [(i,j) | i <- [1..n], j <- [1..k+m]]
                     in zip indices names
       frsh2 =  drop (n*(k+m)) frsh1                     
       newEsGroupedInRows    = map (map snd) $ groupBy ((==) `on` (fst . fst)) newEs -- vect of rows
       newEsGroupedInColumns = map (map snd) $ groupBy ((==) `on` (snd . fst)) $ (sortBy (\((a1,b1),c1) ((a2,b2),c2) -> compare (b1,a1) (b2,a2)) newEs) -- vect of columns
      {-  These are the new Es as a matrix:
      
       E_1,1 ... E_1,k+m
       ...
       E_i,1 ... E_i,k+m
       ...
       E_n,1 ....E_n,k+m
      -}
       
       -- select one of each column of the first k colums
       delta2InitsForChVars :: [[EnvVar]]
       delta2InitsForChVars =
         let cols = take k newEsGroupedInColumns
             go [] = [[]]
             go (col:cols) =  [sel:rest | (sel,_) <- splits col, rest <- (go cols)]
         in go cols
       -- the selection for each E' in Delta2: due to minimality it has to respect the choices before
       ePrimesWithNewEij :: [(EnvVar,[EnvVar])]
       ePrimesWithNewEij = zip bigEsRight (drop k newEsGroupedInColumns)  -- (E_i' |- E...E)

       delta2ForBigEiRights :: [EnvVar] -> [(EnvVar,[EnvVar])] -> [ [EnvVar] ]
       delta2ForBigEiRights someDeltaInit [] = [someDeltaInit]
       delta2ForBigEiRights someDeltaInit ((bigEiRight,newEijs):xxs)
         | bigEiRight `elem` (delta2 del) =
            case [e | e <- someDeltaInit, e `elem` newEijs]
              of [] -> concat [delta2ForBigEiRights (sel:someDeltaInit) xxs | (sel,_) <- splits newEijs]
                 (_:_) -> delta2ForBigEiRights someDeltaInit xxs
         | otherwise = delta2ForBigEiRights someDeltaInit xxs
       delta2Step2ForERights :: [[EnvVar]]         
       delta2Step2ForERights = concat [delta2ForBigEiRights delta2Init ePrimesWithNewEij | delta2Init <- delta2InitsForChVars]

       eWithNewEij :: [(EnvVar,[EnvVar])]
       eWithNewEij = zip bigEsLeft newEsGroupedInRows
       delta2ForBigEiLefts :: [EnvVar] -> [(EnvVar,[EnvVar])] -> [[EnvVar]]
       delta2ForBigEiLefts someDeltaInit [] = [someDeltaInit]
       delta2ForBigEiLefts someDeltaInit ((bigEi,newEijs):xxs)
         | bigEi `elem` (delta2 del) =
            case [e | e <- someDeltaInit, e `elem` newEijs]
              of [] -> concat [delta2ForBigEiLefts (sel:someDeltaInit) xxs | (sel,_) <- splits newEijs]
                 (_:_) -> delta2ForBigEiLefts someDeltaInit xxs
         | otherwise = delta2ForBigEiLefts someDeltaInit xxs
         
       delta2Step3ForELefts = concat [delta2ForBigEiLefts delta2Init eWithNewEij | delta2Init <- delta2Step2ForERights]
       sigma1 = zipWith (\ei es -> MapEnv (ei:|-> empty{metaEnvVars = es})) bigEsLeft newEsGroupedInRows
       bigEsInDelta2WithRow = [(i,ei,es) | (i,ei,es) <- (zipWith (\(i,ei) es -> (i,ei,es)) bigEsLeftNumbered newEsGroupedInRows), ei `elem` (delta2 del)]
       sigma2 = zipWith (\ei es -> MapEnv (ei :|-> empty{metaEnvVars = es})) bigEsRight (drop k newEsGroupedInColumns)
       newEquations = zipWith (\ch es -> empty{metaEnvVars = es} :=?=: empty{metaChains=[ch]}) chainsRight (take k newEsGroupedInColumns)
       sigma = sigma1 ++ sigma2
       sol' = sigma ++ sol
       gam'         = applySolution sigma gam{environmentEq=rest}
       res=
         [problem{ solution=sol'
                   , gamma=gam'
                   , fresh=frsh2
                    ,delta= applySolution sigma (del{ delta2 = (delta2 del) ++ delta2Prime
                                                   , delta4 =(delta4 del){environmentEq = environmentEq (delta4 del) ++ newEquations}})}
           | delta2Prime <-  minimizeSets delta2Step3ForELefts
           ]
    in  (Some res) 
--    
-- no environment rule applicable:
--
 | otherwise = Empty

-- ********************************************************************************************************************
-- expandDelta4Constraints expands constraints E1;E2 =?= EE[z,s] 
-- if there are non-capture constraints (E1,E2) or (E2,E1)
expandDelta4ConstraintsIfPossible :: Problem -> UnifResult
expandDelta4ConstraintsIfPossible problem@(Problem{solution=sol,gamma=gam,delta=del})
 | Just (eq@(left :=?=: right),rest)
     <- splitBy (\eq -> fst (del4constrSplitable (splitCapCon (delta3 del)) eq)) (environmentEq (delta4 del))
 ,[envVar1,envVar2] <- metaEnvVars left
 ,[ch@(MetaChain sub typ ee z s) ] <- metaChains right
 ,[(Env sub1 e1),(Env sub2 e2)] <- if snd  (del4constrSplitable (splitCapCon (delta3 del)) eq)
                                     then [envVar2,envVar1] -- commute
                                      else [envVar1,envVar2]
 =  Some [
      let (_name_ee1:_name_ee2:_name_ctxt:_name_zi:fresh') =  (fresh problem)
          name_ee1 = case typ of 
                      AChain -> "EE" ++ _name_ee1
                      VarChain -> "VV" ++ _name_ee1
          name_ee2 = case typ of 
                      AChain -> "EE" ++ _name_ee2
                      VarChain -> "VV" ++ _name_ee2
          name_ctxt =  "A" ++ _name_ctxt
          name_zi   =  "X" ++ _name_zi
          zi        = ((MetaVarLift emptySubstitution name_zi))
          actxt  = case typ of 
                     AChain ->    \h  -> MetaCtxtSubst emptySubstitution ACtxt name_ctxt h
                     VarChain ->   id
          ee1 = \z -> MetaChain sub typ name_ee1 z (actxt (VarL zi))
          ee2 = \s -> MetaChain sub typ name_ee2 zi s
          solNew = MapCV (MetaChain sub typ ee z s :|-> \z s -> empty{metaChains = [(ee1 z), (ee2 s)]})
          eqsNew = [empty{metaEnvVars=[Env sub1 e1]} :=?=: empty{metaChains=[ee2 s]}
                   ,empty{metaEnvVars=[Env sub2 e2]} :=?=: empty{metaChains=[ee1 z]}]
          del' = applySolution [solNew] (del{delta4=(delta4 del){environmentEq = rest}})
          gam' = gam{environmentEq = (environmentEq gam) ++ eqsNew}
          sol' = solNew:sol 
      in 
           
           problem{gamma=gam',solution=sol',delta=del',fresh=fresh'}
      ,
      let 
          eqsNew = [empty{metaEnvVars=[Env sub1 e1]} :=?=: empty
                   ,empty{metaEnvVars=[Env sub2 e2]} :=?=: empty{metaChains=[ ch ]}]
          gam' = gam{environmentEq = (environmentEq gam) ++ eqsNew}
          del' = (del{delta4=(delta4 del){environmentEq = rest}})
       in problem{gamma=gam',delta=del'}
      ,
      let 
          eqsNew = [empty{metaEnvVars=[Env sub1 e1]} :=?=: empty{metaChains=[ch]}
                   ,empty{metaEnvVars=[Env sub2 e2]} :=?=: empty]
          gam' = gam{environmentEq = (environmentEq gam) ++ eqsNew}
          del' = (del{delta4=(delta4 del){environmentEq = rest}})
       in problem{gamma=gam',delta=del'}
       ]
       
expandDelta4ConstraintsIfPossible problem@(Problem{solution=sol,gamma=gam,delta=del}) = Empty

del4constrSplitable del3splitted (left :=?=: right)
 | null (metaChains left)
 , null (bindings left)
 , null (metaEnvVars right)
 , null (bindings right)
 , [(Env sub1 e1),(Env sub2 e2)] <- metaEnvVars left
 , [MetaChain sub typ ee z s ] <- metaChains right
 , Just set <- Map.lookup (SCMetaEnv (Env sub1 e1)) del3splitted
 , SCMetaEnv (Env sub2 e2) `Set.member` set   
 = (True,False)  -- True and the envVars are in the right order
del4constrSplitable del3splitted (left :=?=: right)
 | null (metaChains left)
 , null (bindings left)
 , null (metaEnvVars right)
 , null (bindings right)
 , [(Env sub2 e2),(Env sub1 e1)] <- metaEnvVars left
 , [MetaChain sub typ ee z s ] <- metaChains right
 , Just set <- Map.lookup (SCMetaEnv (Env sub1 e1)) del3splitted
 , SCMetaEnv (Env sub2 e2) `Set.member` set   
 = (True,True)  -- True but the bindings have to be commuted
 | otherwise 
 = (False,False)
-- ********************************************************************************************************************
  
 
-- ********************************************************************************************************************
-- Check whether Delta4 chains are solvable
-- solvableChains checks whether the environment constraints in Delta4 are solvable
-- It returns Nothing, if the constraints are solvable, and Just text, otherwise, where text is some error message 
-- ********************************************************************************************************************
solvableChains :: Problem -> Maybe String
solvableChains  problem@(Problem{solution=sol,gamma=gam,delta=del}) =  run 
  where   
   run = 
       let distribs fn = goAll fn eqs freshInp
           res fn = let allChecks =
                            [(\x -> case checkNCCs (delta2 $ del) x of 
                                    Nothing -> Nothing -- 16i05 checkWeakDVCDelta3 x
                                    Just err -> Just (show err))  (applySolution sol del3)
                            | sol <- map composeAll $ combine (distribs fn)
                            ]
                    in if any isNothing allChecks 
                                        then Nothing -- found solution
                                        else Just ("Solvability Check failed: Found no solution:\n Delta3:" 
                                                          ++ show del3 ++ "\n" ++ "(Solutions,applied to del3):" 
                                                          ++ show [ (sol,applySolution sol del3) | sol <- map composeAll $ combine (distribs fn)])
                    
       in     res (\n1 n2 -> (n1^2*(n2+1)+n2))
   eqs = (environmentEq $ delta4 del)                               -- constraints in Delta 4   Ei1,...,Ein =?= EEi[z,s]
   leftEI = [v | (left :=?=: _) <- eqs, v <- metaEnvVars left]      -- E11,...,E1n,...,Em1,...Emn = all left hand sides of constraints
   del2EI = [ei | ei <- leftEI, ei `elem` (delta2 del)]             -- {E11,...,E1n,...,Em1,...Emn} cup Delta2, i.e. such Eij that are non-empty
   del3 = (delta3 del)                                              -- Delta 3
   freshInp = fresh problem 
   combine [] = [[]]
   combine (d:ds) = [sol++sols | sol <- d, sols <- (combine ds)]   
   
   goAll fn  [] _ = []
   goAll fn (eq:eqs) f = 
    let (dist,fresh1) = goB fn eq f
    in   (dist:(goAll fn eqs fresh1))
       
   goB f equation@(left :=?=: right) fresh = 
    let
        leftEI = metaEnvVars left
        [MetaChain sub typ ee z s] = metaChains right
        n = length leftEI
        delE = [v | v <- leftEI, v `elem` (del2EI)]   
        n1 = length delE
        n2 = n - n1
        maxbound =  f n1 n2 
        dfs i b fv 
            | i > b = (fv,[])
        dfs i b fv = let (d,fv') = go equation i fv
                         (fv'',ds) = dfs (i+1) b fv'
                     in (fv'',d:ds)
        (freshOut,d) = dfs 1 maxbound fresh           
    in  (concat d,freshOut)
   go equation@(left :=?=: right) bound' fresh0 = 
    let 
        leftEI = metaEnvVars left
        [MetaChain sub typ ee z s] = metaChains right
        (names_a,fresh1) = splitAt (bound'-1) fresh0
        (names_x,fresh2) = splitAt (bound'-1) fresh1
        (x:a:freshrest) = fresh2
        (varsA) = 
                  [case typ of
                     AChain ->  MetaCtxtSubst emptySubstitution  ACtxt ("A" ++ name_a)
                     VarChain -> id
                   | name_a <- names_a]
        (varsX) = [MetaVarLift emptySubstitution  ("X" ++ name_x) | name_x <- names_x]
        newBindingsGEQ3 = zipWith 
                           (\y r -> y:=r) 
                           (z:varsX) 
                           ((zipWith 
                                (\a x -> (a (VarL x))) 
                                (varsA) 
                                varsX)++[s])
        newBindings1 = [z:=s]
        newBindings2 = [z:=case typ of 
                             AChain -> MetaCtxtSubst emptySubstitution ACtxt ("A" ++ a) (VarL (MetaVarLift emptySubstitution ("X" ++ x)))
                             VarChain ->  (VarL (MetaVarLift emptySubstitution  ("X" ++ x)))
                        ,(MetaVarLift emptySubstitution  ("X" ++ x)) := s]
        toBeDistributedBindings = if bound' == 1 then newBindings1 
                                    else if bound' == 2 then newBindings2 
                                          else newBindingsGEQ3
        distribute :: [EnvVar] -> [Binding] -> [[SolutionEntry]]
        distribute [e] bs = [[(MapEnv (e :|-> empty{bindings=bs}))]]
        distribute (e:es) bs = [(MapEnv (e :|-> empty{bindings=here})):rest |  (here,next) <- (divide bs)
                               , rest <- distribute es next 
                               ]
        distributions = let res = distribute leftEI  toBeDistributedBindings in res -- trace ("ALL dist" ++ (unlines $ map show res)) res
        distributions' = let res = [dist | dist <- distributions, isValid dist] in res --trace ("VALID dist" ++ (unlines $ map show res)) res
        isValid dist = null [e | (MapEnv (e :|-> env)) <- dist, isEmpty env && e `elem` del2EI]
    in  (distributions',freshrest)

  
-- ********************************************************************************************************************
-- Commute all equations
-- ********************************************************************************************************************
commuteProblem prob = prob {gamma = gam'}
  where
   gam = gamma prob
   gam' = gam { expressionEq = map commute $ expressionEq gam 
              , bindingEq = map commute $ bindingEq gam
              , varLiftEq = map commute $ varLiftEq gam              
  
              , environmentEq = map commute $ environmentEq gam              
              }
              
commuteDelta4 prob = prob {delta = del'}
  where
   del = delta prob
   del4 = delta4 del
   del4' = del4 { expressionEq = map commute $ expressionEq del4 
                , bindingEq = map commute $ bindingEq del4
                , varLiftEq = map commute $ varLiftEq del4             
                , environmentEq = map commute $ environmentEq del4              
              }
   del' = del{delta4=del4'}
              
-- ********************************************************************************************************************

applyFullGuessRulesEnv problem@(Problem{solution=sol,gamma=gam,delta=del})
 | (q@(Env sub name):r) <- [x | x@(Env sub name) <- (envVarsInSol $ composeAll sol)
                                   , not (x `elem` (delta2 del) || (Env emptySubstitution name) `elem` (delta2 del)) ]
                                   
    = let problem1 = 
                let oldDel_2 = delta2 del
                    newDel_2 = q:oldDel_2
                in  problem{delta = del{delta2 = newDel_2}}
          problem2 = 
                let sol' = (MapEnv $ (Env sub name) :|-> empty):sol
                    gam' = substEnv empty name gam
                    del' = substEnv empty name del
               in problem{solution=sol',gamma=gam', delta=del'}
      in Some [problem1,problem2]
-- no guess rule applicable
 | otherwise = Empty
