-- DONE SYNTAX CHANGE
-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Unification.UnificationProblem
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- The data type for unification problems which is used for the unification algorithm
--
-----------------------------------------------------------------------------
-- ====================================================
module LRSX.Unification.UnificationProblem where
import Data.List
import Data.Maybe
import LRSX.Language.Syntax


-- | the function used for checking the NCCs, 
--   abstracted here to change it into other functions
--   at the moment it is synonymous to 'capturedInHole'
checkBVorNCC :: Context -> [VarLift]
checkBVorNCC e = capturedInHole e

 
-- ============================================================================
-- Data-type for a unification problem
-- ============================================================================
 
data Problem = Problem {
                       solution      :: Solution,               -- ^ the solution, generated by unification
                       gamma         :: Gamma,                  -- ^ the unification equations
                       delta         :: Delta,                  -- ^ the constraints
                       constantDelta :: Delta,                  -- ^ fixed (given) constraints, only used for matching, not for unification
                       fresh         :: [String],               -- ^ a list of fresh names
                       functions     :: FunInfo,                -- ^ the function symbols together with their arity and a list of strict positions
                       inputGamma    :: Gamma,                  -- ^ the input equations
                       inputDelta    :: Delta,                  -- ^ the input constraints in delta
                       alphaBase     :: [(SyntaxComponent,SubstItem)],   -- ^ freshness constraints on an alpha renaming, not used by the unification, but for matching
                       mpLogging     :: String,                          -- ^ a String to store logging information 
                       expandDelta4Constraints :: Bool,                  -- ^ True, if E1;E2 = EE[z,s] should be expanded, if there is an NCC (E1,E2) or (E2,E1)
                       expandAllContexts :: Bool,                        -- ^ True, if all (and not only needed) context variables are guessed empty/non-empty
                                                                         --   set to True, this may help when joining overlaps with matching
                       expandAllEnvs :: Bool} 
 deriving(Show)                    

-- compute the strict positions of a function symbol
getStrictPos :: Problem -> Name -> Maybe [Int]
getStrictPos pb f = 
 case [pos | (n,_,pos,_) <- (functions pb), n == f] of
  []      -> Nothing
  ([pos]) -> Just pos
  (_:_)   -> error "in getStrictPos: Positions are not unique?"

-- ============================================================================
-- Gamma: multisets of different equations:
-- ============================================================================
data Gamma = Gamma {varLiftEq     :: [Equation VarLift]     -- ^ variable equations
                   ,expressionEq  :: [Equation Expression]  -- ^ expression equations, note that it is forbidden to equate contexts with expressions, 
                                                            --   which is not ensured by the modelling                      
                   ,environmentEq :: [Equation Environment] -- ^ enviroment equations
                   ,bindingEq     :: [Equation Binding]     -- ^ binding equations
                   }
                   
-- | check whether an environment is empty                   
nullGamma :: Gamma -> Bool
nullGamma (Gamma{varLiftEq = d1, expressionEq =d2, environmentEq=d3, bindingEq = d4}) = 
   and [null d1,null d2,null d3,null d4]                   
   
-- | construct an empty enviroment   
emptyGamma :: Gamma   
emptyGamma = (Gamma{varLiftEq = [], expressionEq =[], environmentEq=[], bindingEq = []})

-- Show-instance                   
instance Show Gamma where
 show sol   = "{" ++ (concat $ intersperse "," (filter (not.null) $  [ showMap (varLiftEq sol), showMap (expressionEq sol), showMap (environmentEq sol), showMap (bindingEq sol)])) ++ "}"
  where
   showMap [] = ""
   showMap mp =   concat $ intersperse "," $   [show eq | eq <-   mp] 

-- | compute all variables occuring in gamma (meta and concrete variables)
allVariablesInGamma :: Gamma -> [VarLift]
allVariablesInGamma gam = concat $
   [ [l,r] | (l :=?=: r) <- varLiftEq gam]
   ++
   [ (allVariables l) ++ (allVariables r) | (l :=?=: r) <- expressionEq gam]
   ++
   [ (xl):(xr):((allVariables l) ++ (allVariables r))  | ((xl:=l) :=?=: (xr:=r)) <- bindingEq gam]
   ++
   [ allVariablesInEnv env1 ++ allVariablesInEnv env2 | (env1 :=?=: env2) <- environmentEq gam]

instance Constantify a => Constantify (Equation a) where
  constantify (a :=?=: b) = (constantify a) :=?=: (constantify b)
  deconstantify (a :=?=: b) = (deconstantify a) :=?=: (deconstantify b)
  
instance Constantify Gamma where
  constantify gamma = gamma{varLiftEq = map constantify (varLiftEq gamma)
                          
                           ,expressionEq = map constantify (expressionEq gamma)
                           ,bindingEq  = map constantify (bindingEq gamma)
                           ,environmentEq = map constantify (environmentEq gamma)
  } 
  deconstantify gamma = gamma{varLiftEq = map deconstantify (varLiftEq gamma)
                          
                           ,expressionEq = map deconstantify (expressionEq gamma)
                           ,bindingEq  = map deconstantify (bindingEq gamma)
                           ,environmentEq = map deconstantify (environmentEq gamma)
  } 

showNonCapConstraint (Left e,e') = "(" ++ show e ++ "," ++ show e' ++ ")"
showNonCapConstraint (Right e,e') = "[" ++ show e ++ "," ++ show e' ++ "]"
instance Constantify Delta where
  constantify delta = delta{delta1 = constantify (delta1 delta)
                           ,delta2 = delta2 delta
                           ,delta3 = [(constantify l,constantify r) | (l,r) <-delta3 delta]
                           ,delta4 = constantify (delta4 delta)
  }
  deconstantify delta = delta{delta1 = deconstantify (delta1 delta)
                           ,delta2 = delta2 delta
                           ,delta3 = [(deconstantify l,deconstantify r) | (l,r) <-delta3 delta]
                           ,delta4 = deconstantify (delta4 delta)
  }
-- Substitute-instance
instance Substitute Gamma where
 subst e1 e2  gamma = substG e1 e2 gamma 
  where 
    substG e1 e2 gamma =
      gamma {varLiftEq = subst e1 e2 (varLiftEq gamma)
        
            ,expressionEq = subst e1 e2 (expressionEq gamma)
            ,environmentEq = subst e1 e2 (environmentEq gamma)
            ,bindingEq = subst e1 e2 (bindingEq gamma)
            }
 substEmpty n gamma = substG n gamma 
  where 
    substG n gamma =
      gamma {varLiftEq = substEmpty n (varLiftEq gamma)
      
            ,expressionEq = substEmpty n (expressionEq gamma)
            ,environmentEq = substEmpty n (environmentEq gamma)
            ,bindingEq     = substEmpty n (bindingEq gamma)
            }
 substEnv env n gamma = 
      gamma {varLiftEq = substEnv env n (varLiftEq gamma)
            
            ,expressionEq = substEnv env n (expressionEq gamma)
            ,environmentEq = substEnv env n (environmentEq gamma)
            ,bindingEq = substEnv env n (bindingEq gamma)
            }
 substChainVar env n gamma = 
      gamma {varLiftEq = substChainVar env n (varLiftEq gamma)
            
            ,expressionEq = substChainVar env n (expressionEq gamma)
            ,environmentEq = substChainVar env n (environmentEq gamma)
            ,bindingEq = substChainVar env n (bindingEq gamma)
            }
 substEnvConstraint env n gamma = 
      gamma {varLiftEq = substEnvConstraint env n (varLiftEq gamma)
          
            ,expressionEq = substEnvConstraint env n (expressionEq gamma)
            ,environmentEq = substEnvConstraint env n (environmentEq gamma)
            ,bindingEq = substEnvConstraint env n (bindingEq gamma)
            }
   
-- solved checks whether all equations are solved, i.e. all sets are empty
solved :: Gamma -> Bool
solved gamma = 
 and [null $ varLiftEq gamma,  null $ expressionEq gamma, null $ environmentEq gamma, null $ bindingEq gamma] 

 
 
-- Equations:
data Equation a = a :=?=: a
 deriving(Eq)
-- Show-instance
instance Show a => Show (Equation a) where
 show (a :=?=: b) = show a ++ " =?= " ++ show b
instance HasMetaVars a => HasMetaVars (Equation a) where
 computeMetaVars (a :=?=: b) = combineMetaVarsList [computeMetaVars a, computeMetaVars b]
-- Substitute-instance
instance Substitute a => Substitute (Equation a) where
  subst e1 e2 (a :=?=: b) = (subst e1 e2 a) :=?=: (subst e1 e2 b)
  substEmpty n (a :=?=: b) = (substEmpty n  a) :=?=: (substEmpty n b)
  substEnv env n ( a :=?=: b) = (substEnv env n a) :=?=: (substEnv env n b)
  substChainVar env v ( a :=?=: b) = (substChainVar env v a) :=?=: (substChainVar env v b)
  substEnvConstraint env v (a :=?=: b)  = (substEnvConstraint env v a) :=?=: (substEnvConstraint env v b)
instance HasFnSymbol a => HasFnSymbol (Equation a) where
 getFnSymbols (a :=?=: b) = nubFnSyms $ getFnSymbols a ++ getFnSymbols b
 
commute :: Equation a -> Equation a 
commute (left :=?=: right) = right :=?=: left  
 
  
type EquationsWithConstraints = ([Equation Expression],[Expression],[EnvVar],NonCaptureConstraints)
type EquationsWithTwoConstraints = ([Equation Expression],[Expression],[EnvVar],NonCaptureConstraints,[Expression],[EnvVar],NonCaptureConstraints)
-- List of equations, list of nonempty contexts, list of nonempty environments
  
-- check constrained equations
-- check :: EquationsWithConstraints -> Maybe String -- Nothing = ok, Just err = Fail
check pb = 
 let eqs = expressionEq (gamma pb)
     ctxtcon = delta1 (delta pb)
     nonemptyenvvars = delta2 (delta pb)
     capcon = delta3 (delta pb) 
 in checkGamma (Gamma{expressionEq=eqs,bindingEq=[],environmentEq=[],varLiftEq=[]},ctxtcon,nonemptyenvvars,capcon)

  
checkGammaUnifInp (gamma,ctxtcon,nonemptyenvvars)
     | ((v,a,b):_) <- checkDoubleVars = Just $ a ++ " " ++ v ++ " occurs also as " ++ b 
     | ((u,v,eq):_) <- checkChains = Just $ "chain-variables on both sides of the equation " ++ eq ++ " (" ++ u ++ " on the left hand side and " ++ v ++ " on the right hand side)"
     | ((v:_,dec):_) <- checkOtherVars = Just $ dec ++ " " ++ v ++ " occurs more than once in the unification problem"
     | Just err <- checkWeakDVCGamma gamma = Just $ "weak DVC not valid:" ++ err
     | otherwise = Nothing
   where  
    nonemptycxtvars = [v | MetaCtxt s srt v <- ctxtcon]
    eq1 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- varLiftEq gamma]
    eq3 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- expressionEq gamma]
    eq4 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- bindingEq gamma]
    eq5 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- environmentEq gamma]
    allVars = combineMetaVarsList $
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq1]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq3]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq4]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq5]
    checkNonEmptyCtxtVars = [v | v <- nonemptycxtvars, v `notElem` (ctxtvars allVars)]
    checkNonEmptyEnvVars = [v | Env s v <- nonemptyenvvars, v `notElem` (envvars allVars)]
    checkDoubleVars =    [(v,"context-variable","variable")      | v <- (ctxtvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"context-variable","expression variable")        | v <- (ctxtvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"context-variable","environment-variable")       | v <- (ctxtvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"context-variable","chain-variable")             | v <- (ctxtvars allVars), v `elem` (chainvars allVars)]
                      ++ [(v,"chain-variable","bound variable")        | v <- (chainvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"chain-variable","expression variable")          | v <- (chainvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"chain-variable","environment-variable")         | v <- (chainvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"environment-variable","variable")  | v <- (envvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"environment-variable","expression variable")    | v <- (envvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"expression-variable","bound variable")   | v <- (exprvars allVars), v `elem` (liftvars allVars)]
    checkEquations = concat $
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"let-bound variable",eq)    | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"let-bound variable",eq)    | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r)]                   
                         
    checkChains =     [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                                      
    checkExprVars = filter (\x -> length x > 2) $ group $ sort $ exprvars allVars
    checkOtherVars = concat [zip (filter (\x -> length x > 1) $ group $ sort $ vars) (repeat dec)
                         | (vars,dec) <- [(envvars allVars,"environment-variable"),(ctxtvars allVars,"context-variable"),(chainvars allVars,"chain-variable")]]

checkGamma (gamma,ctxtcon,nonemptyenvvars,capcon)
     | ((v,a,b):_) <- checkDoubleVars = Just $ a ++ " " ++ v ++ " occurs also as " ++ b 
     | ((v,var,eq):_) <- checkEquations = Just $ var ++ " " ++ v ++ " occurs on both sides of the equation " ++ eq
     | ((u,v,eq):_) <- checkChains = Just $ "chain-variables on both sides of the equation " ++ eq ++ " (" ++ u ++ " on the left hand side and " ++ v ++ " on the right hand side)"
     | ((v:_):_) <- checkExprVars = Just $ "expression variable " ++ v ++ " occurs more than twice in the unification problem"
     | ((v:_,dec):_) <- checkOtherVars = Just $ dec ++ " " ++ v ++ " occurs more than once in the unification problem"
     | Just err <- checkWeakDVCGamma gamma = Just $ "weak DVC not valid:" ++ err
     | otherwise = Nothing
   where  
    nonemptycxtvars = [v | MetaCtxt s srt v <- ctxtcon]
    eq1 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- varLiftEq gamma]
    eq3 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- expressionEq gamma]
    eq4 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- bindingEq gamma]
    eq5 = [(l,computeMetaVars l) :=?=: (r,computeMetaVars r)| (l :=?=: r) <- environmentEq gamma]
    allVars = combineMetaVarsList $
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq1]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq3]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq4]
             ++
             [l `combineMetaVars` r | (_,l) :=?=: (_,r) <- eq5]
    checkNonEmptyCtxtVars = [v | v <- nonemptycxtvars, v `notElem` (ctxtvars allVars)]
    checkNonEmptyEnvVars = [v | Env s v <- nonemptyenvvars, v `notElem` (envvars allVars)]
    checkDoubleVars =    [(v,"context-variable","bound variable")      | v <- (ctxtvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"context-variable","expression variable")        | v <- (ctxtvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"context-variable","environment-variable")       | v <- (ctxtvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"context-variable","chain-variable")             | v <- (ctxtvars allVars), v `elem` (chainvars allVars)]
                      ++ [(v,"chain-variable","bound variable")        | v <- (chainvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"chain-variable","expression variable")          | v <- (chainvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"chain-variable","environment-variable")         | v <- (chainvars allVars), v `elem` (envvars allVars)]
                      ++ [(v,"environment-variable","bound variable")  | v <- (envvars allVars), v `elem` (liftvars allVars)]
                      ++ [(v,"environment-variable","expression variable")    | v <- (envvars allVars), v `elem` (exprvars allVars)]
                      ++ [(v,"expression-variable","bound variable")   | v <- (exprvars allVars), v `elem` (liftvars allVars)]

    checkEquations = concat $
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                       ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r)]                   
                       ++
                      [    [(v,"context-variable",eq)      | v <- (ctxtvars varsl), v `elem` (ctxtvars varsr)] 
                       ++  [(v,"expression-variable",eq)   | v <- (exprvars varsl), v `elem` (exprvars varsr)] 
                       ++  [(v,"chain-variable",eq)        | v <- (chainvars varsl), v `elem` (chainvars varsr)] 
                                              ++  [(v,"bound variable",eq) | v <- (liftvars varsl), v `elem` (liftvars varsr)] 
                       ++  [(v,"environment-variable",eq)  | v <- (envvars varsl), v `elem` (envvars varsr)] 
                         | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r)]                   
                         
    checkChains =     [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq1, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq3, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq4, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                   
                   ++ [ (u,v,eq) | ((l,varsl) :=?=: (r,varsr)) <- eq5, let eq = show (l :=?=: r), u <- (chainvars varsl), v <- (chainvars varsr)]                                      
    checkExprVars = filter (\x -> length x > 2) $ group $ sort $ exprvars allVars
    checkOtherVars = concat [zip (filter (\x -> length x > 1) $ group $ sort $ vars) (repeat dec)
                         | (vars,dec) <- [(envvars allVars,"environment-variable"),(ctxtvars allVars,"context-variable"),(chainvars allVars,"chain-variable")]]
-- ----------------------------------
-- several helper functions on Gamma
-- ----------------------------------
environmentVarsInGamma :: Gamma -> [EnvVar]
environmentVarsInGamma gam = 
  concat (
      [ search e1 ++ search e2 | e1 :=?=: e2 <- expressionEq gam]
   ++ [ searchEnv env1 ++ searchEnv env2 | env1 :=?=: env2 <- environmentEq gam]
   ++ [ searchBind b1 ++ searchBind b2 | (b1 :=?=: b2) <- bindingEq gam]
  )
 where
    searchArgPos (Binder v e) = search e
    searchArgPos (VarPos _) = []
    search (Fn _ args) = concatMap searchArgPos args
    search (Letrec env e) = searchEnv env ++ search e
    search (MetaCtxtSubst _ _ _ e) = search e
    search (ConstantCtxtSubst _ _ _ e) = search e    
    search _ = []
    searchEnv env =
      let bs = concatMap searchBind $ bindings env
          ms = [Env sub v | Env sub v <- metaEnvVars env] 
          ms' = [Env sub v | Env sub v <- constantEnvVars env] 
          me = concat [search e | MetaChain _ _ _ _ e <- metaChains env]
          ce = concat [search e | ConstantChain _ _ _ _ e <- constantChains env]          
      in bs ++ ms ++ ms' ++ me ++ ce

    searchBind (_ := e) = search e

envVarsInSol  sol = concatMap getEnvvars sol
 where
     getEnvvars (MapVarLift _) = []
     getEnvvars (MapVarLet _) = []
     getEnvvars (MapExpr    (e1 :|-> e2 )) = envVarsInExpr e2
     getEnvvars  (MapEnv     (envVar :|-> env)) = envVarsInEnv env
     getEnvvars (MapCV _) = []
     getEnvvars (MapEnvSpec     (envVar :|-> env)) = envVarsInEnv env
    
envVarsInArgPos (Binder v e) = envVarsInExpr e
envVarsInArgPos (VarPos _) = []
envVarsInExpr :: Expression -> [EnvVar]
envVarsInExpr (Fn _ args) = concatMap envVarsInArgPos args
envVarsInExpr (Letrec env e) = envVarsInEnv env ++ envVarsInExpr e
envVarsInExpr (MetaCtxtSubst sub srt name expr) = (envVarsInExpr expr)
envVarsInExpr (ConstantCtxtSubst sub srt name expr) = (envVarsInExpr expr)
envVarsInExpr (MetaCtxt sub srt name) = []
envVarsInExpr (ConstantCtxt sub srt name) = []
envVarsInExpr _ = []
envVarsInEnv :: Environment -> [EnvVar]
envVarsInEnv env =
      let bs = concatMap envVarsInBind $ bindings env
          me = concat [envVarsInExpr e | MetaChain _ _ _ _ e <- metaChains env]
          ce = concat [envVarsInExpr e | ConstantChain _ _ _ _ e <- constantChains env]          
      in bs ++ me ++ ce ++ (metaEnvVars env) ++ (constantEnvVars env)
envVarsInBind (_ := e) = envVarsInExpr e    


    
contextVarsInSol  sol = concatMap getContextvars sol
 where
     getContextvars (MapVarLift _) = []
     getContextvars (MapVarLet _) = []
     getContextvars (MapExpr    (e1 :|-> e2 )) = contextVarsInExpr e2
     getContextvars (MapEnv     (envVar :|-> env)) = contextVarsInEnv env
     getContextvars  (MapCV _) = []
     getContextvars (MapEnvSpec     (envVar :|-> env)) = contextVarsInEnv env




contextVarsInGamma :: Gamma -> [Context]
contextVarsInGamma gam = 
  concat (
      [ contextVarsInExpr e1 ++ contextVarsInExpr e2 | e1 :=?=: e2 <- expressionEq gam]
   ++ [ contextVarsInEnv env1 ++ contextVarsInEnv env2 | env1 :=?=: env2 <- environmentEq gam]
   ++ [ contextVarsInBind b1 ++ contextVarsInBind b2 | (b1 :=?=: b2) <- bindingEq gam]
  )
 where
     
     
contextVarsInArgPos (Binder v e) = contextVarsInExpr e
contextVarsInArgPos (VarPos _) = []
contextVarsInExpr (Fn _ args) = concatMap contextVarsInArgPos args
contextVarsInExpr (Letrec env e) = contextVarsInEnv env ++ contextVarsInExpr e
contextVarsInExpr (MetaCtxtSubst sub srt name expr) = (MetaCtxt sub srt name):(contextVarsInExpr expr)
contextVarsInExpr (ConstantCtxtSubst sub srt name expr) = (ConstantCtxt sub srt name):(contextVarsInExpr expr)
contextVarsInExpr (MetaCtxt sub srt name) = [(MetaCtxt sub srt name)]
contextVarsInExpr (ConstantCtxt sub srt name) = [(ConstantCtxt sub srt name)]
contextVarsInExpr _ = []
contextVarsInEnv env =
      let bs = concatMap contextVarsInBind $ bindings env
          me = concat [contextVarsInExpr e | MetaChain _ _ _ _ e <- metaChains env]
          ce = concat [contextVarsInExpr e | ConstantChain _ _ _ _ e <- constantChains env]          
      in bs ++ me ++ ce
contextVarsInBind (_ := e) = contextVarsInExpr e    
    
    
    
   
   
-- check whether Gamma fulfills the weak distinct variable convention

checkWeakDVCGamma :: Gamma -> Maybe String
checkWeakDVCGamma gamma
 | any isJust l1 = case head (filter isJust l1)
                     of Just s -> Just (s ++ " (case 1)")
 | any isJust l2 = case head (filter isJust l2)
                     of Just s -> Just (s ++ " (case 2)")
 | any isJust l3 = case head (filter isJust l3)
                     of Just s -> Just (s ++ " (case 3)")
 | otherwise     = Nothing
 where
   l1 =  concatMap (\(x :=?=: y) -> [checkWeakDVC x, checkWeakDVC y]) (expressionEq gamma)
   l2 =  concatMap (\(x :=?=: y) -> [checkWeakDVCEnv x, checkWeakDVCEnv y]) (environmentEq gamma)
   l3 =  concatMap (\(x :=?=: y) -> [checkWeakDVCEnv empty{bindings=[x]}, checkWeakDVCEnv empty{bindings=[y]}]) (bindingEq gamma)
 
-- ============================================================================

-- ============================================================================
-- Delta: Several constraints on the problem 
-- ============================================================================
data Delta = Delta {delta1  :: SetOfCtxtVars
                   ,delta2  :: SetOfEnvironmentVars
                   ,delta3  :: NonCaptureConstraints
                   ,delta4  :: Gamma 
                   }
 deriving(Show)
 
emptyDelta = Delta {delta1 = [], delta2 =[], delta3=[],delta4=emptyGamma} 
type SetOfCtxtVars = [Context]


type SetOfEnvironmentVars =  [EnvVar]
type NonCaptureConstraints = [NonCaptureConstraint] 
type NonCaptureConstraint  = (Either Expression Environment,Context)

instance HasMetaVars Delta where
  computeMetaVars del = 
                      combineMetaVarsList [computeMetaVarsDelta1 (delta1 del)
                                          ,computeMetaVarsDelta2 (delta2 del)
                                          ,computeMetaVarsDelta3 (delta3 del)
                                          ,computeMetaVars (delta4 del)]
computeMetaVarsDelta3 l = combineMetaVarsList (map computeMetaVarsNCC l)
computeMetaVarsNCC (Left e,ctxt) = (computeMetaVars e) `combineMetaVars` (computeMetaVars ctxt)
computeMetaVarsNCC (Right e,ctxt) = (computeMetaVars e) `combineMetaVars` (computeMetaVars ctxt)
computeMetaVarsDelta2 del2 = emptyMVars {envvars = [e | Env sub e <- del2]}
computeMetaVarsDelta1 del1 = combineMetaVarsList (map computeMetaVars del1)


variablesInDelta3 list = concat $
 [ case l of 
        Left e -> allVariables e ++ allVariables r
        Right env -> allVariablesInEnv env ++ allVariables  r
  | (l,r) <- list] 

  
-- Substitute-instance
instance Substitute Delta where
 subst e1 e2 delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = subst e1 e2 (delta3 delta)
        , delta4 = subst e1 e1 (delta4 delta)
        }
 substEmpty n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = substEmpty n (delta3 delta)
        , delta4 = substEmpty n (delta4 delta)
        }
        
 substEnv env n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = substEnv env n (delta3 delta)
        , delta4 = substEnv env n (delta4 delta)
        }
 substChainVar env n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = substChainVar env n (delta3 delta)
        , delta4 = substChainVar env n (delta4 delta)
        }
 
 substEnvConstraint env n  delta =
  delta { delta1 = delta1 delta
        , delta2 = delta2 delta
        , delta3 = delta3 delta -- substChainVar env n (delta3 delta)
        , delta4 = delta4 delta -- substChainVar env n (delta4 delta)
        }
        
-- ============================================================================


-- ============================================================================
-- Solution: Several mappings from meta-variables to objects
-- ============================================================================
-- The solution is a list of mappings
type Solution = [SolutionEntry]

-- several types of mappings
data SolutionEntry = MapVarLift  (Mapping VarLift VarLift)
                   | MapVarLet  (Mapping VarLift VarLift)
                   | MapExpr    (Mapping Expression Expression)
                   | MapEnv     (Mapping EnvVar Environment) -- String is the name of the MetaEnv                  
                   | MapCV      (Mapping ChainVar (VarLift -> Expression -> Environment))                  
                   | MapEnvSpec (Mapping String Environment) -- for mappings to -:s and x.-
                   | MapConstraint (Mapping Environment Environment)
-- a mapping is more or less a pair 
data Mapping a b = a :|-> b
 
-- Show-instances
instance (Show a, Show b) => Show (Mapping a b) where
 show (a :|-> b) = show a ++ " |-> " ++ show b

instance Show SolutionEntry where
  show (MapVarLift m) = show m
  show (MapVarLet m) = show m
  show (MapExpr m) = show m
  show (MapEnv (a :|-> b)) = show a ++ " |-> "  ++ show b
  show (MapCV (l:|-> r)) = showCV l ++ "|->" ++ show (r (MetaVarLift emptySubstitution "[.1]") (MetaExpression emptySubstitution "[.2]")) 
    where
      showCV (MetaChain sub AChain s v e)       = s ++ "[" ++ ".1" ++ "," ++ ".2" ++ "]"
      showCV (MetaChain sub VarChain s v e)     = s ++ "|" ++ ".1" ++ "," ++ ".2" ++ "|"
      showCV (ConstantChain sub AChain s v e)   = s ++ "(cnst)[" ++ ".1" ++ "," ++ ".2" ++ "]"
      showCV (ConstantChain sub VarChain s v e) = s ++ "(cnst)|" ++ ".1" ++ "," ++ ".2" ++ "|"
  show (MapEnvSpec (l :|-> r)) = show l ++ "|->" ++ show r
  show (MapConstraint (l :|-> r)) = show l ++ "|->" ++ show r


-- Substitute-instances
instance (Substitute b) => Substitute (Mapping a b) where
 subst e1 e2 (a :|-> b)         = a :|-> subst e1 e2 b
 substEmpty n (a :|-> b)        = a :|-> substEmpty n b 
 substEnv env n (a :|-> b)      = a :|-> substEnv env n b 
 substChainVar env n (a :|-> b) = a :|-> (substChainVar env n b)
 substEnvConstraint env n (a :|-> b) = a :|-> (substEnvConstraint env n b)

instance Substitute SolutionEntry where
 subst e1 e2 (MapVarLift m)         = MapVarLift (subst e1 e2 m) 
 subst e1 e2 (MapExpr m)           = MapExpr (subst e1 e2 m) 
 subst e1 e2 (MapEnv m)            = MapEnv (subst e1 e2 m)
 subst e1 e2 (MapEnvSpec m)        = MapEnvSpec (subst e1 e2 m)
 subst e1 e2 (MapConstraint m)     = MapConstraint (subst e1 e2 m)
 subst e1 e2 (MapCV (a :|-> fn))   = MapCV (a :|-> \z s -> (subst e1 e2 (fn z s)))
 -- 
 substEnv e1 e2 (MapVarLift m)      = MapVarLift (substEnv e1 e2 m) 
 substEnv e1 e2 (MapExpr m)        = MapExpr (substEnv e1 e2 m) 
 substEnv e1 e2 (MapEnv m)         = MapEnv (substEnv e1 e2 m)
 substEnv e1 e2 (MapEnvSpec m)     = MapEnvSpec (substEnv e1 e2 m)
 substEnv e1 e2 (MapConstraint m)     = MapConstraint (substEnv e1 e2 m)
 substEnv e1 e2 (MapCV (a :|-> fn))= MapCV (a :|-> \z s -> (substEnv e1 e2 (fn z s)))
 --
 substEmpty e2 (MapVarLift m)       = MapVarLift (substEmpty e2 m) 
 substEmpty e2 (MapExpr m)         = MapExpr (substEmpty e2 m) 
 substEmpty e2 (MapEnv m)          = MapEnv (substEmpty e2 m)
 substEmpty e2 (MapEnvSpec m)      = MapEnvSpec (substEmpty e2 m)
 substEmpty e2 (MapConstraint m)      = MapConstraint (substEmpty e2 m)
 substEmpty e2 (MapCV (a :|-> fn)) = MapCV (a :|-> \z s -> (substEmpty e2 (fn z s)))
 --
 substChainVar e1 e2 (MapVarLift m)  = MapVarLift (substChainVar e1 e2 m) 
 substChainVar e1 e2 (MapExpr m)    = MapExpr (substChainVar e1 e2 m) 
 substChainVar e1 e2 (MapEnv m)     = MapEnv (substChainVar e1 e2 m)
 substChainVar e1 e2 (MapEnvSpec m) = MapEnvSpec (substChainVar e1 e2 m)
 substChainVar e1 e2 (MapConstraint m) = MapConstraint (substChainVar e1 e2 m)
 substChainVar e1 e2 (MapCV (a :|-> fn))  = MapCV (a :|-> \z s -> (substChainVar e1 e2 (fn z s)))
  
 substEnvConstraint e1 e2 (MapVarLift m)      = MapVarLift (substEnvConstraint e1 e2 m) 
 substEnvConstraint e1 e2 (MapExpr m)        = MapExpr (substEnvConstraint e1 e2 m) 
 substEnvConstraint e1 e2 (MapEnv m)         = MapEnv (substEnvConstraint e1 e2 m)
 substEnvConstraint e1 e2 (MapEnvSpec m)     = MapEnvSpec (substEnvConstraint e1 e2 m)
 substEnvConstraint e1 e2 (MapConstraint m)  = MapConstraint (substEnvConstraint e1 e2 m)
 substEnvConstraint e1 e2 (MapCV (a :|-> fn)) = MapCV (a :|-> \z s -> (substEnvConstraint e1 e2 (fn z s)))
-- ---------------------------
-- several helper functions
-- ---------------------------

-- composition of the solution, iteratively apply the 
-- substitutions to the solution

composeAll :: Solution -> Solution  
composeAll [] = []  
composeAll maps = let (a:rest) = reverse maps 
                  in reverse $ go [a] rest 
 where 
  go maps [] = maps
  go maps (m@(MapVarLift (l :|-> r)):maps') = 
   go (m:(subst r l maps)) maps'
  go maps (m@(MapExpr (l :|-> r)):maps') = 
   go (m:(subst r l maps)) maps'
  go maps (m@(MapEnv ((Env  sub l) :|-> r)):maps') 
    | isEmpty r                    = go (m:substEmpty l maps) maps'
    | otherwise                    = go (m:(substEnv r l maps)) maps'
  go maps (m@(MapEnvSpec (l :|-> r)):maps') 
    | isEmpty r                    = go (m:substEmpty l maps) maps'
    | otherwise                    = go (m:(substEnv r l maps)) maps'
  go maps (m@(MapCV (l :|-> r)):maps') = 
    go (m:substChainVar r l maps) maps'
  go maps (m@(MapConstraint (l :|-> r)):maps') = 
   go (m:(substEnvConstraint r l maps)) maps'
-- --------------------------------------------------
-- applying a solution, implemented via a type class
-- --------------------------------------------------

class ApplySolution a where
  applySolution :: Solution -> a -> a
    
instance ApplySolution a => ApplySolution (Equation a) where
    applySolution s (a :=?=: b) = (applySolution s a) :=?=: (applySolution s b)
   
instance ApplySolution a => ApplySolution [a] where
    applySolution sol = map (applySolution sol)
instance (ApplySolution a, ApplySolution b) => ApplySolution (a,b) where
    applySolution s (a,b) = (applySolution s a, applySolution s b)

    
instance (ApplySolution a, ApplySolution b) => ApplySolution (Either a b) where
    applySolution s (Left a) = Left $ applySolution s a
    applySolution s (Right a) = Right $ applySolution s a

    
instance ApplySolution Gamma where
    applySolution sol gamma = 
     gamma{varLiftEq = applySolution sol (varLiftEq gamma)
          ,expressionEq = applySolution sol (expressionEq gamma)
          ,bindingEq = applySolution sol (bindingEq gamma)
          }
    
instance HasMetaVars Gamma where
    computeMetaVars gamma = 
         combineMetaVarsList [computeMetaVars (varLiftEq gamma)
                             ,computeMetaVars (expressionEq gamma)
                             ,computeMetaVars (bindingEq gamma)
                             ]
          
instance ApplySolution Delta where
     applySolution sol delta =
      delta{
            delta3  = applySolution  sol (delta3 delta)
           ,delta4  = applySolution  sol (delta4 delta)
           }     


 
instance ApplySolution VarLift where
 applySolution = applySolutionGen
 
instance ApplySolution Binding where
 applySolution = applySolutionGen
  
instance ApplySolution Environment where
 applySolution = applySolutionGen
   
instance ApplySolution Expression where    
 applySolution = applySolutionGen
 
applySolutionGen [] expr = expr
applySolutionGen (m@(MapVarLift (l :|-> r)):maps') expr = 
   applySolutionGen maps' (subst r l expr)
applySolutionGen (m@(MapExpr (l :|-> r)):maps') expr = 
   applySolutionGen maps' (subst r l expr)
applySolutionGen (m@(MapEnv ((Env sub l) :|-> r)):maps') expr
    | isEmpty r = applySolutionGen maps' (substEmpty l expr) 
    | otherwise = applySolutionGen maps' (substEnv r l expr)
applySolutionGen(m@(MapEnvSpec (l :|-> r)):maps') expr
    | isEmpty r = applySolutionGen maps' (substEmpty l expr) 
    | otherwise = applySolutionGen maps' (substEnv r l expr)
applySolutionGen(m@(MapCV (l :|-> r)):maps') expr =
    applySolutionGen maps' (substChainVar r l expr) 
applySolutionGen(m@(MapConstraint (l :|-> r)):maps') expr
    | otherwise = applySolutionGen maps' expr -- (substEnvConstraint r l expr)
        
   
-- ============================================================================

lookupMapVarLift v ((MapVarLift (l :|-> r)):maps) 
 | v == l = Just r
 | otherwise = lookupMapVarLift v maps
lookupMapVarLift v (m:ms) = lookupMapVarLift v ms
lookupMapVarLift v [] = Nothing
 
 
lookupMapExpr v ((MapExpr (l :|-> r)):maps) 
 | v == l = Just r
 | otherwise = lookupMapExpr v maps
lookupMapExpr v (m:ms) = lookupMapExpr v ms
lookupMapExpr v [] = Nothing

lookupMapEnv v ((MapEnv ((Env sub l) :|-> r)):maps)
 | v == l = Just r
 | otherwise = lookupMapEnv v maps
lookupMapEnv v ((MapEnvSpec (l :|-> r)):maps)
 | v == l = Just r
 | otherwise = lookupMapEnv v maps
lookupMapEnv v (m:ms) = lookupMapEnv v ms
lookupMapEnv v [] = Nothing


lookupMapCV (MetaChain sub typ ee z s) ((MapCV ((MetaChain sub' typ' ee' z' s') :|-> r)):maps)
 | ee == ee'  = Just r
lookupMapCV v (m:ms) = lookupMapCV v ms
lookupMapCV v [] = Nothing
 

lookupMapConstraint v ((MapConstraint (l :|-> r)):maps)
 | null [b | b <- bindings v, b `notElem` (bindings l)]
 , null [b | b <- metaEnvVars v, b `notElem` (metaEnvVars l)]
 , null [b | b <- metaChains v, b `notElem` (metaChains l)]
 , null [b | b <- constantEnvVars v, b `notElem` (constantEnvVars l)]
 , null [b | b <- constantChains v, b `notElem` (constantChains l)] = Just (l,r)
 | otherwise = lookupMapConstraint v maps
lookupMapConstraint v (_:maps) = lookupMapConstraint v maps
lookupMapConstraint v [] = Nothing




-- compute all meta variables

-- metaVarsVar (MetaVarLift sub name) = [(SCVar $ MetaVarLift sub name)]
-- metaVarsVar (ConstantVarLift sub name) = [(SCVar $ ConstantVarLift sub name)]
-- metaVarsVar (VarLift sub name) = []

-- metaVarsArgPos (VarPos _) = []
-- metaVarsArgPos (Binder v e) = (concatMap metaVarsVar v) ++ (metaVars e)

-- metaVars (VarL x) = metaVarsVar x
-- metaVars (Fn _ args) = concatMap metaVarsArgPos args
-- metaVars (Letrec env e) = (metaVars e) ++ (metaVarsEnv env)
-- metaVars e@(MetaExpression sub name) = [SCExpr e]
-- metaVars e@(Constant _ _) = [SCExpr e]
-- metaVars (Hole _) = []
-- metaVars d@(MetaCtxt _ _ _) = [SCCtxt d]
-- metaVars d@(ConstantCtxt _ _ _) = [SCCtxt d]
-- metaVars (MetaCtxtSubst a b c expr) = (SCCtxt (MetaCtxt a b c)):(metaVars expr)
-- metaVars (ConstantCtxtSubst a b c expr) = (SCCtxt (ConstantCtxt a b c)):(metaVars expr)

-- metaVarsEnv env =
      -- (concat [ (metaVarsVar x) ++ metaVars e |  (x := e) <- bindings env])
   -- ++ (concat [ (SCMetaChain sub typ name):(metaVarsVar z ++ metaVars s) | (MetaChain sub typ name z s) <- metaChains env])
   -- ++ (concat [ (SCConstantChain sub typ name):(metaVarsVar z ++ metaVars s) | (ConstantChain sub typ name z s) <- constantChains env])
   -- ++ [ (SCMetaEnv v) | v@(Env sub e) <- metaEnvVars env]
   -- ++ [ (SCConstantEnv v) | v@(Env sub e) <- constantEnvVars env]
   
standardSubst (f:fresh) nabla v@(SCVar (ConstantVarLift sub name)) = 
   (MapVarLift $ (ConstantVarLift sub name) :|-> (VarLift emptySubstitution ("x_" ++ f)), fresh)
   
standardSubst (f:fresh) nabla v@(SCExpr (Constant sub name)) = 
  let v = VarLift emptySubstitution ("x_" ++ f)
  in (MapExpr $ (Constant sub name) :|-> Fn "\\" [Binder [v] (VarL v)], fresh)

standardSubst (fresh) nabla v@(SCCtxt (ConstantCtxt sub typ name)) 
  | (ConstantCtxt sub typ name) `elem` (delta1 nabla)
    ||
    (ConstantCtxt emptySubstitution typ name) `elem` (delta1 nabla)
   = (MapExpr $ (ConstantCtxt sub typ name) :|-> (Fn "__internal_dummy_f" [Binder [] (Hole emptySubstitution)]), fresh)
  | otherwise = ((MapExpr $ (ConstantCtxt sub typ name) :|-> Hole emptySubstitution), fresh)
  
standardSubst (fresh) nabla v@(SCConstantChain sub typ name) =
   ((MapCV $ (ConstantChain sub typ name undefined undefined) :|-> \z s -> empty{bindings=[z := s]}),fresh) 
   
standardSubst (f:fresh) nabla v@(SCConstantEnv (Env sub e))
  |   (Env sub e) `elem` (delta2 nabla)
   || (Env emptySubstitution e) `elem` (delta2 nabla)
  = 
   let v = VarLift emptySubstitution ("x_" ++ f)
   in ((MapEnv $ (Env sub e) :|-> (empty{bindings =  [v := (VarL v)]})),fresh) 
  | otherwise = (MapEnv ((Env sub e) :|-> empty), (f:fresh))

mkStandardSubst fresh nabla [] = ([],fresh)
mkStandardSubst fresh nabla (v:vs) =
  let (sub,fresh')   = standardSubst fresh nabla v 
      (subs,fresh'') = mkStandardSubst fresh' nabla vs 
  in (sub:subs,fresh'')

   
  